/*
 * File: AR_Config.hpp
 * Created: 12.09.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#ifndef AR_CONFIG_HPP
#define AR_CONFIG_HPP


#include <string>
#include <stdlib.h>



class AR_Config {
	public:
		AR_Config(): m_home(getenv("HOME")), m_dir(m_home + "/.tcoffee/"),
			m_cache(m_dir + "cache/"), m_tmp(m_dir + "tmp/"), m_methods(m_dir + "methods/"),
			m_aln_width(60), m_extract_pdb("extract_pdb"), m_sap("sap")
		{
		}

		void read(char *env[]);

		void tc_cache(const std::string &cache) { m_cache = cache; }
		void tc_tmp(const std::string &tmp) { m_tmp = tmp; }
		void tc_methods(const std::string &meth) { m_methods = meth; }

		const std::string &home() const { return m_home; }
		const std::string &tc_dir() const { return m_dir; }
		const std::string &tc_cache() const { return m_cache; }
		const std::string &tc_tmp() const { return m_tmp; }
		const std::string &tc_methods() const { return m_methods; }
		const std::string &tc_master() const { return m_master; }
		const std::string &tc_defaults() const { return m_defaults; }
		const std::string &tc_extra_params() const { return m_extra_params; }
		unsigned int aln_width() const { return m_aln_width; }
		const std::string &extract_pdb() const { return m_extract_pdb; }
		const std::string &sap() const { return m_sap; }

	private:
		const std::string m_home;   // HOME
		std::string m_dir;          // TCOFFEE_DIR
		std::string m_cache;        // TCOFFEE_CACHE
		std::string m_tmp;          // TCOFFEE_TMP
		std::string m_methods;      // TCOFFEE_METHODS
		std::string m_master;       // TCOFFEE_MASTER
		std::string m_defaults;     // TCOFFEE_DEFAULTS
		std::string m_extra_params; // TCOFFEE_EXTRA_PARAMS
		unsigned int m_aln_width;   // TCOFFEE_ALN_WIDTH
		std::string m_extract_pdb;  // TCOFFEE_EXTRACT_PDB
		std::string m_sap;          // TCOFFEE_SAP
};


extern AR_Config g_conf;



#endif // AR_CONFIG_HPP
