/***
 *  $Id: FakeArr.hpp 770 2009-07-26 19:37:20Z zola $
 **
 *  File: FakeArr.hpp
 *  Created: Nov 18, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#ifndef AR_FAKEARR_HPP
#define AR_FAKEARR_HPP

#include <stdlib.h>
#include <string.h>


class FakeArr {
public:
    // must be consistent with TC_ConstraintList::entry_len
    enum { entry_size = 7 };

    FakeArr() : size(0), tab(0) { }
    FakeArr(int n) : size(0), tab(0) { }

    operator bool() const { return tab; }

    short int* operator[](int n) { return tab + entry_size * n; }
    short int* operator[](int n) const { return tab + entry_size * n; }

    void allocate(int s, int m) {
	this->free();
	tab = (short int*)malloc(s * m * sizeof(short int));
	size = s;
    } // allocate

    void reallocate(int s, int m) {
	int tmp_size = size + s;
	tab = (short int*)realloc(tab, tmp_size * m * sizeof(short int));
	size = tmp_size;
    } // reallocate

    void free() { 
	::free(tab);
	tab = 0;
	size = 0;
    } // free
        
    int size;
    short int* tab;
   
}; // class FakeArr

class FakeEntry {
public:
    // must be consistent with TC_ConstraintList::entry_len
    enum { entry_size = 7 };

    bool operator<(const FakeEntry& rhs) const {
	if (entry[0] < rhs.entry[0]) return true;
	if (rhs.entry[0] < entry[0]) return false;
	if (entry[1] < rhs.entry[1]) return true;
	if (rhs.entry[1] < entry[1]) return false;
	if (entry[2] < rhs.entry[2]) return true;
	if (rhs.entry[2] < entry[2]) return false;
	if (entry[3] < rhs.entry[3]) return true;
	if (rhs.entry[3] < entry[3]) return false;
	if (entry[4] < rhs.entry[4]) return true;
	if (rhs.entry[4] < entry[4]) return false;
	return false;
    }

    short int entry[entry_size];

}; // class FakeEntry

#endif // AR_FAKEARR_HPP
