/***
 *  $Id: JZ_residue_cache.hpp 706 2006-12-20 18:10:43Z zola $
 **
 *  File: JZ_residue_cache.hpp
 *  Created: Dec 15, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#ifndef JZ_RESIDUE_CACHE_HPP
#define JZ_RESIDUE_CACHE_HPP

#include <map>

struct res_key {
    res_key(int as1, int ar1, int as2, int ar2) 
	: s1(as1), r1(ar1), s2(as2), r2(ar2) { }

    bool operator<(const res_key& l) const {
	if (s1 < l.s1) return true;
	else if (l.s1 < s1) return false;
	else if (r1 < l.r1) return true;
	return false;
    } // operator<

    int s1, r1, s2, r2;
}; // struct res_key

struct ResidueCache {

    void insert(int s1, int r1, int s2, int r2, int s) {
	res_key k(s1, r1, s2, r2);
	storage.insert(std::pair<res_key, int>(k, s));
    } // insert

    bool find(int s1, int r1, int s2, int r2, int& s) {
	res_key k(s1, r1, s2, r2);
	std::map<res_key, int>::iterator iter(storage.find(k));
	if (iter != storage.end()) {
	    s = iter->second;
	    return true;
	}
	return false;
    } // insert

    std::map<res_key, int> storage;

}; // struct ResidueCache

#endif // JZ_RESIDUE_CACHE_HPP
