/*
 * File: TC_Alignment.hpp
 * Created: 06.06.2006
 * Last Modified: 23.08.2006
 * Author: Adrian Rospondek
 */


#ifndef TC_ALIGNMENT_HPP
#define TC_ALIGNMENT_HPP

#include "config.hpp"


struct TC_ConstraintList;
struct TC_Profile;
struct TC_Sequences;


struct TC_Alignment {
/* Size */
	int max_len;
	int min_len;   
	int *len;
	int declared_len;
	int max_n_seq;
	int nseq;
	int len_aln;

/* Sequence Information */
	char **file;
	char **seq_comment;
	char **aln_comment;
	char **name;

	char **expanded_order;
	char **tree_order;
	char **seq_al;
	int  **order;
	TC_Profile *P;
	TC_Sequences *S;
	TC_ConstraintList *CL;

	int **seq_cache; /* Contains the index of the residues. */
	/* The sequence numbering is relative to the sequences, and not to the alignmnent
	 * seq_cache[0][1] = 3
	 * indicates that in the aln residue (0) 1 corresponds to [order[0][0]][3]
	 * residues: 1...N
	 * sequences 0...M */
	int **cdna_cache; /* Contains the information about wheather a nucleotide is coding or not. */
					  /* Only defined if used */
/* Score */
	int *score_seq;
	int **score_res;
	int score_aln;
	int score;
		
	int cpu;
	int finished;

/* Input/Output Options */
	int output_res_num;
	int residue_case; /* 1 for lower, 0 for upper, 2 for keeping unchanged */
	int expand;

/* Must Not be copied */
	int used;
	int num;     
	int **pos; 

/* For linked lists */
	TC_Alignment *A;

/* Misc */
	int random_tag;
};



TC_Alignment *TC_declare_aln(TC_Sequences *S);
TC_Alignment *TC_declare_aln2(int nseq, int len);
TC_Alignment *TC_realloc_aln(TC_Alignment *A, int new_len);
TC_Alignment *TC_realloc_aln2(TC_Alignment *A, int n_nseq, int n_len);
TC_Alignment *TC_copy_aln(const TC_Alignment *A, TC_Alignment *B);
TC_Sequences *TC_free_aln(TC_Alignment *LA);

TC_Alignment *TC_seq2aln(TC_Sequences *S, TC_Alignment *A, int rm_gap);
TC_Alignment *TC_stack_aln(TC_Alignment *A, TC_Alignment *B);

TC_Alignment *TC_get_aln_type(TC_Alignment *A);
TC_Alignment *TC_reorder_aln(TC_Alignment *A, char **name, int nseq);
TC_Alignment *TC_padd_aln(TC_Alignment *A);
TC_Alignment *TC_ungap_aln(TC_Alignment *A);
TC_Alignment *TC_ungap_aln_n(TC_Alignment *A, int p);
TC_Alignment *TC_ungap_sub_aln(TC_Alignment *A, int ns, int *ls);
void TC_compress_aln(TC_Alignment *A);
TC_Alignment *TC_trim_aln_with_seq(TC_Alignment *S, TC_Alignment *P);
TC_Alignment *TC_extract_sub_aln(const TC_Alignment *B, int nseq, int *list);
TC_Alignment *TC_fix_aln_seq(TC_Alignment *A, const TC_Sequences *S);
TC_Alignment *TC_expand_aln(TC_Alignment *A);
TC_Alignment *TC_update_aln_random_tag(TC_Alignment *A);

int TC_analyse_aln_column(TC_Alignment *B, int col);
int **TC_aln2pos_simple(const TC_Alignment *A, int n_nseq, ...);
char *TC_aln2cons_seq(const TC_Alignment *A, int ns, int *ls, int n_groups, char **group_list);
char *TC_aln2cons_seq_mat(const TC_Alignment *A, const char *mat_name);
char *TC_sub_aln2cons_seq_mat(const TC_Alignment *A, int ns, int *ls, const char *mat_name);
int **TC_aln2count_mat(const TC_Alignment *A);
int **TC_aln2count_mat2(const TC_Alignment *A);
int **TC_sub_aln2count_mat2(const TC_Alignment *A, int ns, int *l_s);
int **TC_seq2aln_pos(const TC_Alignment *A, int *ns, int **l_s);

int TC_output_format_aln(const char *format, const TC_Alignment *inA, const TC_Alignment *inEA,
		const char *name);

#if DEBUG_TREE
void print_A(TC_Alignment* A);
#else
#define print_A(A)
#endif

#endif // TC_ALIGNMENT_HPP
