/*
 * File: TC_ConstraintList.hpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#ifndef TC_CONSTRAINTLIST_HPP
#define TC_CONSTRAINTLIST_HPP

#include <numeric>
#include <string>
#include <vector>
#include "config.hpp"
#include "dp_drivers.hpp"
#include "FakeArr.hpp"
#include "evaluate.hpp"
#include "ResidueIndex.hpp"



#define SEQ1 0
#define SEQ2 1
#define R1 2
#define R2 3
#define WE 4
#define CONS 5
#define MISC 6



class MPI_Env;
struct TC_Alignment;
struct TC_DistanceMatrix;
struct TC_Method;
struct TC_Sequences;
struct TC_Weights;


struct TC_ConstraintList {

    TC_ConstraintList() {
	copy_mode = 0;
	pCL = 0;
	S = 0;
	STRUC_LIST = 0;
	L = 0;
	W = 0;
	DM = 0;
	ktupDM = 0;
	translation = 0;
	packed_seq_lu = 0;
	ne = 0;
	M = 0;
	normalise = 0;
	max_ext_value = 0;
	max_value = 0;
	overweight = 0;
	pw_parameters_set = 0;
	gop = 0;
	gep = 0;
	f_gop = 0;
	f_gep = 0;
	nm_gop = 0;
	nm_gep = 0;
	nomatch = 0;
	TG_MODE = 0;
	F_TG_MODE = 0;
	maximise = 0;
	diagonal_threshold = 0;
	ktup = 0;
	use_fragments = 0;
	fasta_step = 0;
	lalign_n_top = 0;
	sw_min_dist = 0;
	matrices_list = 0;
	n_matrices = 0;
	tree_aln = 0;

	get_dp_cost = 0;
	profile_mode = 0;
	evaluate_residue_pair = 0;
	pair_wise = 0;

	weight_field = 0;
	max_n_pair = 0;

	split = 0;
	split_nseq_thres = 0;
	split_score_thres = 0;
	
	forbiden_pair_list = 0;
	
	seq_for_quadruplet = 0;
	nseq_for_quadruplet = 0;

	extend_jit = 0;
	extend_threshold = 0;
	do_self = 0;

	residue_indexed = 0;

#if USE_RMA
#else
	residue_index = 0;
#endif

	residue_field = 0;

	seq_indexed = 0;
	start_index = 0;
	end_index = 0;
	max_L_len = 0;
	chunk = 0;

	cpu = 0;

	memset(align_pdb_param_file, 0, 100);
	memset(align_pdb_hasch_mode, 0, 100);
	memset(dp_mode, 0, 100);
	memset(matrix_for_aa_group, 0, 100);
	memset(tree_mode, 0, 100);
	memset(distance_matrix_mode, 0, 100);
	memset(profile_comparison, 0, 100);
	memset(evaluate_mode, 0, 100);
	memset(extend_clean_mode, 0, 100);
	memset(extend_compact_mode, 0, 100);
	memset(lib_list, 0, 100);
    } // TC_ConstraintList




	int copy_mode;
	const TC_ConstraintList *pCL; 
	TC_Sequences *S; // Total sequences
	TC_Sequences *STRUC_LIST; // Name of the sequences with a Structure
	char align_pdb_param_file[100];
	char align_pdb_hasch_mode[100];
	TC_Weights *W; // Sequence Weights
	TC_DistanceMatrix *DM; // Accurate Distance Matrix
	TC_DistanceMatrix *ktupDM; // Fast Distance Matrix
	int *translation;   
	int **packed_seq_lu; // Packing Sequence: To use with domain analysis

/* List Information */
	int ne; // Number of elements in the list
	static const unsigned int entry_len = 7; // Size of an entry

/* DATA */
	FakeArr L; // Library
	int **M; // Substitution matrix

/* Normalisation information */
	int normalise;
	int max_ext_value;
	int max_value;
	int overweight;

/* Pair wise alignment method */
	int pw_parameters_set;
	int gop;
	int gep;
	int f_gop;
	int f_gep;
	int nm_gop;
	int nm_gep;
	int nomatch;
	int TG_MODE;
	int F_TG_MODE;
	char dp_mode[100];
	int maximise;
	char matrix_for_aa_group[100];
	float diagonal_threshold;
	int ktup;
	int use_fragments;
	int fasta_step;
	int lalign_n_top;
	int sw_min_dist;
	char **matrices_list;
	int n_matrices;
	char tree_mode[100];
	char distance_matrix_mode[100];
	TC_Alignment *tree_aln;

/* Functions used for Dynamic Programming and Evaluation */
/* 1: Function for evaluating the cost of a column */
	TC_col_cost_func_type get_dp_cost;
	TC_prf_cost_func_type profile_mode;
	char profile_comparison [100];
/* 2: Function for evaluating the cost of a pair of residues */
	TC_pair_cost_func_type evaluate_residue_pair;
/* 3: Function for making dynamic programming */
	TC_pw_func_type pair_wise;

	int weight_field;
	int max_n_pair; // Maximum number of pairs when aligning two profiles

/* Split parameters */
	int split;
	int split_nseq_thres;
	int split_score_thres;

	char evaluate_mode[100];

/* Functions for hiding forbiden pairs of residues */
	int ****forbiden_pair_list;
	// pair_list[S1][S2][L1][L2] = 1 -> forbiden
	// pair_list[S1][S2][L1][L2] = 0 -> allowed
	// pair_list[S1][S2][L1] = NULL  -> all pairs S1L1, S2 allowed
	// S -> sequences, 0..N
	// L -> residues, 1..L-1

/* Extention properties: copy */
	int *seq_for_quadruplet;
	int nseq_for_quadruplet;

/* Extention properties: Do Not copy */
	int extend_jit; // Extend only on request
	int extend_threshold; // Do not extend pairs below the Theshold
	int do_self; // Extend a sequence against itself
	char extend_clean_mode[100];  
	char extend_compact_mode[100];

/* Lookup table parameteres */
/* !! Do not copy in duplication !! */
	// Residue Index contains residue_index[nseq][seq_len][0]->number of links
	// [seq][res][x  ] -> target seq (0->N-1)
	// [seq][res][x+1] -> traget res (1->len
	// [seq][res][x+2] -> target weight
	// It is automatically recomputed when L residue_indexed is set to 0
	int residue_indexed;

    // TODO: replace by RT mechanism
#if USE_RMA
        ResIdx residue_index;
#else
	int ***residue_index;
#endif

	int residue_field;	

/* Index of the pairs of sequences within L */
	int seq_indexed;
	int **start_index;
	int **end_index;
	int max_L_len;
	int chunk;

/* MISC */
	char lib_list[100];
	int cpu;
};



struct AR_lib_job {

    AR_lib_job(): method(-1)
    {
	seq[0] = seq[1] = -1;
    }

    int method;
    int seq[2];
};



class AR_job_cmp {
public:
    explicit AR_job_cmp(const std::vector<int> &jc): m_job_cnt(jc)
    {
    }

    bool operator()(const AR_lib_job &lhs, const AR_lib_job &rhs) const
    {
	return m_job_cnt[lhs.method] > m_job_cnt[rhs.method];
    }

private:
    const std::vector<int> &m_job_cnt;
};



TC_ConstraintList *TC_declare_constraint_list(TC_Sequences *S, short int **L, int ne, int **M);
TC_ConstraintList *TC_duplicate_constraint_list(const TC_ConstraintList *CL);
TC_ConstraintList *TC_duplicate_constraint_list_soft(const TC_ConstraintList *CL);
TC_Sequences *TC_free_constraint_list(TC_ConstraintList *CL);
void TC_free_constraint_list_full(TC_ConstraintList *CL);

short int TC_vwrite_clist(TC_ConstraintList *CL, int a, int b, short int x);
int TC_vread_clist(const TC_ConstraintList *CL, int a, int b);
short int *TC_extract_entry(short int *entry, int pos, const TC_ConstraintList *CL);
TC_ConstraintList *TC_add_entry2list(const short int *entry, TC_ConstraintList *CL);

TC_ConstraintList *TC_aln2constraint_list(const TC_Alignment *A, TC_ConstraintList *CL, const char *in_weight_mode);

TC_ConstraintList *TC_filter_list(TC_ConstraintList *CL, int start, int len, int T);
TC_ConstraintList *TC_weight_constraint_list(TC_ConstraintList *CL, const char *seq_weight);
TC_ConstraintList *TC_evaluate_constraint_list_reference(TC_ConstraintList *CL);
TC_ConstraintList *TC_sort_constraint_list(TC_ConstraintList *CL, int start, int len);
TC_ConstraintList *TC_sort_constraint_list_on_n_fields(TC_ConstraintList *CL, int start, int len, int first_field,
		int n_fields);
TC_ConstraintList *TC_sort_constraint_list_inv(TC_ConstraintList *CL, int start, int len);
TC_ConstraintList *TC_compact_list(TC_ConstraintList *CL, const int start, const int len,
		const char *compact_mode);
TC_ConstraintList *TC_merge_constraint_list(const TC_ConstraintList *SL, TC_ConstraintList *ML);
TC_ConstraintList *TC_index_constraint_list(TC_ConstraintList *CL);
TC_ConstraintList *TC_index_res_constraint_list(TC_ConstraintList *CL, int field);
int **TC_main_search_in_list_constraint(int *key, int *p, int k_len, TC_ConstraintList *CL);


void AR_method_list2job_list(const std::vector<TC_Method *> &meth_list, const TC_Sequences *S,
			     const TC_DistanceMatrix *DM, const char *lib_list, std::vector<AR_lib_job> &job_list, std::vector<int> &job_cnt);


TC_ConstraintList *AR_par_compile_constraint_list(TC_ConstraintList *CL, const std::vector<std::string> &in_files,
						  const char *weight, const MPI_Env &mpi_env);

void AR_send_jobs(const int recv, const MPI_Env &mpi_env, unsigned int chunk_size, unsigned int &curr_job,
		  const std::vector<AR_lib_job> &job_list, int *ibuf);


void AR_process_jobs(const unsigned int n_jobs, const int *job_desc, const std::vector<TC_Method *> &methods,
		     const std::vector<TC_ConstraintList *> &libs);


#if USE_RMA
TC_ConstraintList* AR_index_res_constraint_list(TC_ConstraintList*, int);
#endif

#if DEBUG_TREE
void print_CL(TC_ConstraintList* CL);
#else
#define print_CL(CL)
#endif


#endif // TC_CONSTRAINTLIST_HPP
