/*
 * File: TC_Sequences.hpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#ifndef TC_SEQUENCES_HPP
#define TC_SEQUENCES_HPP


#include <iostream>
#include <vector>
#include <stdio.h>



struct TC_Alignment;
struct TC_ConstraintList;
struct TC_Templates;
struct TC_Weights;
struct TC_X_template;


struct TC_Sequences {
	char **file; // file[Nseq][FILENAMELEN] name of the file that contributed each sequence 
	char **seq_comment; // seq_comment[Nseq][LONG_STRING] comment read in the file
	char **aln_comment; // id
	char **seq; // seq[Nseq][sequence] sequences
	int *len; // len[Nseq] length of each sequence
	int max_len; // Lenght of the longest seq
	int min_len; // Length of the shortest seq
	int nseq; // nseq
	int max_nseq; // Maximum number of sequences in the datastruct
	char **name; // name[Nseq][MAXNAMELEN]
	TC_ConstraintList *CL; // Constraint list
	int contains_gap; // set to 1 if gaps are to be kept
	char type[15]; // PROTEIN, DNA
	TC_Weights *W; // Associated weights
	TC_Templates **T;
};



TC_Sequences *TC_declare_sequence(int min, int max, int nseq);
TC_Sequences *TC_duplicate_sequence(const TC_Sequences *S);
TC_Sequences *TC_realloc_sequence(TC_Sequences *OUT, int new_nseq, int max_len);
void TC_free_sequence(TC_Sequences *LS);

TC_Sequences *TC_fill_sequence_struc(int nseq, char **sequences, char **seq_name);
TC_Sequences *TC_aln2seq(const TC_Alignment *A);

int TC_seq_name2index(const char *name, const TC_Sequences *S);
TC_Sequences *TC_seq2unique_name_seq(TC_Sequences *S);
TC_Sequences *AR_seq2template_seq(TC_Sequences *S, const char *template_list,
		std::vector<TC_X_template *> &remote_templates);
TC_Sequences *AR_profile_seq2template_seq(TC_Sequences *S, const char *template_list,
		std::vector<TC_X_template *> &remote_templates);
TC_Sequences *TC_seq2template_type(TC_Sequences *Seq);

int TC_seq2n_X_template(const TC_Sequences *S, const char *type);
int TC_prf_in_seq(const TC_Sequences *S);
TC_X_template *TC_seq_has_template(const TC_Sequences *S, int n, const char *mode);
TC_Alignment *TC_seq2R_template_profile(const TC_Sequences *S, int n);
char *TC_seq2P_template_file(const TC_Sequences *S, int n);

TC_Sequences *TC_get_sequence_type(TC_Sequences *S);
TC_Sequences *TC_ungap_seq(TC_Sequences *S);
TC_Sequences *TC_reorder_seq(TC_Sequences *A, char **name, int nseq);
TC_Sequences *TC_merge_seq(const TC_Sequences *IN, TC_Sequences *OUT);
TC_Sequences *TC_add_sequence(const TC_Sequences *IN, TC_Sequences *OUT, int i);
int **TC_index_seq_name(const TC_Sequences *S1, const TC_Sequences *S2);
int **TC_index_seq_res(const TC_Sequences *S1, const TC_Sequences *S2, int **name_index);

FILE *TC_display_sequences_names(const TC_Sequences *S, FILE *fp, int check_pdb_status, int print_templates);
std::ostream &TC_display_sequence_templates(const TC_Sequences *S, int i, std::ostream &out);



#endif // TC_SEQUENCES_HPP
