/***
 *  $Id: cache_manager.hpp 710 2007-01-11 22:28:40Z zola $
 **
 *  File: cache_manager.hpp
 *  Created: Dec 19, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#ifndef JZ_CACHE_MANAGER_HPP
#define JZ_CACHE_MANAGER_HPP

#include <algorithm>
#include <iostream>
#include <vector>
#include <map>
#include <math.h>


static inline double my_timer() {
    static double time = 0.0;
    time += 0.001;
    return time;
} // my_timer


template <typename T> struct policy {
    bool operator()(const T& lhs, const T& rhs) const {
//	return lhs->second.count < rhs->second.count;
//	return lhs->second.gds < rhs->second.gds;
	return rhs->second.time < lhs->second.time;
    } // operator()
}; // lfu


template <typename T> struct entry {
    entry() : data(), size(0), time(0.0) { }

    T data;
    unsigned int size;

    double time;

//     double log_size;
//     double gds;
//     unsigned int count;
    
    void rm() {
	delete data.first;
	delete[] data.second;
    }
}; // entry


template <typename Key, typename Data>
class cache_manager {
public:
    typedef entry<Data> entry_type;
    typedef std::map<Key, entry_type> storage_type;
    typedef typename storage_type::iterator storage_iterator;


    cache_manager(std::size_t C = 0) 
	: capacity_(C), size_(0), update_(true), P_(), keys_(), storage_() { }


    storage_iterator begin() { return storage_.begin(); }
    
    storage_iterator end() { return storage_.end(); }


    bool empty() const { return storage_.empty(); }
    
    void clear() {
	size_ = 0;
	update_ = true;

	keys_.clear();
	
	storage_iterator iter(storage_.begin());
	for (; iter != storage_.end(); ++iter) iter->second.rm();

	storage_.clear();
    } // clear


    void insert(Key key, const Data& data, std::size_t size) {
	// clean storage
	if (capacity_ < size_ + size + 1) {
	    std::size_t rm_cnt = 0;

	    if (update_ == true) {
		std::make_heap(keys_.begin(), keys_.end(), P_);
	    }

	    do {
		size_ -= keys_[0]->second.size;
		if (*(keys_[0]->second.data.first) == 0) {
		    keys_[0]->second.rm();
		}
		else {
		    *(keys_[0]->second.data.first) = 2; // mark to be removed
		}
		storage_.erase(keys_[0]);
		std::pop_heap(keys_.begin(), keys_.end() - rm_cnt, P_);
		++rm_cnt;
	    }
	    while (capacity_ < size_ + size + 1);

	    keys_.erase(keys_.end() - rm_cnt, keys_.end());
	    update_ = false;
	}

	// insert new element
	size_ += size;

	entry_type e;
	e.data = data;
	e.size = size;

	e.time = my_timer();
// 	e.log_size = log(size);
// 	e.count = 1;
// 	e.gds = 1 / e.log_size;

	std::pair<Key, entry_type> dat(key, e);
	std::pair<storage_iterator, bool> res(storage_.insert(dat));
	
	keys_.push_back(res.first);
	
	if (update_ == false) std::push_heap(keys_.begin(), keys_.end(), P_);
    } // insert

    bool find(Key key, Data& data) {
	storage_iterator iter(storage_.find(key));
	if (iter == storage_.end()) return false;

	data = iter->second.data;

//	++(iter->second.count);
//	iter->second.gds = iter->second.count / iter->second.log_size;
	iter->second.time = my_timer();

	update_ = true;

	return true;
    } // find


private:
    std::size_t capacity_;
    std::size_t size_;
    
    bool update_;
    
    policy<storage_iterator> P_;

    std::vector<storage_iterator> keys_;
    storage_type storage_;

}; // cache_manager

#endif // JZ_CACHE_MANAGER_HPP
