/*
 * File: evaluate.hpp
 * Created: 06.06.2006
 * Last Modified: 26.08.2006
 * Author: Adrian Rospondek
 */


#ifndef EVALUATE_HPP
#define EVALUATE_HPP



struct TC_Alignment;
struct TC_ConstraintList;


typedef int (*TC_pair_cost_func_type)(TC_ConstraintList *, int, int, int, int);
typedef int (*TC_col_cost_func_type)(TC_Alignment *, int **, int, int *, int, int **, int, int *, int,
		TC_ConstraintList *);
typedef int (*TC_prf_cost_func_type)(const int *, const int *, TC_ConstraintList *);



int TC_evaluate_matrix_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2);
int TC_evaluate_cdna_matrix_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2);
int TC_evaluate_aln_profile_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2);
int TC_extend_residue_pair(TC_ConstraintList *CL, int s1, int r1, int s2, int r2);
int TC_residue_pair_extended_list(TC_ConstraintList *CL, int s1, int r1, int s2, int r2);

int TC_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2, int ns2,
		int *list2, int col2, TC_ConstraintList *CL);
int TC_fast_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
		int ns2, int *list2, int col2, TC_ConstraintList *CL);
int TC_fast_get_dp_cost_2(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
		int ns2, int *list2, int col2, TC_ConstraintList *CL);
int TC_slow_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
		int ns2, int *list2, int col2, TC_ConstraintList *CL);
int TC_consensus_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
		int ns2, int *list2, int col2, TC_ConstraintList *CL);
int TC_cw_profile_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
		int ns2, int *list2, int col2, TC_ConstraintList *CL);

TC_ConstraintList *TC_choose_extension_mode(const char *extend_mode, TC_ConstraintList *CL);


int TC_cw_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL);
int TC_muscle_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL);
int TC_channel_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL);

TC_prf_cost_func_type TC_get_profile_mode_function(char *name, TC_prf_cost_func_type func);


int TC_sub_aln2sub_aln_score(TC_Alignment *A, TC_ConstraintList *CL, int *ns, int **ls);


TC_Alignment *TC_main_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL, const char *mode);
TC_Alignment *TC_categories_evaluate_output(const TC_Alignment *IN);
TC_Alignment *TC_matrix_evaluate_output(const TC_Alignment *IN, TC_ConstraintList *CL);
TC_Alignment *TC_sar_evaluate_output(const TC_Alignment *IN, TC_ConstraintList *CL);
TC_Alignment *TC_boxshade_evaluate_output(const TC_Alignment *IN);
TC_Alignment *TC_fast_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL);
TC_Alignment *TC_slow_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL);


float TC_get_avg_matrix_mm(int **matrix, const char *alphabet);

int **TC_clean_residue_pair_hasch(int s1, int r1, int s2, int r2, int **hasch,
				  TC_ConstraintList *CL);

#endif // EVALUATE_HPP
