/*
 * File: globals.hpp
 * Created: 06.06.2006
 * Last Modified: 28.08.2006
 * Author: Adrian Rospondek
 */


#ifndef GLOBALS_HPP
#define GLOBALS_HPP


#include <ctype.h>
#include <float.h>
#include "util_str.hpp"



#define PROGRAM "Parallel-T-COFFEE"
#define VERSION "1.914"
#define DATE ""
#define AUTHOR "Adrian Rospondek, Jaroslaw Zola, Xiao Yang"
#define URL ""
#define MAIL "a.rospondek@poczta.fm, jaroslaw.zola@gmail.com, xyang@iastate.edu"

#define FALSE 0
#define TRUE 1

#define MY_EPS (1000 * DBL_EPSILON)

#define MAX_N_PARAM 2000
#define MAX_N_SEQ 1
#define MAX_LEN_ALN 1
#define MAXNAMES 100
#define STRING 300
#define LONG_STRING 1000
#define VERY_LONG_STRING 10000
#define FILENAMELEN 500

#define AA_ALPHABET "acdefghiklmnpqrstvwy-ACDEFGHIKLMNPQRSTVWY"
#define DNA_ALPHABET "AGCTUNRYMKSWHBVD-agctunrymkswhbvd"
#define GAP_LIST "-.#*~"
#define NAMES_ALPHABET "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_|-!%@&#-+="

#define UNGAPED_POSITIONS 1
#define ALIGNED_POSITIONS 2
#define AVERAGE_POSITIONS 3

#define GAP_CODE 60
#define MAXID 100
#define SCORE_K 10
#define PAVIE_MAT_FACTOR 1000

#define SOFT_COPY 1
#define HARD_COPY 2

#define CHECK 0
#define NO_CHECK 1
#define FORCE 2
#define STORE 3
#define FLUSH 4
#define GET 7
#define ON 8
#define OFF 9

#define ALLOWED 0
#define FORBIDEN -9999
#define OPTIONAL 1

#define TC_UNDEFINED FORBIDEN
#define UNDEFINED_2 0
#define UNDEFINED_RESIDUE '>'

#define LOWER_CASE 0
#define UPPER_CASE 1
#define KEEP_CASE 2
#define CHANGE_CASE 3

#define KEEP_NAME 1
#define KEEP_GAP 0
#define RM_GAP 1

#define NOTEMPLATES 0
#define TEMPLATES 1
#define NO_COLOR_RESIDUE 127

#define SEPARATORS "\n \t,;"
#define NULL_DEVICE "/dev/null"
#define TO_NULL_DEVICE ">/dev/null 2>&1"

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#define FABS(x) ((x < 0) ? (-x) : (x))
#define SWAPP(x, y, tp) { tp = y; y = x; x = tp; }
#define GET_CASE(f, c) ((f == UPPER_CASE) ? toupper(c) : ((f == LOWER_CASE) ? tolower(c) : c))

#define TC_a_better_than_b(x, y, m) ((m == 1) ? (((x) > (y)) ? 1 : 0) : (((x) < (y)) ? 1 : 0))
#define TC_best_of_a_b(x, y, m) ((m == 1) ? MAX((x), (y)) : MIN((x), (y)))

#define TC_is_in_range(x, min, max) ((min <= x && x <= max) ? 1 : 0)
#define TC_is_parameter(x) (x[0] == '-' && !isdigit(x[1])) 

#define TC_strm(x, y) (TC_vstrcmp((x), (y)) == 0)
#define TC_strm2(a, b, c) (TC_strm(a, b) || TC_strm(a, c))
#define TC_strm3(a, b, c, d) (TC_strm2(a, b, c) || TC_strm(a, d))
#define TC_strm4(a, b, c, d, e) (TC_strm2(a, b, c) || TC_strm2(a, d, e))
#define TC_strm5(a, b, c, d, e, f) (TC_strm2(a, b, c) || TC_strm3(a, d, e, f))
#define TC_strm6(a, b, c, d, e, f, g) (TC_strm3(a, b, c, d) || TC_strm3(a, e, f, g))
#define TC_strnm(x, y, n) (TC_vstrncmp((x), (y), (n)) == 0)



#endif // GLOBALS_HPP
