/*
 * File: input.hpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#ifndef INPUT_HPP
#define INPUT_HPP


#include <string>
#include <vector>
#include "TC_tree.hpp"



struct TC_Alignment;
struct TC_ConstraintList;
struct TC_Sequences;
struct TC_Weights;



void AR_identify_list_format(std::vector<std::string>::iterator beg, std::vector<std::string>::iterator end);
char *TC_name2type_name(const char *name);

int TC_is_seq(const char *name);
int TC_is_aln(const char *name);
int TC_is_lib(const char *name);
int TC_is_lib_01(const char *name);
int TC_is_matrix(const char *name);
int TC_is_blast_matrix(const char *name);
int TC_is_clustalw_matrix(const char *name);
int TC_is_pavie_matrix(const char *name);
int TC_is_method(const char *file);
int TC_is_pdb_name(const char *name);
inline int AR_is_pdb_name(const std::string &name)
{
	return TC_is_pdb_name(name.c_str());
}
int TC_is_pdb_file(const char *name);
inline int AR_is_pdb_file(const std::string &name)
{
	return TC_is_pdb_file(name.c_str());
}
int TC_is_single_seq_weight_file(const char *name);

char *TC_identify_seq_format(const char *file);

int TC_format_is_fasta(const char *file);
int TC_format_is_fasta_seq(const char *file);
int TC_format_is_fasta_aln(const char *file);
int TC_format_is_pir(const char *file);
int TC_format_is_pir_seq(const char *file);
int TC_format_is_pir_aln(const char *file);
int TC_format_is_msf(const char *file);
int TC_format_is_swissprot(const char *name);
int TC_format_is_saga(const char *file);
int TC_format_is_oligo(const char *file);


TC_Sequences *AR_read_seq_from_files(std::vector<std::string>::iterator files_beg,
		std::vector<std::string>::iterator files_end, const char *type, const char *SeqMode);
TC_Sequences *TC_main_read_seq(const char *name);
TC_Sequences *TC_get_fasta_sequence(const char *fname);
TC_Sequences *TC_get_pir_sequence(const char *fname);
TC_Sequences *TC_get_swissprot_sequence(const char *fname);
TC_Sequences *TC_get_pdb_sequence(const char *fname);
int TC_read_seq_in_list(const char *fname, int *nseq, char ***sequences, char ***seq_name);

TC_Alignment *TC_main_read_aln(const char *name, TC_Alignment *A);
void TC_read_aln(const char *file_name, TC_Alignment *A);
void TC_read_msf_aln(const char *fname, TC_Alignment *A);

TC_ConstraintList *TC_read_constraint_list(TC_ConstraintList *CL, const char *fname, const char *mode,
		const char *weight_mode);
TC_ConstraintList *TC_read_constraint_list_file(TC_ConstraintList *CL, const char *fname);
TC_ConstraintList *TC_fast_read_constraint_list_file(TC_ConstraintList *CL, const char *fname);
TC_ConstraintList *TC_aln_file2constraint_list(const char *alname, TC_ConstraintList *CL, const char *weight_mode);

int **TC_read_matrice(const char *mat_name_in);
int **TC_read_blast_matrix(const char *mat_name);
int **TC_read_pavie_matrix(const char *mat_name);

char *TC_get_pdb_id(const char *file);
inline char *AR_get_pdb_id(const std::string &file)
{
	return TC_get_pdb_id(file.c_str());
}

TC_Weights *TC_read_seq_weight(char **name, int nseq, const char *seq_weight);

TC_NT_node **TC_read_tree(const char *treefile, int *tot_node, int nseq, char **seq_names);



#endif // INPUT_HPP
