/***
 *  $Id: par_tree_aln.hpp 714 2007-01-29 15:37:47Z zola $
 **
 *  File: par_tree_aln.hpp
 *  Created: Nov 06, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Xiao Yang <xyang@iastate.edu>
 *          Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#ifndef PTC_PAR_TREE_ALN_HPP
#define PTC_PAR_TREE_ALN_HPP

#include <algorithm>
#include <cstdlib>
#include <deque>
#include <new>
#include <queue>
#include <vector>
#include <string.h>
#include <MPI_ibuffer.hpp>
#include <MPI_obuffer.hpp>
#include "config.hpp"
#include "globals.hpp"
#include "JZ_ConstraintList.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "TC_tree.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"
#include "util_str.hpp"
#include "time_log.hpp"


struct JobNode {
    JobNode() : id(-1), pNode(0), tot_leaf(0), l_leaf_num(0), r_leaf_num(0), 
		parent(0), l_child(0), r_child(0),
		in_jobs(-1), l_0(0), l_1(0), name("") {
    } // JobNode

    int id;
    TC_NT_node pNode;
    int tot_leaf;
    int l_leaf_num;
    int r_leaf_num;
    JobNode* parent;
    JobNode* l_child;
    JobNode* r_child;
    short int in_jobs; // number of precedence jobs finished
    int* l_0;
    int* l_1;

    std::string name;

}; // struct JobNode

typedef JobNode* JobNodePtr;
typedef std::vector<JobNode> job_list_type;

enum { PA_JOB_TAG = 113, PA_RES_TAG = 213 };


#if USE_TIMER
extern yat::TimeLog TL;
#endif


struct host_cmp : public std::binary_function<int, int, bool> {

    bool operator()(int lhs, int rhs) const {
	int l_mod = lhs % ptc_proc;
	int r_mod = rhs % ptc_proc;

	if (r_mod < l_mod) return true;
	if (l_mod < r_mod) return false;

	return lhs < rhs;
    }

    int ptc_proc;
};


// should lhs be scheduled before rhs
inline bool job_node_compare(const JobNodePtr lhs, const JobNodePtr rhs) {
    if (rhs->in_jobs < lhs->in_jobs) return true;
    if (lhs->in_jobs < rhs->in_jobs) return false;
    
    // if (lhs->id < rhs->id) return true;

    // faked tree
    int lhs_nl = lhs->tot_leaf;
    int rhs_nl = rhs->tot_leaf;
    
    if (lhs_nl < rhs_nl) return true;
    if (rhs_nl < lhs_nl) return false;

    // real tree
    int lhs_bal = lhs->l_leaf_num * lhs->r_leaf_num;
    int rhs_bal = rhs->l_leaf_num * rhs->r_leaf_num;

    if (lhs_bal < rhs_bal) return true;
    
    // here we can add another condition
    // based on length of alignment
    // ...

    return lhs->id < rhs->id;
} // job_node_compare

inline bool job_node_compare_neg(const JobNodePtr lhs, const JobNodePtr rhs) {
    return !job_node_compare(lhs, rhs);
} // job_node_compare_neg


std::size_t traverse_phyl_tree(TC_NT_node root, job_list_type& j_list);

TC_NT_node* master_par_tree_aln(TC_NT_node** T, TC_Alignment*& A,
				TC_ConstraintList*& CL, const MPI_Env& mpi_env);

void worker_par_tree_aln(TC_ConstraintList*& CL, const MPI_Env& mpi_env);

TC_NT_node* AR_par_tree_aln(TC_NT_node** T, TC_Alignment*& A,
			    TC_ConstraintList*& CL,
			    const MPI_Env& mpi_env);

#endif // PTC_PAR_TREE_ALN_HPP
