/*
 * File: util_mpi.hpp
 * Created: 25.07.2006
 * Last Modified: 23.08.2006
 * Author: Adrian Rospondek
 */


#ifndef UTIL_MPI_HPP
#define UTIL_MPI_HPP


#include <vector>
#include <mpi.h>
#include <MPI_env.hpp>



struct TC_Sequences;


enum MPI_TAGS {
	CHUNK_TAG = 1001, // used when chunk of dist. matrix computations is sent
	DM_PART_TAG = 2001, // used when results of dist. matrix computations are sent
	LIB_JOB_TAG = 3001, // used when lib jobs description is sent
	LIB_TAG = 4001 // used when precompiled lib is sent
};


class MPI_Env: public empi::MPI_env {
	public:
		explicit MPI_Env(MPI_Comm comm = MPI_COMM_WORLD): empi::MPI_env(comm)
		{
			proc_ = 1;
			m_rank_map.resize(size_);
			for (unsigned int i = 0; i < m_rank_map.size(); ++i)
				m_rank_map[i] = i;
		}

		bool set_map(const std::vector<int> &map)
		{
			if (size_ != map.size())
				return false;
			m_rank_map = map;
			return true;
		}

		int map(unsigned int i) const
		{
			return m_rank_map[i];
		}

/*
		void set_error_handler(MPI_Comm_errhandler_fn fn) const
		{
			MPI_Errhandler h;

			MPI_Comm_create_errhandler(fn, &h);
			MPI_Comm_set_errhandler(comm_, h);
		}
*/
		
		void set_proc(int p) { proc_ = p; }
		
		int get_proc() const { return proc_; }

                void set_mem(int m) { low_mem_ = m; }
		
		bool get_mem() const { return low_mem_; }


	private:
		int proc_;
                int low_mem_;
		std::vector<int> m_rank_map;    
};



TC_Sequences *AR_bcast_seq(TC_Sequences *S, const MPI_Env &mpi_env);



#endif // UTIL_MPI_HPP
