/*
 * File: AR_Config.cpp
 * Created: 12.09.2006
 * Last Modified: 14.09.2006
 * Author: Adrian Rospondek
 */


#include "AR_Config.hpp"
#include "util_str.hpp"



void AR_Config::read(char *env[])
{
	std::string tmp;
	std::string::size_type pos;

	bool fd, fc, ft, fm, fmas, fdef, fext, faw, fpdb, fsap;
	fd = fc = ft = fm = fmas = fdef = fext = faw = fpdb = fsap = false;

	for (char **var = env; *var; ++var) {
		tmp = *var;

		if (!fd && (pos = tmp.find("TCOFFEE_DIR")) != std::string::npos) {
			pos += 12;
			if (pos != tmp.size()) {
				m_dir = tmp.substr(pos);

				if (*m_dir.rbegin() != '/')
					m_dir += '/';
			}
			fd = true;
		}
		else if (!fc && (pos = tmp.find("TCOFFEE_CACHE")) != std::string::npos) {
			pos += 14;
			if (pos != tmp.size()) {
				m_cache = tmp.substr(pos);

				if (*m_cache.rbegin() != '/')
					m_cache += '/';
			}
			fc = true;
		}
		else if (!ft && (pos = tmp.find("TCOFFEE_TMP")) != std::string::npos) {
			pos += 12;
			if (pos != tmp.size()) {
				m_tmp = tmp.substr(pos);

				if (*m_tmp.rbegin() != '/')
					m_tmp += '/';
			}
			ft = true;
		}
		else if (!fm && (pos = tmp.find("TCOFFEE_METHODS")) != std::string::npos) {
			pos += 16;
			if (pos != tmp.size()) {
				m_methods = tmp.substr(pos);

				if (*m_methods.rbegin() != '/')
					m_methods += '/';
			}
			fm = true;
		}
		else if (!fmas && (pos = tmp.find("TCOFFEE_MASTER")) != std::string::npos) {
			pos += 15;
			if (pos != tmp.size())
				m_master = tmp.substr(pos);
			fmas = true;
		}
		else if (!fdef && (pos = tmp.find("TCOFFEE_DEFAULTS")) != std::string::npos) {
			pos += 17;
			if (pos != tmp.size())
				m_defaults = tmp.substr(pos);
			fdef = true;
		}
		else if (!fext && (pos = tmp.find("TCOFFEE_EXTRA_PARAMS")) != std::string::npos) {
			pos += 21;
			if (pos != tmp.size())
				m_extra_params = tmp.substr(pos);
			fext = true;
		}
		else if (!faw && (pos = tmp.find("TCOFFEE_ALN_WIDTH")) != std::string::npos) {
			pos += 18;
			if (pos != tmp.size()) {
				const char *val = tmp.substr(pos).c_str();

				if (TC_is_number(val))
					m_aln_width = atoi(val);
			}
			faw = true;
		}
		else if (!fpdb && (pos = tmp.find("TCOFFEE_EXTRACT_PDB")) != std::string::npos) {
			pos += 20;
			if (pos != tmp.size())
				m_extract_pdb = tmp.substr(pos);
			fpdb = true;
		}
		else if (!fsap && (pos = tmp.find("TCOFFEE_SAP")) != std::string::npos) {
			pos += 12;
			if (pos != tmp.size())
				m_sap = tmp.substr(pos);
			fsap = true;
		}

		if (fd && fc && ft && fm && fmas && fdef && fext && faw && fpdb && fsap)
			break;
	} // for (var)

	if (fd) {
		if (!fc) {
			m_cache = m_dir + "cache/";
		}
		if (!ft) {
			m_tmp = m_dir + "tmp/";
		}
		if (!fm) {
			m_methods = m_dir + "methods/";
		}
	}
} // AR_Config::read


AR_Config g_conf;
