/*
 * File: TC_Method.cpp
 * Created: 06.06.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_Method.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "globals.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_str.hpp"



TC_Method *TC_method_file2tc_method(const char *method)
{
	TC_Method *m;
	FILE *fp;
	char subcommand[LONG_STRING + 1];
	char line[LONG_STRING + 1];
	char *p;
	int c;

	m = (TC_Method *)AR_allocate(1, sizeof(TC_Method));

	// set default parameter values
	m->gop = m->gep = TC_UNDEFINED;
	strncpy(m->seq_type, "S", 100 - 1);
	strncpy(m->weight, "sim", 100 - 1);
	m->minid = 0;
	m->maxid = 100;

	fp = TC_vfopen(method, "r");
	while ((c = fgetc(fp)) != EOF) {
		ungetc(c, fp);
		fgets(line, LONG_STRING, fp);

		if (line && (line[0] == '*' || line[0] == '#' || line[0] == '$' || line[0] == ' ' || line[0] == '\0'))
			subcommand[0] = '\0';
		else if (p = strstr(line, "EXECUTABLE2"))
			sscanf(p, "EXECUTABLE2 %s", m->executable2);
		else if (p = strstr(line, "EXECUTABLE"))
			sscanf(p, "EXECUTABLE %s", m->executable);
		else if (p = strstr(line, "IN_FLAG2"))
			sscanf(p, "IN_FLAG2 %s", m->in_flag2);
		else if (p = strstr(line, "IN_FLAG"))
			sscanf(p, "IN_FLAG %s", m->in_flag);
		else if (p = strstr(line, "OUT_FLAG"))
			sscanf(p, "OUT_FLAG %s", m->out_flag);
		else if (p = strstr(line, "OUT_MODE"))
			sscanf(p, "OUT_MODE %s", m->out_mode);
		else if (p = strstr(line, "ALN_MODE"))
			sscanf(p, "ALN_MODE %s", m->aln_mode);
		else if (p = strstr(line, "SEQ_TYPE"))
			sscanf(p, "SEQ_TYPE %s", m->seq_type);
		else if (p = strstr(line, "WEIGHT"))
			sscanf(p, "WEIGHT %s", m->weight);
		else if (p = strstr(line, "MATRIX"))
			sscanf(p, "MATRIX %s", m->matrix);
		else if (p = strstr(line, "GOP"))
			sscanf(p, "GOP %d", &m->gop);
		else if (p = strstr(line, "GEP"))
			sscanf(p, "GEP %d", &m->gep);
		else if (p = strstr(line, "MAXID"))
			sscanf(p, "MAXID %d", &m->maxid);
		else if (p = strstr(line, "MINID"))
			sscanf(p, "MINID %d", &m->minid);
		else if (p = strstr(line, "PARAM1")) {
			sprintf(subcommand, " %s ", p + 6);
			strcat(m->param1, subcommand);
		}
		else if (p = strstr(line, "PARAM")) {
			sprintf(subcommand, " %s ", p + 5);
			strcat(m->param, subcommand);
		}
	}
	TC_vfclose(fp);

	return m;
}

char *TC_method_name2method_file(const char *method)
{
	FILE *fp;
	char *fname;

	if (AR_check_file_exists(method) || (toupper(method[0]) == method[0] && AR_check_file_exists(method + 1)))
		return 0;

	fname = AR_tmpname();

	// IMPORTANT: methods must start with a lowercase
	// internal methods
	if (strstr(method, "slow_pair")) {
		fp = TC_vfopen(fname, "w");
		fprintf(fp, "EXECUTABLE slow_pair\n");
		fprintf(fp, "ALN_MODE   pairwise\n");
		fprintf(fp, "OUT_MODE   fL\n");
		fprintf(fp, "IN_FLAG    no_name\n");
		fprintf(fp, "OUT_FLAG   no_name\n");
		fprintf(fp, "SEQ_TYPE   S\n");
		TC_vfclose(fp);
	}
	else if (strstr(method, "lalign_id_pair")) {
		fp = TC_vfopen(fname, "w");
		fprintf(fp, "EXECUTABLE lalign_id_pair\n");
		fprintf(fp, "ALN_MODE   pairwise\n");
		fprintf(fp, "OUT_MODE   fL\n");
		fprintf(fp, "IN_FLAG    no_name\n");
		fprintf(fp, "OUT_FLAG   no_name\n");
		fprintf(fp, "MATRIX     blosum50mt\n");
		fprintf(fp, "GOP   -10\n");
		fprintf(fp, "GEP   -4\n");
		fprintf(fp, "MAXID  50\n");
		fprintf(fp, "SEQ_TYPE   S\n");
		TC_vfclose(fp);
	}
	// structure super-imposition
	else if (strstr(method, "fugue_pair")) {
		fp = TC_vfopen(fname, "w");
		fprintf(fp, "EXECUTABLE fugue_pair\n");
		fprintf(fp, "ALN_MODE   pairwise\n");
		fprintf(fp, "OUT_MODE   fL\n");
		fprintf(fp, "IN_FLAG    no_name\n");
		fprintf(fp, "OUT_FLAG   no_name\n");
		fprintf(fp, "SEQ_TYPE   Ps\n");
		TC_vfclose(fp);
	}
	else if (strstr(method, "sap_pair")) {
		fp = TC_vfopen(fname, "w");
		fprintf(fp, "EXECUTABLE sap_pair\n");
		fprintf(fp, "ALN_MODE   pairwise\n");
		fprintf(fp, "OUT_MODE   fL\n");
		fprintf(fp, "IN_FLAG    no_name\n");
		fprintf(fp, "OUT_FLAG   no_name\n");
		fprintf(fp, "SEQ_TYPE   P\n");
		TC_vfclose(fp);
	}
	else {
		return 0;
	}

	// add extra parameters
	if (method = strstr(method, "EP@")) {
		char **list;
		int n, a;

		list = TC_string2list2(method, "@");
		n = atoi(list[0]);

		fp = TC_vfopen(fname, "a");
		for (a = 2; a < n; a += 2) {
			fprintf(fp, "%s %s\n", list[a], list[a + 1]);
		}
		TC_vfclose(fp);
		TC_free_char(list);
	}

	return fname;
}

bool AR_is_preset_method(const char *method)
{
	const char *new_name = 0;
	bool is_preset = false;

	new_name = TC_method_name2method_file(method);
	if (new_name) {
		TC_vremove(new_name);
		is_preset = true;
	}

	return is_preset;
}

int TC_method_uses_structure(const TC_Method *M)
{
	if (strchr(M->seq_type, 'P'))
		return 1;
	else
		return 0;
}

int TC_method_uses_profile(const TC_Method *M)
{
	if (strchr(M->seq_type, 'R'))
		return 1;
	else
		return 0;
}

/**
 * @return 1 if the sequence S is associated with a profile containing the right sequence
 */
static int TC_check_profile_seq_type(const TC_Sequences *S, int i, char t)
{
	TC_Alignment *A;
	TC_Templates *T;
	int a;

	A = TC_seq2R_template_profile(S, i);
	if (A == 0 || A->S == 0)
		return 0;

	for (a = 0; a < A->nseq; a++) {
		T = (A->S)->T[a];
		if (T && strchr(T->seq_type, t))
			return 1;
	}

	return 0;
}

int TC_check_seq_type(const TC_Method *M, const char *list, const TC_Sequences *S)
{
	TC_Templates *T1, *T2;
	char t1, t2;
	int *slist;
	int s1, s2, n1, nseq, ntype, i;

	slist = TC_string2num_list(list);
	nseq = slist[1];
	ntype = strlen(M->seq_type);
	t1 = M->seq_type[0];
	t2 = M->seq_type[1];
	n1 = 0;

	/* Profiles and Sequences MUST NOT be distinguished
	 * so that sequences and profiles can easily be aligned */

	if (tolower(t1) == 'r')
		t1 = 'S';
	if (tolower(t2) == 'r')
		t2 = 'S';

	if (TC_strm(M->aln_mode, "pairwise") && nseq > 2)
		n1 = 0;
	else if (TC_strm(M->aln_mode, "multiple") && ntype > 1)
		n1 = 0;
	else if (ntype == 1) {
		for (n1 = 0, i = 0; i < nseq; i++) {
			s1 = slist[i + 2];
			T1 = S->T[s1];
			n1 += (strchr(T1->seq_type, t1) || TC_check_profile_seq_type(S, s1, t1)) ? 1 : 0;
		}
		n1 = (n1 == nseq) ? 1 : 0;
	}
	else if (ntype == 2) {
		int s1_has_t1;
		int s1_has_t2;
		int s2_has_t1;
		int s2_has_t2;

		s1 = slist[2];
		s2 = slist[3];
		T1 = (S->T[s1]);
		T2 = (S->T[s2]);

		s1_has_t1 = (strchr(T1->seq_type, t1) || TC_check_profile_seq_type(S, s1, t1)) ? 1 : 0;
		s1_has_t2 = (strchr(T1->seq_type, t2) || TC_check_profile_seq_type(S, s1, t2)) ? 1 : 0;
		s2_has_t1 = (strchr(T2->seq_type, t1) || TC_check_profile_seq_type(S, s2, t1)) ? 1 : 0;
		s2_has_t2 = (strchr(T2->seq_type, t2) || TC_check_profile_seq_type(S, s2, t2)) ? 1 : 0;
		n1 = ((s1_has_t1 && s2_has_t2) || (s1_has_t2 && s2_has_t1)) ? 1 : 0;
	}

	AR_free(slist);

	return n1;
}
