/*
 * File: TC_Profile.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "globals.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"



TC_Profile *TC_declare_profile(const char *alphabet, int len)
{
	TC_Profile *P;
	
	P = (TC_Profile *)AR_allocate(1, sizeof(TC_Profile));
	P->max_len = len;
	P->alp_size = strlen(alphabet);
	P->alphabet = (char *)AR_allocate(P->alp_size + 2, sizeof(char));
	strncpy(P->alphabet, alphabet, P->alp_size + 1);
	P->count = TC_declare_int(P->alp_size + 2, len);
	P->count2 = TC_declare_int(100, len);
	P->count3 = TC_declare_int(100, len);
	
	return P;
}

TC_Profile *TC_copy_profile(const TC_Profile *P1)
{
	TC_Profile *P;
	
	if (!P1)
		return 0;
	
	P = TC_declare_profile(P1->alphabet, P1->max_len);
	P->count = TC_copy_int(P1->count, P->count, -1, -1);
	P->count2 = TC_copy_int(P1->count2, P->count2, -1, -1);
	P->count3 = TC_copy_int(P1->count3, P->count3, -1, -1);
	
	return P;
}

TC_Profile *TC_free_profile(TC_Profile *P)
{
	if (!P)
		return 0;
	else {
		AR_free(P->alphabet);
		TC_free_int(P->count);
		TC_free_int(P->count2);
		TC_free_int(P->count3);
		AR_free(P);
	}
	
	return 0;
}

TC_Alignment *TC_seq2profile(TC_Sequences *S, int i)
{
	TC_Alignment *A;
	
	A = TC_seq2R_template_profile(S, i);
	
	if (!A) {
		FILE *fp;
		char *tmp;
		
		tmp = AR_tmpname();
		fp = TC_vfopen(tmp, "w");
		fprintf(fp, ">%s\n%s\n", S->name[i], S->seq[i]);
		TC_vfclose(fp);
		
		(S->T[i])->R = TC_fill_R_template(S->name[i], tmp, S);
		
		A = TC_seq2R_template_profile(S, i);
	}

	return A;
}

TC_Alignment *TC_aln2profile(TC_Alignment *A)
{
	TC_Alignment *B = 0;
	char *cons;
	
	if (!A->P)
		A->P = TC_declare_profile(AA_ALPHABET, A->len_aln + 1);
	B = TC_copy_aln(A, B);
	
	TC_free_int((A->P)->count);
	TC_free_int((A->P)->count2);
	TC_free_int((A->P)->count3);
	
	(A->P)->count = TC_aln2count_mat(A);
	(A->P)->count2 = TC_aln2count_mat2(A);

	cons = TC_aln2cons_seq_mat(A, "blosum62mt");
	strncpy(B->seq_al[0], cons, B->declared_len - 1);
	B->nseq = 1;
	(A->P)->count3 = TC_aln2count_mat2(B);
	
	AR_free(cons);
	TC_free_aln(B);
	
	return A;
}
