/*
 * File: TC_tree.cpp
 * Created: 06.06.2006
 * Last Modified: 19.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "TC_tree.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_str.hpp"



static int g_nseqs;
static char **g_names;
static int g_first_seq, g_last_seq;
static int *g_tkill;
static double **g_tmat;
static double *g_av;
static double *g_left_branch, *g_right_branch;


TC_NT_node TC_declare_tree_node(int nseq)
{
	TC_NT_node p;

	p = (TC_NT_node)AR_allocate(1, sizeof(TC_tree_node));
	p->left = 0;
	p->right = 0;
	p->parent = 0;
	p->dist = 0.0;
	p->leaf = 0;
	p->order = 0;
	p->maxnseq = nseq;
	p->name = (char *)AR_allocate(MAXNAMES + 1, sizeof(char));
	p->name[0] = '\0';
	p->lseq = (int *)AR_allocate(nseq, sizeof(int));

	return p;
}

void TC_create_tree_node(TC_NT_node pptr, TC_NT_node parent)
{
	pptr->parent = parent;
	pptr->left = TC_declare_tree_node(pptr->maxnseq);
	(pptr->left)->parent = pptr;
	pptr->right = TC_declare_tree_node(pptr->maxnseq);
	(pptr->right)->parent = pptr;
}

TC_NT_node TC_insert_tree_node(TC_NT_node pptr)
{
	TC_NT_node newnode;

	newnode = TC_declare_tree_node(pptr->maxnseq);

	TC_create_tree_node(newnode, pptr->parent);
	newnode->left = pptr;
	pptr->parent = newnode;
	TC_set_info(newnode, pptr->parent, 0, "", 0.0, 0);

	return newnode;
}

void TC_set_info(TC_NT_node p, TC_NT_node parent, int pleaf, const char *pname, float pdist, float bootstrap)
{
	p->parent = parent;
	p->leaf = pleaf;
	p->dist = pdist;
	p->bootstrap = bootstrap;
	p->order = 0;
	strncpy(p->name, pname, MAXNAMES);
	if (pleaf == 1) {
		p->left = 0;
		p->right = 0;
	}
}

static void TC_slow_nj_tree(char **tree_description)
{
	int l[4], nude, k;
	int nc, mini, minj, i, j, ii, jj;
	double fnseqs, fnseqs2, sumd;
	double diq, djq, dij, d2r, dr, dio, djo, da;
	double tmin, total, dmin;
	double bi, bj, b1, b2, b3, branch[4];
	int typei, typej;		/* 0 = node; 1 = OTU */

	fnseqs = (double)g_nseqs;

/* First Initialisation */
	mini = minj = 0;
	g_left_branch = (double *)AR_allocate(g_nseqs + 2, sizeof(double));
	g_right_branch = (double *)AR_allocate(g_nseqs + 2, sizeof(double));
	g_tkill = (int *)AR_allocate(g_nseqs + 1, sizeof(int));
	g_av = (double *)AR_allocate(g_nseqs + 1, sizeof(double));

	for (i = 1; i <= g_nseqs; ++i) {
		g_tmat[i][i] = g_av[i] = 0.0;
		g_tkill[i] = 0;
	}

/* Enter The Main Cycle */
	/* start main cycle */
	for (nc = 1; nc <= (g_nseqs - 3); ++nc) {
		sumd = 0.0;
		for (j = 2; j <= g_nseqs; ++j)
			for (i = 1; i < j; ++i) {
				g_tmat[j][i] = g_tmat[i][j];
				sumd = sumd + g_tmat[i][j];
			}
		tmin = 99999.0;
		/* compute SMATij values and find the smallest one */
		for (jj = 2; jj <= g_nseqs; ++jj)
			if (g_tkill[jj] != 1)
				for (ii = 1; ii < jj; ++ii)
					if (g_tkill[ii] != 1) {
						diq = djq = 0.0;
						for (i = 1; i <= g_nseqs; ++i) {
							diq = diq + g_tmat[i][ii];
							djq = djq + g_tmat[i][jj];
						}
						dij = g_tmat[ii][jj];
						d2r = diq + djq - (2.0 * dij);
						dr = sumd - dij - d2r;
						fnseqs2 = fnseqs - 2.0;
						total = d2r + fnseqs2 * dij + dr * 2.0;
						total = total / (2.0 * fnseqs2);
						if (total < tmin) {
							if (tmin - total > MY_EPS)
								tmin = total;
							mini = ii;
							minj = jj;
						}
					}
		/* compute branch lengths and print the results */
		dio = djo = 0.0;
		for (i = 1; i <= g_nseqs; ++i) {
			dio = dio + g_tmat[i][mini];
			djo = djo + g_tmat[i][minj];
		}
		dmin = g_tmat[mini][minj];
		dio = (dio - dmin) / fnseqs2;
		djo = (djo - dmin) / fnseqs2;
		bi = (dmin + dio - djo) * 0.5;
		bj = dmin - bi;
		bi = bi - g_av[mini];
		bj = bj - g_av[minj];
		if (g_av[mini] > 0.0)
			typei = 0;
		else
			typei = 1;
		if (g_av[minj] > 0.0)
			typej = 0;
		else
			typej = 1;
		/* Set negative branch lengths to zero;
		 * also set any tiny positive branch lengths to zero. */
		if (fabs(bi) < 0.0001)
			bi = 0.0;
		if (fabs(bj) < 0.0001)
			bj = 0.0;
		g_left_branch[nc] = bi;
		g_right_branch[nc] = bj;
		for (i = 1; i <= g_nseqs; i++)
			tree_description[nc][i] = 0;
		if (typei == 0) {
			for (i = nc - 1; i >= 1; i--)
				if (tree_description[i][mini] == 1) {
					for (j = 1; j <= g_nseqs; j++)
						if (tree_description[i][j] == 1)
							tree_description[nc][j] = 1;
					break;
				}
		}
		else
			tree_description[nc][mini] = 1;
		if (typej == 0) {
			for (i = nc - 1; i >= 1; i--)
				if (tree_description[i][minj] == 1) {
					for (j = 1; j <= g_nseqs; j++)
						if (tree_description[i][j] == 1)
							tree_description[nc][j] = 1;
					break;
				}
		}
		else
			tree_description[nc][minj] = 1;
		/* Here is where the -0.00005 branch lengths come from for 3 or more identical seqs. */
		if (dmin <= 0.0)
			dmin = 0.000001;
		g_av[mini] = dmin * 0.5;

		/* re-initialisation */
		fnseqs = fnseqs - 1.0;
		g_tkill[minj] = 1;
		for (j = 1; j <= g_nseqs; ++j)
			if (g_tkill[j] != 1) {
				da = (g_tmat[mini][j] + g_tmat[minj][j]) * 0.5;
				if ((mini - j) < 0)
					g_tmat[mini][j] = da;
				if ((mini - j) > 0)
					g_tmat[j][mini] = da;
			}
		for (j = 1; j <= g_nseqs; ++j)
			g_tmat[minj][j] = g_tmat[j][minj] = 0.0;
	}
	/* end main cycle */

/* Last Cycle (3 seqs. left) */
	nude = 1;
	for (i = 1; i <= g_nseqs; ++i)
		if (g_tkill[i] != 1) {
			l[nude] = i;
			nude = nude + 1;
		}
	b1 = (g_tmat[l[1]][l[2]] + g_tmat[l[1]][l[3]] - g_tmat[l[2]][l[3]]) * 0.5;
	b2 = g_tmat[l[1]][l[2]] - b1;
	b3 = g_tmat[l[1]][l[3]] - b1;
	branch[1] = b1 - g_av[l[1]];
	branch[2] = b2 - g_av[l[2]];
	branch[3] = b3 - g_av[l[3]];

	/* Reset tiny negative and positive branch lengths to zero */
	if (fabs(branch[1]) < 0.0001)
		branch[1] = 0.0;
	if (fabs(branch[2]) < 0.0001)
		branch[2] = 0.0;
	if (fabs(branch[3]) < 0.0001)
		branch[3] = 0.0;

	g_left_branch[g_nseqs - 2] = branch[1];
	g_left_branch[g_nseqs - 1] = branch[2];
	g_left_branch[g_nseqs] = branch[3];
	for (i = 1; i <= g_nseqs; i++)
		tree_description[g_nseqs - 2][i] = 0;
	for (i = 1; i <= 3; ++i) {
		if (g_av[l[i]] > 0.0) {
			for (k = g_nseqs - 3; k >= 1; k--)
				if (tree_description[k][l[i]] == 1) {
					for (j = 1; j <= g_nseqs; j++)
						if (tree_description[k][j] == 1)
							tree_description[g_nseqs - 2][j] = i;
					break;
				}
		}
		else
			tree_description[g_nseqs - 2][l[i]] = i;
	}
}

static void TC_fast_nj_tree(char **tree_description)
{
	int l[4], nude, k;
	int nc, mini, minj, i, j, ii, jj;
	double fnseqs, fnseqs2 = 0, sumd;
	double diq, djq, dij, dio, djo, da;
	double tmin, total, dmin;
	double bi, bj, b1, b2, b3, branch[4];
	int typei, typej;		/* 0 = node; 1 = OTU */

/* IMPROVEMENT 1, STEP 0: declare  variables */
	double *sum_cols, *sum_rows, *join;

/* IMPROVEMENT 2, STEP 0: declare  variables */
	int loop_limit;
	struct ValidNodeID {
		int n;
		ValidNodeID *prev;
		ValidNodeID *next;
	};
	ValidNodeID *tvalid, *lpi, *lpj, *lpii, *lpjj, *lp_prev, *lp_next;

	/*
	 * correspondence of the loop counter variables.
	 *   i .. lpi->n,  ii .. lpii->n
	 *   j .. lpj->n,  jj .. lpjj->n
	 */

	fnseqs = (double)(g_last_seq - g_first_seq + 1);

/* First Initialisation */
	if (fnseqs == 2)
		return;

	mini = minj = 0;
	g_left_branch = (double *)AR_allocate((g_nseqs + 2), sizeof(double));
	g_right_branch = (double *)AR_allocate((g_nseqs + 2), sizeof(double));
	g_tkill = (int *)AR_allocate((g_nseqs + 1), sizeof(int));
	g_av = (double *)AR_allocate((g_nseqs + 1), sizeof(double));

/* IMPROVEMENT 1, STEP 1: Allocate memory */
	sum_cols = (double *)AR_allocate((g_nseqs + 1), sizeof(double));
	sum_rows = (double *)AR_allocate((g_nseqs + 1), sizeof(double));
	join = (double *)AR_allocate((g_nseqs + 1), sizeof(double));
/* IMPROVEMENT 2, STEP 1: Allocate memory */
	tvalid = (ValidNodeID *) AR_allocate((g_nseqs + 1), sizeof(ValidNodeID));
	/* tvalid[0] is special entry in array, it points a header of valid entry list */
	tvalid[0].n = 0;
	tvalid[0].prev = 0;
	tvalid[0].next = &tvalid[1];

/* IMPROVEMENT 2, STEP 2: Construct and initialize the entry chain list */
	for (i = 1, loop_limit = g_last_seq - g_first_seq + 1, lpi = &tvalid[1], lp_prev = &tvalid[0],
		lp_next = &tvalid[2]; i <= loop_limit; ++i, ++lpi, ++lp_prev, ++lp_next) {
		g_tmat[i][i] = g_av[i] = 0.0;
		g_tkill[i] = 0;
		lpi->n = i;
		lpi->prev = lp_prev;
		lpi->next = lp_next;
/* IMPROVEMENT 1, STEP 2: Initialize arrays */
		sum_cols[i] = sum_rows[i] = join[i] = 0.0;
	}
	tvalid[loop_limit].next = 0;

/* IMPROVEMENT 1, STEP 3: Calculate the sum of score value that is sequence[i] to others. */
	sumd = 0.0;
	for (lpj = tvalid[0].next; lpj != 0; lpj = lpj->next) {
		double tmp_sum = 0.0;

		j = lpj->n;
		/* calculate sum_rows[j] */
		for (lpi = tvalid[0].next; lpi->n < j; lpi = lpi->next) {
			i = lpi->n;
			tmp_sum += g_tmat[i][j];
		}
		sum_rows[j] = tmp_sum;
		tmp_sum = 0.0;
		/* set lpi to that lpi->n is greater than j */
		if (lpi != 0 && lpi->n == j)
			lpi = lpi->next;
		/* calculate sum_cols[j] */
		for (; lpi != 0; lpi = lpi->next) {
			i = lpi->n;
			tmp_sum += g_tmat[j][i];
		}
		sum_cols[j] = tmp_sum;
	}

/* Enter The Main Cycle */
	for (nc = 1, loop_limit = (g_last_seq - g_first_seq + 1 - 3); nc <= loop_limit; ++nc) {
		sumd = 0.0;
/* IMPROVEMENT 1, STEP 4: use sum value */
		for (lpj = tvalid[0].next; lpj != 0; lpj = lpj->next)
			sumd += sum_cols[lpj->n];
/* IMPROVEMENT 3, STEP 0: multiply tmin and 2*fnseqs2 */
		fnseqs2 = fnseqs - 2.0;	/* Set fnseqs2 at this point. */
		tmin = 99999.0 * 2.0 * fnseqs2;

		/* compute SMATij values and find the smallest one */
		mini = minj = 0;

		/* jj must starts at least 2 */
		if (tvalid[0].next != 0 && tvalid[0].next->n == 1)
			lpjj = tvalid[0].next->next;
		else
			lpjj = tvalid[0].next;
		for (; lpjj != 0; lpjj = lpjj->next) {
			jj = lpjj->n;
			for (lpii = tvalid[0].next; lpii->n < jj; lpii = lpii->next) {
				ii = lpii->n;
				diq = djq = 0.0;
/* IMPROVEMENT 1, STEP 4: use sum value */
				diq = sum_cols[ii] + sum_rows[ii];
				djq = sum_cols[jj] + sum_rows[jj];
				/* always ii < jj in this point.
				 * Use upper triangle of score matrix. */
				dij = g_tmat[ii][jj];
/* IMPROVEMENT 3, STEP 1: fnseqs2 is already calculated. */
				//fnseqs2 = fnseqs - 2.0

/* IMPROVEMENT 4: transform the equation */
/* --------------------------------------------------------------- *
 * OPTIMIZE of expression 'total = d2r + fnseqs2 * dij + dr * 2.0' *
 *   total = d2r + fnseq2*dij + 2.0*dr                             *
 *         = d2r + fnseq2*dij + 2(sumd - dij - d2r)                *
 *         = d2r + fnseq2*dij + 2*sumd - 2*dij - 2*d2r             *
 *         =       fnseq2*dij + 2*sumd - 2*dij - 2*d2r + d2r       *
 *         = fnseq2*dij + 2*sumd - 2*dij - d2r                     *
 *         = fnseq2*dij + 2*sumd - 2*dij - (diq + djq - 2*dij)     *
 *         = fnseq2*dij + 2*sumd - 2*dij - diq - djq + 2*dij       *
 *         = fnseq2*dij + 2*sumd - 2*dij + 2*dij - diq - djq       *
 *         = fnseq2*dij + 2*sumd  - diq - djq                      *
 * --------------------------------------------------------------- */
				total = fnseqs2 * dij + 2.0 * sumd - diq - djq;

/* IMPROVEMENT 3, STEP 2: abbrevlate the division on comparison between total and tmin. */
				//total = total / (2.0 * fnseqs2);
				if (total < tmin) {
					tmin = total;
					mini = ii;
					minj = jj;
				}
			}
		}
		/* MEMO: always ii < jj in avobe loop, so mini < minj */

		/* compute branch lengths and print the results */
		dio = djo = 0.0;

/* IMPROVEMENT 1, STEP 4: use sum value */
		dio = sum_cols[mini] + sum_rows[mini];
		djo = sum_cols[minj] + sum_rows[minj];
		dmin = g_tmat[mini][minj];
		dio = (dio - dmin) / fnseqs2;
		djo = (djo - dmin) / fnseqs2;
		bi = (dmin + dio - djo) * 0.5;
		bj = dmin - bi;
		bi = bi - g_av[mini];
		bj = bj - g_av[minj];
		if (g_av[mini] > 0.0)
			typei = 0;
		else
			typei = 1;
		if (g_av[minj] > 0.0)
			typej = 0;
		else
			typej = 1;
		/* Set negative branch lengths to zero;
		 * also set any tiny positive branch lengths to zero. */
		if (fabs(bi) < 0.0001)
			bi = 0.0;
		if (fabs(bj) < 0.0001)
			bj = 0.0;
		g_left_branch[nc] = bi;
		g_right_branch[nc] = bj;
		for (i = 1; i <= g_last_seq - g_first_seq + 1; i++)
			tree_description[nc][i] = 0;
		if (typei == 0) {
			for (i = nc - 1; i >= 1; i--)
				if (tree_description[i][mini] == 1) {
					for (j = 1; j <= g_last_seq - g_first_seq + 1; j++)
						if (tree_description[i][j] == 1)
							tree_description[nc][j] = 1;
					break;
				}
		}
		else
			tree_description[nc][mini] = 1;
		if (typej == 0) {
			for (i = nc - 1; i >= 1; i--)
				if (tree_description[i][minj] == 1) {
					for (j = 1; j <= g_last_seq - g_first_seq + 1; j++)
						if (tree_description[i][j] == 1)
							tree_description[nc][j] = 1;
					break;
				}
		}
		else
			tree_description[nc][minj] = 1;
		/* Here is where the -0.00005 branch lengths come from for 3 or more identical seqs. */
		if (dmin <= 0.0)
			dmin = 0.000001;
		g_av[mini] = dmin * 0.5;

		/* re-initialisation */
		fnseqs = fnseqs - 1.0;
		g_tkill[minj] = 1;

/* IMPROVEMENT 2, STEP 3: Remove tvalid[minj] from chain list. */
		(tvalid[minj].prev)->next = tvalid[minj].next;
		if (tvalid[minj].next != 0)
			(tvalid[minj].next)->prev = tvalid[minj].prev;
		tvalid[minj].prev = tvalid[minj].next = 0;
/* IMPROVEMENT 1, STEP 5: re-calculate sum values. */
		for (lpj = tvalid[0].next; lpj != 0; lpj = lpj->next) {
			double tmp_di = 0.0;
			double tmp_dj = 0.0;

			j = lpj->n;
			/* subtrace a score value related with 'minj' from sum arrays. */
			if (j < minj) {
				tmp_dj = g_tmat[j][minj];
				sum_cols[j] -= tmp_dj;
			}
			else if (j > minj) {
				tmp_dj = g_tmat[minj][j];
				sum_rows[j] -= tmp_dj;
			}
			/* nothing to do when j is equal to minj. */
			/* subtrace a score value related with 'mini' from sum arrays. */
			if (j < mini) {
				tmp_di = g_tmat[j][mini];
				sum_cols[j] -= tmp_di;
			}
			else if (j > mini) {
				tmp_di = g_tmat[mini][j];
				sum_rows[j] -= tmp_di;
			}
			/* nothing to do when j is equal to mini. */
			/* calculate a score value of the new inner node;
			 * then, store it temporary to join[] array. */
			join[j] = (tmp_dj + tmp_di) * 0.5;
		}
		/* 
		 * 1) Set the score values (stored in join[]) into the matrix, row/column position is 'mini'.
		 * 2) Add a score value of the new inner node to sum arrays.
		 */
		for (lpj = tvalid[0].next; lpj != 0; lpj = lpj->next) {
			j = lpj->n;
			if (j < mini) {
				g_tmat[j][mini] = join[j];
				sum_cols[j] += join[j];
			}
			else if (j > mini) {
				g_tmat[mini][j] = join[j];
				sum_rows[j] += join[j];
			}
			/* nothing to do when j is equal to mini. */
		}
		/* re-calculate sum_rows[mini], sum_cols[mini]. */
		sum_cols[mini] = sum_rows[mini] = 0.0;

		/* calculate sum_rows[mini] */
		da = 0.0;
		for (lpj = tvalid[0].next; lpj->n < mini; lpj = lpj->next)
			da += join[lpj->n];
		sum_rows[mini] = da;
		/* skip if 'lpj->n' is equal to 'mini' */
		if (lpj != 0 && lpj->n == mini)
			lpj = lpj->next;

		/* calculate sum_cols[mini] */
		da = 0.0;
		for (; lpj != 0; lpj = lpj->next)
			da += join[lpj->n];
		sum_cols[mini] = da;

		/* Clean up sum_rows[minj], sum_cols[minj] and score matrix related with 'minj'. */
		sum_cols[minj] = sum_rows[minj] = 0.0;
		for (j = 1; j <= g_last_seq - g_first_seq + 1; ++j)
			g_tmat[minj][j] = g_tmat[j][minj] = join[j] = 0.0;
	}
	/* end main cycle */

/* Last Cycle (3 seqs. left) */
	nude = 1;
	for (lpi = tvalid[0].next; lpi != 0; lpi = lpi->next) {
		l[nude] = lpi->n;
		++nude;
	}
	b1 = (g_tmat[l[1]][l[2]] + g_tmat[l[1]][l[3]] - g_tmat[l[2]][l[3]]) * 0.5;
	b2 = g_tmat[l[1]][l[2]] - b1;
	b3 = g_tmat[l[1]][l[3]] - b1;
	branch[1] = b1 - g_av[l[1]];
	branch[2] = b2 - g_av[l[2]];
	branch[3] = b3 - g_av[l[3]];

	/* Reset tiny negative and positive branch lengths to zero */
	if (fabs(branch[1]) < 0.0001)
		branch[1] = 0.0;
	if (fabs(branch[2]) < 0.0001)
		branch[2] = 0.0;
	if (fabs(branch[3]) < 0.0001)
		branch[3] = 0.0;

	g_left_branch[g_last_seq - g_first_seq + 1 - 2] = branch[1];
	g_left_branch[g_last_seq - g_first_seq + 1 - 1] = branch[2];
	g_left_branch[g_last_seq - g_first_seq + 1] = branch[3];
	for (i = 1; i <= g_last_seq - g_first_seq + 1; i++)
		tree_description[g_last_seq - g_first_seq + 1 - 2][i] = 0;
	for (i = 1; i <= 3; ++i) {
		if (g_av[l[i]] > 0.0) {
			for (k = g_last_seq - g_first_seq + 1 - 3; k >= 1; k--)
				if (tree_description[k][l[i]] == 1) {
					for (j = 1; j <= g_last_seq - g_first_seq + 1; j++)
						if (tree_description[k][j] == 1)
							tree_description[g_last_seq - g_first_seq + 1 - 2][j] = i;
					break;
				}
		}
		else
			tree_description[g_last_seq - g_first_seq + 1 - 2][l[i]] = i;
	}
	AR_free(sum_cols);
	AR_free(sum_rows);
	AR_free(join);
	AR_free(tvalid);
}

static void TC_nj_tree(char **tree_description, int nseq)
{
	if (nseq < 100)
		TC_slow_nj_tree(tree_description);
	else
		TC_fast_nj_tree(tree_description);
}

static void TC_two_way_split(char **tree_description, FILE *tree, int start_row, int flag,
	int bootstrap)
{
	int row, new_row, col, test_col;
	int single_seq;

	if (start_row != g_nseqs - 2)
		fprintf(tree, "(\n");
	for (col = 1; col <= g_nseqs; col++) {
		if (tree_description[start_row][col] == flag) {
			test_col = col;
			break;
		}
	}
	single_seq = TRUE;
	for (row = start_row - 1; row >= 1; row--)
		if (tree_description[row][test_col] == 1) {
			single_seq = FALSE;
			new_row = row;
			break;
		}
	if (single_seq) {
		tree_description[start_row][test_col] = 0;
		fprintf(tree, "%s", g_names[test_col + 0 - 1]);
	}
	else {
		for (col = 1; col <= g_nseqs; col++) {
			if (tree_description[start_row][col] == 1 && tree_description[new_row][col] == 1)
				tree_description[start_row][col] = 0;
		}
		TC_two_way_split(tree_description, tree, new_row, 1, bootstrap);
	}
	if (start_row == g_nseqs - 2) {
		return;
	}

	fprintf(tree, ":%7.5f,\n", g_left_branch[start_row]);
	for (col = 1; col <= g_nseqs; col++)
		if (tree_description[start_row][col] == flag) {
			test_col = col;
			break;
		}
	single_seq = TRUE;
	for (row = start_row - 1; row >= 1; row--)
		if (tree_description[row][test_col] == 1) {
			single_seq = FALSE;
			new_row = row;
			break;
		}
	if (single_seq) {
		tree_description[start_row][test_col] = 0;
		fprintf(tree, "%s", g_names[test_col + 0 - 1]);
	}
	else {
		for (col = 1; col <= g_nseqs; col++) {
			if (tree_description[start_row][col] == 1 && tree_description[new_row][col] == 1)
				tree_description[start_row][col] = 0;
		}
		TC_two_way_split(tree_description, tree, new_row, 1, bootstrap);
	}
	fprintf(tree, ":%7.5f)\n", g_right_branch[start_row]);
}

static void TC_print_phylip_tree(char **tree_description, FILE *tree, int bootstrap)
{
	fprintf(tree, "(\n");
	TC_two_way_split(tree_description, tree, g_nseqs - 2, 1, bootstrap);
	fprintf(tree, ":%7.5f,\n", g_left_branch[g_nseqs - 2]);
	TC_two_way_split(tree_description, tree, g_nseqs - 2, 2, bootstrap);
	fprintf(tree, ":%7.5f,\n", g_left_branch[g_nseqs - 1]);
	TC_two_way_split(tree_description, tree, g_nseqs - 2, 3, bootstrap);
	fprintf(tree, ":%7.5f)", g_left_branch[g_nseqs]);
	if (bootstrap)
		fprintf(tree, "TRICHOTOMY");
	fprintf(tree, ";\n");
}

/**
 * @brief Routine for producing unrooted NJ trees from seperately aligned pairwise distances.
 * This produces the GUIDE DENDROGRAMS in PHYLIP format.
 */
static void TC_guide_tree(const char *fname, double **saga_tmat, char **saga_seq_name,
	int saga_nseq)
{
	FILE *fp;
	char **standard_tree;
	int i;

	g_nseqs = saga_nseq;
	g_first_seq = 1;
	g_last_seq = g_nseqs;
	g_names = saga_seq_name;
	g_tmat = saga_tmat;
	standard_tree = (char **)AR_allocate((g_nseqs + 1), sizeof(char *));
	for (i = 0; i < g_nseqs + 1; i++)
		standard_tree[i] = (char *)AR_allocate((g_nseqs + 1), sizeof(char));

	TC_nj_tree(standard_tree, g_nseqs);
	fp = fopen(fname, "w");
	TC_print_phylip_tree(standard_tree, fp, FALSE);
	fclose(fp);

	AR_free(g_left_branch);
	AR_free(g_right_branch);
	AR_free(g_tkill);
	AR_free(g_av);
	for (i = 0; i < g_nseqs + 1; i++)
		AR_free(standard_tree[i]);
	AR_free(standard_tree);
}

static TC_NT_node **TC_dist2nj_tree(double **distances, char **out_seq_name, int out_nseq,
	const char *tree_file)
{
	double **d_dist;
	int tot_node = 0;
	int a, b;

	if (!tree_file)
		tree_file = AR_tmpname();
	d_dist = TC_declare_double(out_nseq + 1, out_nseq + 1);

	for (a = 0; a < out_nseq; a++)
		for (b = 0; b < out_nseq; b++) {
			if (a != b)
				d_dist[a + 1][b + 1] = distances[a][b] / MAXID;
			else
				d_dist[a + 1][b + 1] = 0;
		}
	TC_guide_tree(tree_file, d_dist, out_seq_name, out_nseq);
	TC_free_double(d_dist);

	return TC_read_tree(tree_file, &tot_node, out_nseq, out_seq_name);
}

static TC_NT_node **TC_int_dist2nj_tree(int **distances, char **out_seq_name, int out_nseq,
	const char *tree_file)
{
	TC_NT_node **T;
	double **d;
	int a, b;

	d = TC_declare_double(out_nseq, out_nseq);
	for (a = 0; a < out_nseq; a++)
		for (b = 0; b < out_nseq; b++)
			d[a][b] = distances[a][b];

	T = TC_dist2nj_tree(d, out_seq_name, out_nseq, tree_file);
	TC_free_double(d);

	return T;
}

static inline TC_NT_node **TC_make_nj_tree(int **distances, char **out_seq_name, int out_nseq, const char *tree_file)
{
	return TC_int_dist2nj_tree(distances, out_seq_name, out_nseq, tree_file);
}

TC_NT_node **TC_make_tree(TC_ConstraintList *CL, TC_Sequences *S, const char *tree_file)
{
	TC_NT_node **T = 0;

	if (!CL || !CL->tree_mode[0])
		return T;

	std::cout << "\nMAKING GUIDE TREE (mode: " << CL->tree_mode << ").....";

	if (TC_strm(CL->tree_mode, "cwph")) {
		std::cerr << '\n' << PROGRAM << ": WARNING: Tree mode '" << CL->tree_mode
				  << "' is not supported in current version of " << PROGRAM << ".\n";
	}
	else if (TC_strm(CL->tree_mode, "nj") || TC_strm(CL->tree_mode, "upgma")) {
		if (!CL->DM) {
			std::cerr << '\n' << PROGRAM << ": ERROR: No distance matrix required in tree mode '"
					  << CL->tree_mode << "'.\n";
			return T;
		}

		int out_nseq = S->nseq;
		char **out_seq_name = S->name;
		int **distances;
		int a, b, ra, rb;

		if (CL->S != S) {
			// Shrink the distance matrix so that it only contains the required sequences
			distances = TC_declare_int(S->nseq, S->nseq);

			for (a = 0; a < S->nseq; ++a) {
				ra = TC_name_is_in_list(S->name[a], (CL->S)->name, (CL->S)->nseq, 100);
				for (b = 0; b < S->nseq; ++b) {
					rb = TC_name_is_in_list(S->name[b], (CL->S)->name, (CL->S)->nseq, 100);
					distances[a][b] = (CL->DM)->score_similarity_matrix[ra][rb];
				}
			}
		}
		else {
			distances = TC_duplicate_int((CL->DM)->score_similarity_matrix, -1, -1);
		}

		distances = TC_sim_array2dist_array(distances, 100);
		if (TC_strm(CL->tree_mode, "upgma")) {
			for (a = 0; a < S->nseq; ++a)
				for (b = 0; b < S->nseq; ++b)
					distances[b][a] = 100;
		}

		T = TC_make_nj_tree(distances, out_seq_name, out_nseq, tree_file);
		TC_free_int(distances);
	}
	std::cout << "[Done]\n";

	return T;
}

static float TC_calc_root_mean(TC_NT_node root, float *maxdist, int nseq, TC_NT_node **lu)
{
	TC_NT_node p;
	float dist, lsum = 0.0, rsum = 0.0, lmean, rmean, diff;
	int nl, nr;
	int direction;
	int i;

	dist = (*maxdist) = 0;
	nl = nr = 0;
	for (i = 0; i < nseq; i++) {
		p = lu[2][i];
		dist = 0.0;
		while (p->parent != root) {
			dist += p->dist;
			p = p->parent;
		}
		if (p == root->left)
			direction = LEFT;
		else
			direction = RIGHT;
		dist += p->dist;
		if (direction == LEFT) {
			lsum += dist;
			nl++;
		}
		else {
			rsum += dist;
			nr++;
		}
		if (dist > (*maxdist))
			*maxdist = dist;
	}
	lmean = lsum / nl;
	rmean = rsum / nr;
	diff = lmean - rmean;
	return diff;
}

static float TC_calc_mean(TC_NT_node nptr, float *maxdist, int nseq, TC_NT_node **lu)
{
	TC_NT_node p, *path2root;
	float *dist2node;
	float dist, lsum = 0.0, rsum = 0.0, lmean, rmean, diff;
	int nl, nr;
	int direction, found;
	int depth = 0, i, j, n;

	path2root = (TC_NT_node *)AR_allocate(nseq, sizeof(TC_NT_node));
	dist2node = (float *)AR_allocate(nseq, sizeof(float));

	(*maxdist) = dist = 0;
	depth = 0;
	nl = nr = 0;
	p = nptr;
	while (p != 0) {
		path2root[depth] = p;
		dist += p->dist;
		dist2node[depth] = dist;
		p = p->parent;
		depth++;
	}
	/***************************************************************************
	 * *nl = *nr = 0;
	 * for each leaf, determine whether the leaf is left or right of the node.
	 * (RIGHT = descendant, LEFT = not descendant)
	 **************************************************************************/
	for (i = 0; i < nseq; i++) {
		p = lu[2][i];
		if (p == nptr) {
			direction = RIGHT;
			dist = 0.0;
		}
		else {
			direction = LEFT;
			dist = 0.0;
			found = FALSE;
			n = 0;
			while (found == FALSE && p->parent != 0) {
				for (j = 0; j < depth; j++)
					if (p->parent == path2root[j]) {
						found = TRUE;
						n = j;
					}
				dist += p->dist;
				p = p->parent;
			}
			if (p == nptr)
				direction = RIGHT;
		}
		if (direction == LEFT) {
			lsum += dist;
			lsum += dist2node[n - 1];
			nl++;
		}
		else {
			rsum += dist;
			nr++;
		}
		if (dist > (*maxdist))
			*maxdist = dist;
	}
	AR_free(dist2node);
	AR_free(path2root);

	if (nl == 0 || nr == 0) {
		std::cerr << '\n' << PROGRAM << ": ERROR: (calc_mean).\n";
		exit(-1);
	}

	lmean = lsum / nl;
	rmean = rsum / nr;
	diff = lmean - rmean;

	return diff;
}

static TC_NT_node TC_insert_root(TC_NT_node p, float diff)
{
	TC_NT_node newp, prev, q, t;
	float dist, prevdist, td;

	newp = TC_declare_tree_node(p->maxnseq);

	t = p->parent;
	prevdist = t->dist;
	p->parent = newp;
	dist = p->dist;
	p->dist = diff / 2;
	if (p->dist < 0.0)
		p->dist = 0.0;
	if (p->dist > dist)
		p->dist = dist;
	t->dist = dist - p->dist;
	newp->left = t;
	newp->right = p;
	newp->parent = 0;
	newp->dist = 0.0;
	newp->leaf = NODE;
	if (t->left == p)
		t->left = t->parent;
	else
		t->right = t->parent;
	prev = t;
	q = t->parent;
	t->parent = newp;

	while (q != 0) {
		if (q->left == prev) {
			q->left = q->parent;
			q->parent = prev;
			td = q->dist;
			q->dist = prevdist;
			prevdist = td;
			prev = q;
			q = q->left;
		}
		else {
			q->right = q->parent;
			q->parent = prev;
			td = q->dist;
			q->dist = prevdist;
			prevdist = td;
			prev = q;
			q = q->right;
		}
	}

	/* remove the old root node */
	q = prev;
	if (q->left == 0) {
		dist = q->dist;
		q = q->right;
		q->dist += dist;
		q->parent = prev->parent;
		if (prev->parent->left == prev)
			prev->parent->left = q;
		else
			prev->parent->right = q;
		prev->right = 0;
	}
	else {
		dist = q->dist;
		q = q->left;
		q->dist += dist;
		q->parent = prev->parent;
		if (prev->parent->left == prev)
			prev->parent->left = q;
		else
			prev->parent->right = q;
		prev->left = 0;
	}

	return newp;
}

TC_NT_node TC_reroot(TC_NT_node ptree, int nseq, int ntotal, TC_NT_node **lu)
{
	TC_NT_node p, rootnode, rootptr;
	float diff, mindiff, mindepth = 1.0, maxdist;
	int first = TRUE;
	int i;

	rootptr = ptree;
	for (i = 0; i < ntotal; i++) {
		p = lu[0][i];
		if (p->parent == 0)
			diff = TC_calc_root_mean(p, &maxdist, nseq, lu);
		else
			diff = TC_calc_mean(p, &maxdist, nseq, lu);
		if (diff == 0 || (diff > 0 && diff < 2 * p->dist)) {
			if (maxdist < mindepth || first == TRUE) {
				first = FALSE;
				rootptr = p;
				mindepth = maxdist;
				mindiff = diff;
			}
		}
	}
	if (rootptr == ptree) {
		mindiff = rootptr->left->dist + rootptr->right->dist;
		rootptr = rootptr->right;
	}
	rootnode = TC_insert_root(rootptr, mindiff);
	diff = TC_calc_root_mean(rootnode, &maxdist, nseq, lu);

	return rootnode;
}
