/*
 * File: dp_gotoh_nw.cpp
 * Created: 06.06.2006
 * Last Modified: 23.08.2006
 * Author: Adrian Rospondek
 */


#include <string.h>
#include <stdio.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "dp_gotoh_nw.hpp"
#include "globals.hpp"
#include "util_algo.hpp"
#include "util_mem.hpp"
#include "util_str.hpp"

#define TRACE_TYPE int


static int TC_idscore_pairseq(char *s1, char *s2, int gop, int gep, int **m, const char *comp_mode)
{
	int **I, **D, **M, *P;
	int i, j, l1, l2, l, score, id, igop, match;

	l1 = strlen(s1);
	l2 = strlen(s2);
	TC_lower_string(s1);
	TC_lower_string(s2);

	I = TC_declare_int(6, l2 + 1);
	D = TC_declare_int(6, l2 + 1);
	M = TC_declare_int(6, l2 + 1);

	for (j = 0; j <= l2; j++) {
		D[0][j] = gep * j;
		M[0][j] = 2 * gep * j;
		D[4][j] = 0;
	}
	for (i = 1; i <= l1; i++) {
		I[1][0] = i * gep;
		M[1][0] = 2 * i * gep;
		for (j = 1; j <= l2; j++) {
			score = m[s1[i - 1] - 'a'][s2[j - 1] - 'a'];
			id = (s1[i - 1] == s2[j - 1]) ? 1 : 0;
			igop = (i == l1 || j == l2) ? 0 : gop;

			if ((D[0][j] + gep) > (M[0][j] + igop + gep)) {
				D[1][j] = D[0][j] + gep;
				D[3][j] = D[2][j];
				D[5][j] = D[4][j];
			}
			else {
				D[1][j] = M[0][j] + igop + gep;
				D[3][j] = M[2][j];
				D[5][j] = M[4][j];
			}

			if ((I[1][j - 1] + gep) > (M[1][j - 1] + igop + gep)) {
				I[1][j] = I[1][j - 1] + gep;
				I[3][j] = I[3][j - 1];
				I[5][j] = I[5][j - 1];
			}
			else {
				I[1][j] = M[1][j - 1] + igop + gep;
				I[3][j] = M[3][j - 1];
				I[5][j] = M[5][j - 1];
			}

			match = M[0][j - 1] + score;
			if (I[1][j] > match && I[1][j] > D[1][j]) {
				M[1][j] = I[1][j];
				M[3][j] = I[3][j];
				M[5][j] = I[5][j];
			}
			else if (D[1][j] > match) {
				M[1][j] = D[1][j];
				M[3][j] = D[3][j];
				M[5][j] = D[5][j];
			}
			else {
				M[1][j] = match;
				M[3][j] = M[2][j - 1] + id;
				M[5][j] = M[4][j - 1] + 1;
			}
		}
		P = I[0];
		I[0] = I[1];
		I[1] = P;
		P = I[2];
		I[2] = I[3];
		I[3] = P;
		P = I[4];
		I[4] = I[5];
		I[5] = P;

		P = D[0];
		D[0] = D[1];
		D[1] = P;
		P = D[2];
		D[2] = D[3];
		D[3] = P;
		P = D[4];
		D[4] = D[5];
		D[5] = P;

		P = M[0];
		M[0] = M[1];
		M[1] = P;
		P = M[2];
		M[2] = M[3];
		M[3] = P;
		P = M[4];
		M[4] = M[5];
		M[5] = P;
	}
	l = MIN(l1, l2);

	if (TC_strm(comp_mode, "sim")) {
		score = (l == 0) ? 0 : (M[2][l2] * 100) / M[4][l2];
	}
	else if (TC_strm(comp_mode, "sim2")) {
		score = (l == 0) ? 0 : (M[2][l2] * 100) / MAX(l1, l2);
	}
	else if (TC_strm(comp_mode, "cov")) {
		score = (l == 0) ? 0 : ((M[4][l2] * 100) / MAX(l1, l2));
	}

	TC_free_int(I);
	TC_free_int(D);
	TC_free_int(M);

	return score;
}

int TC_idscore_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	A->score_aln = A->score =
		TC_idscore_pairseq(A->seq_al[l_s[0][0]], A->seq_al[l_s[1][0]], CL->gop, CL->gep, CL->M, "sim");
	return A->score_aln;
}

int TC_gotoh_pair_wise (TC_Alignment *A,int*ns, int **l_s, TC_ConstraintList *CL) {


    int TG_MODE;
    int l_gop, l_gep;
    int gop, gep;
    int maximise;
/*VARIANLES FOR THE MULTIPLE SEQUENCE ALIGNMENT*/
    int a, b, i, j;

    int *cc;	
    int *dd,*ddg;
    int   e, eg;

    int lenal[2], len;
    int t, c,s, ch;
    int sub;
    int fop;
    int score=0;
    int **pos0;
    static char **al;
    char **aln;
    int ala, alb,LEN;
    char *buffer;
    char *char_buf;
/*trace back variables       */
    FILE       *long_trace=NULL;
    TRACE_TYPE *buf_trace=NULL;
    static TRACE_TYPE **trace;
    TRACE_TYPE k;
    TRACE_TYPE *tr;
    int long_trace_flag=0;
    int dim;
/********Prepare penalties*******/
    gop=CL->gop*SCORE_K;
    gep=CL->gep*SCORE_K;
    TG_MODE=CL->TG_MODE;
    maximise=CL->maximise;
	
	
/********************************/	
/*CLEAN UP AFTER USE*/
    if ( A==NULL)
    {
	TC_free_int (trace);
	trace=NULL;
	TC_free_char (al);
	al=NULL;
	return 0;
    }		

/*DO MEMORY ALLOCATION FOR DP*/

    lenal[0]=strlen (A->seq_al[l_s[0][0]]);
    lenal[1]=strlen (A->seq_al[l_s[1][0]]);
    len= MAX(lenal[0],lenal[1])+1;
    buf_trace=(TRACE_TYPE*)AR_allocate( len, sizeof (TRACE_TYPE));
    buffer=(char*)AR_allocate ( 2*len, sizeof (char));	
    al=TC_declare_char (2, 2*len);  
	
    char_buf= (char*)AR_allocate (2*len, sizeof (char));	
	

    dd = (int*)AR_allocate (len, sizeof (int));
	

    cc = (int*)AR_allocate (len, sizeof (int));
    ddg= (int*)AR_allocate (len, sizeof (int));
	

	
//     if ( len>=MAX_LEN_FOR_DP)
//     {
// 	long_trace_flag=1;
// 	long_trace=vtmpfile();
//     }
//     else
//     {
	   
	dim=(trace==NULL)?0:TC_read_array_size( trace,sizeof (int*));
	trace    =TC_realloc_int ( trace,dim,dim,MAX(0,len-dim), MAX(0,len-dim));
//    }
	
/*END OF MEMORY ALLOCATION*/
	
		       
    pos0=TC_aln2pos_simple ( A,-1, ns, l_s);


    cc[0]=0;		
    tr=(long_trace_flag)?buf_trace:trace[0];
    tr[0]=(TRACE_TYPE)1;
    for ( j=1; j<=lenal[1]; j++)tr[j]=(TRACE_TYPE)-1;

//    if (long_trace_flag)fwrite (buf_trace, sizeof ( TRACE_TYPE),lenal[1]+1, long_trace);
	
	
    t=(TG_MODE==0)?gop:0;
	

    for (cc[0]=0,j=1; j<=lenal[1]; j++)
    {
	    
	l_gop=(TG_MODE==0)?gop:0;
	l_gep=(TG_MODE==2)?0:gep;

	cc[j]=t=t+l_gep;
	dd[j]=  t+  gop;
    }

    t=(TG_MODE==0)?gop:0;	
	
    for (i=1; i<=lenal[0];i++)
    {			
	tr=(long_trace_flag)?buf_trace:trace[i];
	s=cc[0];

	l_gop=(TG_MODE==0)?gop:0;
	l_gep=(TG_MODE==2)?0:gep;
			
			
			
	cc[0]=c=t=t+l_gep;
	e=t+  gop;
	tr[0]=(TRACE_TYPE)1;

			

	for (eg=0,j=1; j<=lenal[1];j++)
	{				   
	    if (CL->evaluate_residue_pair==TC_evaluate_matrix_score && ns[0]==1 && ns[1]==1 && 1==2)
	    {
		sub=SCORE_K*CL->M[A->seq_al[l_s[0][0]][i]-'A'][A->seq_al[l_s[1][0]][j]-'A'] -SCORE_K*CL->nomatch;
	    }
	    else
	    {
		sub=(CL->get_dp_cost) (A, pos0, ns[0], l_s[0], i-1, pos0, ns[1], l_s[1],j-1,CL);	
	    }
	    /*get the best Insertion*/
	    l_gop=(i==lenal[0] || i==1 )?((TG_MODE==0)?gop:0):gop;
	    l_gep=(i==lenal[0] || i==1)?((TG_MODE==2)?0:gep):gep;
			

	    if ( TC_a_better_than_b ( e,c+l_gop, maximise)) eg++;
	    else eg=1;	
	    e=TC_best_of_a_b (e, c+l_gop, maximise)+l_gep;
				
	    /*Get the best deletion*/
	    l_gop=(j==lenal[1] || j==1)?((TG_MODE==0)?gop:0):gop;
	    l_gep=(j==lenal[1] || j==1)?((TG_MODE==2)?0:gep):gep;
				

	    if ( TC_a_better_than_b ( dd[j], cc[j]+l_gop, maximise))ddg[j]++;
	    else ddg[j]=1;
	    dd[j]=TC_best_of_a_b( dd[j], cc[j]+l_gop,maximise)+l_gep;
				


	    c=TC_best_int(3,maximise,&fop, e, s+sub,dd[j]);
	    /*Chose Substitution for tie breaking*/
	    if ( fop==0 && (s+sub)==e)fop=1;
	    else if ( fop==2 && (s+sub)==dd[j])fop=1;
	    /*Chose Deletion for tie breaking*/
	    else if ( fop==2 && e==dd[j])fop=1;

	    fop-=1;
	    s=cc[j];
	    cc[j]=c;	

	
	    if ( fop<0)
	    {tr[j]=(TRACE_TYPE)fop*eg;
	    }
	    else if ( fop>0)
	    {tr[j]=(TRACE_TYPE)fop*ddg[j];
	    }
	    else if (fop==0)
	    {tr[j]=(TRACE_TYPE)0;	
	    }					
	    fop= -2;
	}
//	if (long_trace_flag)
//	{
//	    fwrite ( buf_trace, sizeof (TRACE_TYPE), lenal[1]+1, long_trace);
//	}
    }
	
		
    score=c;
	
    i=lenal[0];
    j=lenal[1];
    ala=alb=0;
	

    while (i>=0 && j>=0 && ((i+j)!=0))
    {
	if ( i==0)
	    k=-1;
	else if ( j==0)
	    k=1;
	else if ( j==0 && i==0)
	    k=1;	
	else
	{
// 	    if (long_trace_flag)
// 	    {
// 		fseek ( long_trace, sizeof (TRACE_TYPE)*((lenal[1]+1)*(i)+j),SEEK_SET);
// 		fread ( &k, sizeof (TRACE_TYPE), 1, long_trace);
// 	    }
// 	    else
// 	    {
				   
		k=trace[i][j];
//	    }
	}
				
				
	if (k==0)
	{
				
	    al[0][ala++]=1;
	    al[1][alb++]=1;
	    i--;
	    j--;
	}		
	else if (k>0)
	{
				
	    for ( a=0; a< k; a++)
	    {
		al[0][ala++]=1;
		al[1][alb++]=0;
		i--;
	    }
	}
	else if (k<0)
	{
				
	    for ( a=0; a>k; a--)
	    {
		al[0][ala++]=0;
		al[1][alb++]=1;
		j--;
	    }
	}
    }
      
    LEN=ala;	
    c=LEN-1;  
	
	

    TC_invert_list_char ( al[0], LEN);
    TC_invert_list_char ( al[1], LEN);

    if ( A->declared_len<=LEN)A=TC_realloc_aln2  ( A,A->max_n_seq, 2*LEN);	
    aln=A->seq_al;

    for ( c=0; c< 2; c++)
    {
	for ( a=0; a< ns[c]; a++) 
	{		
	    ch=0;
	    for ( b=0; b< LEN; b++)
	    {		   
		if (al[c][b]==1)
		    char_buf[b]=aln[l_s[c][a]][ch++];
		else
		    char_buf[b]='-';
	    }
	    char_buf[b]='\0';
	    sprintf (aln[l_s[c][a]],"%s", char_buf);
	}
    }
	
	
    A->len_aln=LEN;
    A->nseq=ns[0]+ns[1];
	

    AR_free ( cc);
    AR_free (dd);		
    AR_free (ddg);
    AR_free (buffer);
    AR_free (char_buf); 
    AR_free (buf_trace);
    TC_free_char (al);
    TC_free_int (pos0);

//    if (long_trace_flag) fclose(long_trace);

    return score;
}
