/*
 * File: dp_mm_nw.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <stdlib.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "dp_mm_nw.hpp"
#include "evaluate.hpp"
#include "globals.hpp"
#include "util_mem.hpp"



static int g_last, *g_sapp;


#define gap(k) ((k) <= 0 ? 0 : (g + h * (k)))	/* k-symbol indel cost */

#define DEL(k)\
{\
	if (g_last < 0)\
		g_last = g_sapp[-1] -= (k);\
	else\
		g_last = *g_sapp++ = -(k);\
}

#define INS(k)\
{\
	if (g_last < 0) {\
		g_sapp[-1] = (k);\
		*g_sapp++ = g_last;\
	}\
	else\
		g_last = *g_sapp++ = (k);\
}

#define REP \
{\
	g_last = *g_sapp++ = 0;\
}

static int TC_diff(TC_Alignment *A, int *ns, int **l_s, int s1, int M, int s2, int N, int tb,
	int te, TC_ConstraintList *CL, int **pos)
{
	static int *CC;
	static int *DD;
	/* Forward cost-only vectors */
	static int *RR;
	static int *SS;
	/* Reverse cost-only vectors */
	int midi, midj, type;	/* Midpoint, type, and cost */
	int midc;

	if (!CC) {
		int L;

		L = M + N + 1;
		CC = (int *)AR_allocate(L, sizeof(int));
		DD = (int *)AR_allocate(L, sizeof(int));
		RR = (int *)AR_allocate(L, sizeof(int));
		SS = (int *)AR_allocate(L, sizeof(int));
	}

	if (A == 0) {
		AR_free(CC);
		AR_free(DD);
		AR_free(RR);
		AR_free(SS);
		CC = DD = RR = SS = 0;
		return 0;
	}

	{
		int i, j;
		int c, e, d, s;
		int t, g, h, m;

		g = CL->gop * SCORE_K;
		h = CL->gep * SCORE_K;
		m = g + h;

		if (N <= 0) {
			if (M > 0)
				DEL(M);
			return gap(M);
		}
		if (M <= 1) {
			if (M <= 0) {
				INS(N);
				return gap(N);
			}
			if (tb > te)
				tb = te;
			midc = (tb + h) + gap(N);
			midj = 0;
			for (j = 1; j <= N; j++) {
				if (CL->evaluate_residue_pair == TC_evaluate_matrix_score && ns[0] == 1
					&& ns[1] == 1) {
					c = gap(j - 1) + (SCORE_K * CL->M[A->seq_al[l_s[0][0]][s1] - 'A']
						[A->seq_al[l_s[1][0]][j - 1 + s2] - 'A'] - SCORE_K * CL->nomatch) + gap(N - j);
				}
				else {
					c = gap(j - 1) + (CL->get_dp_cost)(A, pos, ns[0], l_s[0], s1, pos, ns[1],
						l_s[1], j - 1 + s2, CL) + gap(N - j);
				}
				if (c > midc) {
					midc = c;
					midj = j;
				}
			}
			if (midj == 0) {
				DEL(1);
				INS(N);
			}
			else {
				if (midj > 1)
					INS(midj - 1);
				REP;
				if (midj < N)
					INS(N - midj);
			}

			return midc;
		}
		/* Divide: Find optimum midpoint (midi,midj) of cost midc */
		midi = M / 2;			/* Forward phase: */
		CC[0] = 0;				/* Compute C(M/2,k) & D(M/2,k) for all k */
		t = tb;
		for (j = 1; j <= N; j++) {
			CC[j] = t = t + h;
			DD[j] = t + g;
		}
		t = tb;
		for (i = 1; i <= midi; i++) {
			s = CC[0];
			CC[0] = c = t = t + h;
			e = t + g;
			for (j = 1; j <= N; j++) {
				if ((c = c + m) > (e = e + h))
					e = c;
				if ((c = CC[j] + m) > (d = DD[j] + h))
					d = c;

				if (CL->evaluate_residue_pair == TC_evaluate_matrix_score && ns[0] == 1
					&& ns[1] == 1) {
					c = s + (SCORE_K * CL->M[A->seq_al[l_s[0][0]][i - 1 + s1] - 'A']
						[A->seq_al[l_s[1][0]][j - 1 + s2] - 'A'] - SCORE_K * CL->nomatch);
				}
				else {
					c = s + (CL->get_dp_cost)(A, pos, ns[0], l_s[0], i - 1 + s1, pos, ns[1],
						l_s[1], j - 1 + s2, CL);
				}
				if (e > c)
					c = e;
				if (d > c)
					c = d;
				s = CC[j];
				CC[j] = c;
				DD[j] = d;
			}
		}
		DD[0] = CC[0];
		RR[N] = 0;		/* Reverse phase: */
		t = te;

		for (j = N - 1; j >= 0; j--) {
			RR[j] = t = t + h;
			SS[j] = t + g;
		}
		t = te;
		for (i = M - 1; i >= midi; i--) {
			s = RR[N];
			RR[N] = c = t = t + h;
			e = t + g;
			for (j = N - 1; j >= 0; j--) {
				if ((c = c + m) > (e = e + h))
					e = c;
				if ((c = RR[j] + m) > (d = SS[j] + h))
					d = c;

				if (CL->evaluate_residue_pair == TC_evaluate_matrix_score && ns[0] == 1
					&& ns[1] == 1 && 1 == 2) {
					c = s + SCORE_K * CL->M[A->seq_al[l_s[0][0]][i + s1] - 'A']
						[A->seq_al[l_s[1][0]][j + s2] - 'A'] - SCORE_K * CL->nomatch;
				}
				else {
					c = s + (CL->get_dp_cost)(A, pos, ns[0], l_s[0], i + s1, pos, ns[1], l_s[1],
						j + s2, CL);
				}
				if (e > c)
					c = e;
				if (d > c)
					c = d;
				s = RR[j];
				RR[j] = c;
				SS[j] = d;
			}
		}
		SS[N] = RR[N];
		midc = CC[0] + RR[0];	/* Find optimal midpoint */
		midj = 0;
		type = 1;
		for (j = 0; j <= N; j++)
			if ((c = CC[j] + RR[j]) >= midc)
				if (c > midc || CC[j] != DD[j] && RR[j] == SS[j]) {
					midc = c;
					midj = j;
				}
		for (j = N; j >= 0; j--)
			if ((c = DD[j] + SS[j] - g) > midc) {
				midc = c;
				midj = j;
				type = 2;
			}
	}
	/* Conquer: recursively around midpoint */
	if (type == 1) {
		TC_diff(A, ns, l_s, s1, midi, s2, midj, tb, CL->gop * SCORE_K, CL, pos);
		TC_diff(A, ns, l_s, s1 + midi, M - midi, s2 + midj, N - midj, CL->gop * SCORE_K, te, CL, pos);
	}
	else {
		TC_diff(A, ns, l_s, s1, midi - 1, s2, midj, tb, 0, CL, pos);
		DEL(2);
		TC_diff(A, ns, l_s, s1 + midi + 1, M - midi - 1, s2 + midj, N - midj, 0, te, CL, pos);
	}

	return midc;
}

int TC_myers_miller_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	char **char_buf;
	int *S;
	int **pos;
	int score;
	int a, b, i, j, l, l1, l2, len;

	pos = TC_aln2pos_simple(A, -1, ns, l_s);

	l1 = strlen(A->seq_al[l_s[0][0]]);
	l2 = strlen(A->seq_al[l_s[1][0]]);
	S = (int *)AR_allocate(l1 + l2 + 1, sizeof(int));
	g_last = 0;
	g_sapp = S;

	score = TC_diff(A, ns, l_s, 0, l1, 0, l2, 0, 0, CL, pos);
	TC_diff(0, ns, l_s, 0, l1, 0, l2, 0, 0, CL, pos);

	i = 0;
	j = 0;
	g_sapp = S;
	len = 0;
	while (!(i == l1 && j == l2)) {
		if (*g_sapp == 0) {
			i++;
			j++;
			len++;
		}
		else if (*g_sapp < 0) {
			i -= *g_sapp;
			len -= *g_sapp;
		}
		else if (*g_sapp > 0) {
			j += *g_sapp;
			len += *g_sapp;
		}
		g_sapp++;
	}

	A = TC_realloc_aln2(A, A->max_n_seq, len + 1);
	char_buf = TC_declare_char(A->max_n_seq, len + 1);

	i = 0;
	j = 0;
	g_sapp = S;
	len = 0;
	while (!(i == l1 && j == l2)) {
		if (*g_sapp == 0) {
			for (b = 0; b < ns[0]; b++)
				char_buf[l_s[0][b]][len] = A->seq_al[l_s[0][b]][i];
			for (b = 0; b < ns[1]; b++)
				char_buf[l_s[1][b]][len] = A->seq_al[l_s[1][b]][j];
			i++;
			j++;
			len++;
		}
		else if (*g_sapp > 0) {
			l = *g_sapp;
			for (a = 0; a < l; a++, j++, len++) {
				for (b = 0; b < ns[0]; b++)
					char_buf[l_s[0][b]][len] = '-';
				for (b = 0; b < ns[1]; b++)
					char_buf[l_s[1][b]][len] = A->seq_al[l_s[1][b]][j];
			}
		}
		else if (*g_sapp < 0) {
			l = -*g_sapp;
			for (a = 0; a < l; a++, i++, len++) {
				for (b = 0; b < ns[0]; b++)
					char_buf[l_s[0][b]][len] = A->seq_al[l_s[0][b]][i];;
				for (b = 0; b < ns[1]; b++)
					char_buf[l_s[1][b]][len] = '-';
			}
		}
		g_sapp++;
	}

	A->len_aln = len;
	A->nseq = ns[0] + ns[1];

	for (a = 0; a < ns[0]; a++) {
		char_buf[l_s[0][a]][len] = '\0';
		strncpy(A->seq_al[l_s[0][a]], char_buf[l_s[0][a]], A->declared_len - 1);
	}
	for (a = 0; a < ns[1]; a++) {
		char_buf[l_s[1][a]][len] = '\0';
		strncpy(A->seq_al[l_s[1][a]], char_buf[l_s[1][a]], A->declared_len - 1);
	}

	AR_free(S);
	TC_free_char(char_buf);
	TC_free_int(pos);

	l1 = strlen(A->seq_al[l_s[0][0]]);
	l2 = strlen(A->seq_al[l_s[1][0]]);

	if (l1 != l2)
		score = 0;

	return score;
}
