/*
 * File: dp_sim.cpp
 * Created: 06.06.2006
 * Last Modified: 26.08.2006
 * Author: Adrian Rospondek
 */


#include <stdio.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "dp_sim.hpp"
#include "globals.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"



#define SIM_GAP -1

#define TC_SCORE_2(x, y)\
	(SCORE_K * g_CL->M[g_Aln->seq_al[g_l_s[0][0]][x] - 'A'][g_Aln->seq_al[g_l_s[1][0]][y] - 'A']\
	 - SCORE_K * g_CL->nomatch)
#define TC_SCORE_N(x, y)\
	((g_CL->get_dp_cost)(g_Aln, g_pos, g_ns[0], g_l_s[0], x, g_pos, g_ns[1], g_l_s[1], y, g_CL))
#define TC_SCORE(x, y)\
	((g_CL->evaluate_residue_pair == TC_evaluate_matrix_score && g_ns[0] == 1 && g_ns[1] == 1) ?\
	 	TC_SCORE_2(x, y) : TC_SCORE_N(x, y))

#define min(x, y) ((x) <= (y) ? (x) : (y))
#define gap(k) ((k) <= 0 ? 0 : g_q + g_r * (k))



/* Memory management */
struct TC_mem {
	void *p;
	TC_mem *next;
};


struct TC_ONE {
	int COL;
	TC_ONE *NEXT;
};
typedef TC_ONE TC_pair, *TC_pair_ptr;

#define PAIRNULL (TC_pair_ptr)0


struct TC_SIM_NODE {
	int SIM_SCORE;
	int SIM_STARI;
	int SIM_STARJ;
	int SIM_ENDI;
	int SIM_ENDJ;
	int SIM_TOP;
	int SIM_BOT;
	int SIM_LEFT;
	int SIM_RIGHT;
};
typedef TC_SIM_NODE TC_vertex, *TC_vertex_ptr;


static TC_mem *g_first_mem, *g_last_mem;
static TC_Alignment *g_Aln;
static int *g_ns, **g_l_s;
static TC_ConstraintList *g_CL;
static int g_q, g_r; /* gap penalties */
static int g_qr;     /* qr = q + r */
static int **g_pos;
static int *g_seqc0, *g_seqc1;
static int g_min0, g_min1, g_max0, g_max1, g_mins;
static int *g_CC, *g_DD;               /* saving matrix scores */
static int *g_RR, *g_SS, *g_EE, *g_FF; /* saving start-points */
static int *g_HH, *g_WW;               /* saving matrix scores */
static int *g_II, *g_JJ, *g_XX, *g_YY; /* saving start-points */
static int g_m1, g_mm, g_n1, g_nn; /* boundaries of recomputed area */
static int g_rl, g_cl;             /* left and top boundaries */
static TC_pair_ptr *g_row, g_z; /* for saving used aligned pairs */
static TC_vertex_ptr *g_LIST;    /* an array for saving k best scores */
static TC_vertex_ptr g_low = 0;  /* lowest score node in LIST */
static TC_vertex_ptr g_most = 0; /* latestly accessed node in LIST */
static int g_numnode; /* the number of nodes in LIST */
static int g_lmin;    /* minimum score in LIST */
static int *g_sapp; /* Current script append ptr */
static int g_last;  /* Last script op appended */
static int g_I, g_J; /* current positions of A, B */
static int g_no_mat; /* number of matches */
static int g_no_mis; /* number of mismatches */
static int g_al_len; /* length of alignment */
static int g_flag; /* indicate if recomputation necessary */
static int g_tt;
static int g_zero = 0;



#define DEL(k)\
{\
	g_I += k;\
	g_al_len += k;\
	if (g_last < 0)\
		g_last = g_sapp[-1] -= (k);\
	else\
		g_last = *g_sapp++ = -(k);\
}

#define INS(k)\
{\
	g_J += k;\
	g_al_len += k;\
	if (g_last < 0) {\
		g_sapp[-1] = (k);\
		*g_sapp++ = g_last;\
	}\
	else\
		g_last = *g_sapp++ = (k);\
}

#define REP \
{\
	g_last = *g_sapp++ = 0;\
	g_al_len += 1;\
}

/* assigns value to x if (ii, jj) is never used before */
#define DIAG(ii, jj, x, value)\
{\
	for (g_tt = 1, g_z = g_row[(ii)]; g_z != PAIRNULL; g_z = g_z->NEXT)\
		if (g_z->COL == (jj)) {\
			g_tt = 0;\
			break;\
		}\
	if (g_tt)\
		x = (value);\
}

/* replace (ss1, xx1, yy1) by (ss2, xx2, yy2) if the latter is large */
#define ORDER(ss1, xx1, yy1, ss2, xx2, yy2)\
{\
	if (ss1 < ss2) {\
		ss1 = ss2;\
		xx1 = xx2;\
		yy1 = yy2;\
	}\
	else if (ss1 == ss2) {\
		if (xx1 < xx2) {\
			xx1 = xx2;\
			yy1 = yy2;\
		}\
		else if (xx1 == xx2 && yy1 < yy2)\
			yy1 = yy2;\
	}\
}

static void TC_sim_reset_static_variables()
{
	g_CC = g_DD = g_RR = g_SS = g_EE = g_FF = g_HH = g_WW = g_II = g_JJ = g_XX = g_YY = g_sapp = 0;
	g_min0 = g_min1 = g_max0 = g_max1 = g_mins = 0;
	g_q = g_r = g_qr = g_numnode = 0;
	g_m1 = g_n1 = g_nn = g_rl = g_cl = g_lmin = g_flag = g_last = g_I = g_J = g_tt = g_zero = 0;
	g_no_mat = g_no_mis = g_al_len = 0;
	g_most = g_low = 0; /* Very important: cause a bug if not reset */
	g_LIST = 0;         /* Very important: cause a bug if not reset */
}

static void *TC_sim_vcalloc(size_t nobj, size_t size)
{
	void *p;
	TC_mem *new_mem;
	
	p = AR_allocate(nobj, size);
	new_mem = (TC_mem *)AR_allocate(1, sizeof(TC_mem));
	
	if (g_last_mem == 0)
		g_first_mem = g_last_mem = new_mem;
	else {
		g_last_mem->next = new_mem;
		g_last_mem = new_mem;
	}
	g_last_mem->p = p;
	
	return p;
}

static void TC_sim_free_all()
{
	TC_mem *p1, *p2;
	
	p1 = g_first_mem;
	while (p1) {
		p2 = p1->next;
		AR_free(p1->p);
		AR_free(p1);
		p1 = p2;
	}
	
	g_first_mem = g_last_mem = 0;
	TC_sim_reset_static_variables();
}

/* Add a new node into list */
static int TC_addnode(int c, int ci, int cj, int i, int j, int K, int cost)
{
	int found; /* 1 if the node is in LIST */
	register int d;
	
	found = 0;
	if (g_most != 0 && g_most->SIM_STARI == ci && g_most->SIM_STARJ == cj)
		found = 1;
	else
		for (d = 0; d < g_numnode; d++) {
			g_most = g_LIST[d];
			if (g_most->SIM_STARI == ci && g_most->SIM_STARJ == cj) {
				found = 1;
				break;
			}
		}
	if (found) {
		if (g_most->SIM_SCORE < c) {
			g_most->SIM_SCORE = c;
			g_most->SIM_ENDI = i;
			g_most->SIM_ENDJ = j;
		}
		if (g_most->SIM_TOP > i)
			g_most->SIM_TOP = i;
		if (g_most->SIM_BOT < i)
			g_most->SIM_BOT = i;
		if (g_most->SIM_LEFT > j)
			g_most->SIM_LEFT = j;
		if (g_most->SIM_RIGHT < j)
			g_most->SIM_RIGHT = j;
	}
	else {
		if (g_numnode == K) /* list full */
			g_most = g_low;
		else
			g_most = g_LIST[g_numnode++];
		g_most->SIM_SCORE = c;
		g_most->SIM_STARI = ci;
		g_most->SIM_STARJ = cj;
		g_most->SIM_ENDI = i;
		g_most->SIM_ENDJ = j;
		g_most->SIM_TOP = g_most->SIM_BOT = i;
		g_most->SIM_LEFT = g_most->SIM_RIGHT = j;
	}
	if (g_numnode == K) {
		if (g_low == g_most || !g_low) {
			for (g_low = g_LIST[0], d = 1; d < g_numnode; d++)
				if (g_LIST[d]->SIM_SCORE < g_low->SIM_SCORE)
					g_low = g_LIST[d];
		}
		
		return g_low->SIM_SCORE;
	}
	else
		return cost;
}

/* A big pass to compute K best classes */
static int TC_big_pass(int A[], int B[], int M, int N, int K, int nseq)
{
	register int i, j; /* row and column indices */
	register int c; /* best score at current point */
	register int f; /* best score ending with insertion */
	register int d; /* best score ending with deletion */
	register int p; /* best score at (i-1, j-1) */
	register int ci, cj; /* end-point associated with c */
	register int di, dj; /* end-point associated with d */
	register int fi, fj; /* end-point associated with f */
	register int pi, pj; /* end-point associated with p */
	
	/* Compute the matrix and save the top K best scores in LIST
	 * CC : the scores of the current row
	 * RR and EE : the starting point that leads to score CC
	 * DD : the scores of the current row, ending with deletion
	 * SS and FF : the starting point that leads to score DD */
	
	/* Initialize the 0th row */
	for (j = 1; j <= N; j++) {
		g_CC[j] = 0;
		g_RR[j] = 0;
		g_EE[j] = j;
		g_DD[j] = -(g_q);
		g_SS[j] = 0;
		g_FF[j] = j;
	}
	for (i = 1; i <= M; i++) {
		c = 0; /* Initialize column 0 */
		f = -(g_q);
		ci = fi = i;
		if (nseq == 2) {
			p = 0;
			pi = i - 1;
			cj = fj = pj = 0;
		}
		else {
			p = g_CC[i];
			pi = g_RR[i];
			pj = g_EE[i];
			cj = fj = i;
		}
		for (j = (nseq == 2 ? 1 : (i + 1)); j <= N; j++) {
			f = f - g_r;
			c = c - g_qr;
			ORDER(f, fi, fj, c, ci, cj);
			c = g_CC[j] - g_qr;
			ci = g_RR[j];
			cj = g_EE[j];
			d = g_DD[j] - g_r;
			di = g_SS[j];
			dj = g_FF[j];
			ORDER(d, di, dj, c, ci, cj);
			c = 0;
			DIAG(i, j, c, p + TC_SCORE(A[i - 1], B[j - 1])); /* diagonal */
			
			if (c <= 0) {
				c = 0;
				ci = i;
				cj = j;
			}
			else {
				ci = pi;
				cj = pj;
			}
			ORDER(c, ci, cj, d, di, dj);
			ORDER(c, ci, cj, f, fi, fj);
			p = g_CC[j];
			g_CC[j] = c;
			pi = g_RR[j];
			pj = g_EE[j];
			g_RR[j] = ci;
			g_EE[j] = cj;
			g_DD[j] = d;
			g_SS[j] = di;
			g_FF[j] = dj;
			if (c > g_lmin) /* add the score into list */
				g_lmin = TC_addnode(c, ci, cj, i, j, K, g_lmin);
		}
	}
	
	return 1;
}

/* Recompute the area on forward pass */
static int TC_small_pass(int A[], int B[], int count, int nseq)
{
	register int i, j; /* row and column indices */
	register int c; /* best score at current point */
	register int f; /* best score ending with insertion */
	register int d; /* best score ending with deletion */
	register int p; /* best score at (i-1, j-1) */
	register int ci, cj; /* end-point associated with c */
	register int di, dj; /* end-point associated with d */
	register int fi, fj; /* end-point associated with f */
	register int pi, pj; /* end-point associated with p */
	int limit; /* lower bound on j */
	
	for (j = g_n1 + 1; j <= g_nn; j++) {
		g_CC[j] = 0;
		g_RR[j] = g_m1;
		g_EE[j] = j;
		g_DD[j] = -(g_q);
		g_SS[j] = g_m1;
		g_FF[j] = j;
	}
	for (i = g_m1 + 1; i <= g_mm; i++) {
		c = 0; /* Initialize column 0 */
		f = -(g_q);
		ci = fi = i;
		if (nseq == 2 || i <= g_n1) {
			p = 0;
			pi = i - 1;
			cj = fj = pj = g_n1;
			limit = g_n1 + 1;
		}
		else {
			p = g_CC[i];
			pi = g_RR[i];
			pj = g_EE[i];
			cj = fj = i;
			limit = i + 1;
		}
		for (j = limit; j <= g_nn; j++) {
			f = f - g_r;
			c = c - g_qr;
			ORDER(f, fi, fj, c, ci, cj);
			c = g_CC[j] - g_qr;
			ci = g_RR[j];
			cj = g_EE[j];
			d = g_DD[j] - g_r;
			di = g_SS[j];
			dj = g_FF[j];
			ORDER(d, di, dj, c, ci, cj);
			c = 0;
			DIAG(i, j, c, p + TC_SCORE(A[i - 1], B[j - 1])); /* diagonal */
			
			if (c <= 0) {
				c = 0;
				ci = i;
				cj = j;
			}
			else {
				ci = pi;
				cj = pj;
			}
			ORDER(c, ci, cj, d, di, dj);
			ORDER(c, ci, cj, f, fi, fj);
			p = g_CC[j];
			g_CC[j] = c;
			pi = g_RR[j];
			pj = g_EE[j];
			g_RR[j] = ci;
			g_EE[j] = cj;
			g_DD[j] = d;
			g_SS[j] = di;
			g_FF[j] = dj;
			if (c > g_lmin) /* add the score into list */
				g_lmin = TC_addnode(c, ci, cj, i, j, count, g_lmin);
		}
	}
	
	return 1;
}

/* Find and remove the largest score in list */
static TC_vertex_ptr TC_findmax()
{
	TC_vertex_ptr cur;
	register int i, j;
	
	for (j = 0, i = 1; i < g_numnode; i++)
		if (g_LIST[i]->SIM_SCORE > g_LIST[j]->SIM_SCORE)
			j = i;
	cur = g_LIST[j];
	
	if (j != --g_numnode) {
		g_LIST[j] = g_LIST[g_numnode];
		g_LIST[g_numnode] = cur;
	}
	g_most = g_LIST[0];
	
	if (g_low == cur)
		g_low = g_LIST[0];
	
	return cur;
}

/*
 * diff(A,B,M,N,tb,te) returns the score of an optimum conversion between
 * A[1..M] and B[1..N] that begins(ends) with a delete if tb(te) is zero
 * and appends such a conversion to the current script.
 */
static int TC_diff_sim(int A[], int B[], int M, int N, int tb, int te)
{
	int midi, midj, type; /* midpoint, type and cost */
	int midc;
	
	{
		register int i, j;
		register int c, e, d, s;
		int t;
		
		/* Boundary cases: M <= 1 or N == 0 */
		if (N <= 0) {
			if (M > 0)
				DEL(M);
			
			return -gap(M);
		}
		if (M <= 1) {
			if (M <= 0) {
				INS(N);
				
				return -gap(N);
			}
			if (tb > te)
				tb = te;
			midc = -(tb + g_r + gap(N));
			midj = 0;
			for (j = 1; j <= N; j++) {
				for (g_tt = 1, g_z = g_row[g_I + 1]; g_z != PAIRNULL; g_z = g_z->NEXT)
					if (g_z->COL == j + g_J) {
						g_tt = 0;
						break;
					}
				if (g_tt) {
					c = TC_SCORE(A[0], B[j - 1]) - (gap(j - 1) + gap(N - j));
					if (c > midc) {
						midc = c;
						midj = j;
					}
				}
			}
			if (midj == 0) {
				INS(N);
				DEL(1);
			}
			else {
				if (midj > 1)
					INS(midj - 1);
				REP;
				if (A[1] == B[midj])
					g_no_mat += 1;
				else
					g_no_mis += 1;
				/* mark (A[I],B[J]) as used: put J into list row[I] */
				g_I++;
				g_J++;
				g_z = (TC_pair_ptr)TC_sim_vcalloc(1, sizeof(TC_pair));
				g_z->COL = g_J;
				g_z->NEXT = g_row[g_I];
				g_row[g_I] = g_z;
				if (midj < N)
					INS(N - midj);
			}
			
			return midc;
		}
		
		/* Divide: Find optimum midpoint (midi,midj) of cost midc */
		midi = M / 2; /* Forward phase: */
		g_CC[0] = 0;  /* Compute C(M/2,k) & D(M/2,k) for all k */
		t = -g_q;
		for (j = 1; j <= N; j++) {
			g_CC[j] = t = t - g_r;
			g_DD[j] = t - g_q;
		}
		t = -tb;
		for (i = 1; i <= midi; i++) {
			s = g_CC[0];
			g_CC[0] = c = t = t - g_r;
			e = t - g_q;
			for (j = 1; j <= N; j++) {
				if ((c = c - g_qr) > (e = e - g_r))
					e = c;
				if ((c = g_CC[j] - g_qr) > (d = g_DD[j] - g_r))
					d = c;
				DIAG(i + g_I, j + g_J, c, s + TC_SCORE(A[i - 1], B[j - 1]));
				if (c < d)
					c = d;
				if (c < e)
					c = e;
				s = g_CC[j];
				g_CC[j] = c;
				g_DD[j] = d;
			}
		}
		g_DD[0] = g_CC[0];
		
		g_RR[N] = 0; /* Reverse phase: */
		t = -g_q;    /* Compute R(M/2,k) & S(M/2,k) for all k */
		for (j = N - 1; j >= 0; j--) {
			g_RR[j] = t = t - g_r;
			g_SS[j] = t - g_q;
		}
		t = -te;
		for (i = M - 1; i >= midi; i--) {
			s = g_RR[N];
			g_RR[N] = c = t = t - g_r;
			e = t - g_q;
			for (j = N - 1; j >= 0; j--) {
				if ((c = c - g_qr) > (e = e - g_r))
					e = c;
				if ((c = g_RR[j] - g_qr) > (d = g_SS[j] - g_r))
					d = c;
				DIAG(i + 1 + g_I, j + 1 + g_J, c, s + TC_SCORE(A[i], B[j])); /* not -1 on purpose */
				if (c < d)
					c = d;
				if (c < e)
					c = e;
				s = g_RR[j];
				g_RR[j] = c;
				g_SS[j] = d;
			}
		}
		g_SS[N] = g_RR[N];
		
		midc = g_CC[0] + g_RR[0]; /* Find optimal midpoint */
		midj = 0;
		type = 1;
		for (j = 0; j <= N; j++)
			if ((c = g_CC[j] + g_RR[j]) >= midc)
				if (c > midc || g_CC[j] != g_DD[j] && g_RR[j] == g_SS[j]) {
					midc = c;
					midj = j;
				}
		for (j = N; j >= 0; j--)
			if ((c = g_DD[j] + g_SS[j] + g_q) > midc) {
				midc = c;
				midj = j;
				type = 2;
			}
	}
	
	/* Conquer: recursively around midpoint */
	if (type == 1) {
		TC_diff_sim(A, B, midi, midj, tb, g_q);
		TC_diff_sim(A + midi, B + midj, M - midi, N - midj, g_q, te);
	}
	else {
		TC_diff_sim(A, B, midi - 1, midj, tb, g_zero);
		DEL(2);
		TC_diff_sim(A + midi + 1, B + midj, M - midi - 1, N - midj, g_zero, te);
	}
	
	return midc;
}

static int TC_calcons(int *aa0, int *aa1, int *res, int *nc, int *nident, int *ns,
	int **l_s)
{
	int i0, i1;
	int op, nid, lenc, nd;
	int *sp0, *sp1;
	int *rp;
	int a, b, id_col, tot_col, r0, r1;
	
	g_min0--;
	g_min1--;
	
	sp0 = g_seqc0 + g_mins;
	sp1 = g_seqc1 + g_mins;
	rp = res;
	lenc = nid = op = 0;
	i0 = g_min0;
	i1 = g_min1;
	
	while (i0 < g_max0 || i1 < g_max1) {
		if (op == 0 && *rp == 0) {
			op = *rp++;
			*sp0 = aa0[i0++];
			*sp1 = aa1[i1++];
			for (id_col = tot_col = 0, a = 0; a < ns[0]; a++)
				for (b = 0; b < ns[1]; b++) {
					r0 = g_Aln->seq_al[l_s[0][a]][*sp0 - 1];
					r1 = g_Aln->seq_al[l_s[1][a]][*sp1 - 1];
					if (!TC_is_gap(r0) && r1 == r0)
						id_col++;
					if (!TC_is_gap(r0) && !TC_is_gap(r1))
						tot_col++;
				}
			nid += (tot_col) ? (id_col / tot_col) : 0;
			lenc++;
			sp0++;
			sp1++;
		}
		else {
			if (op == 0)
				op = *rp++;
			if (op > 0) {
				*sp0++ = SIM_GAP;
				*sp1++ = aa1[i1++];
				op--;
				lenc++;
			}
			else {
				*sp0++ = aa0[i0++];
				*sp1++ = SIM_GAP;
				op++;
				lenc++;
			}
		}
	}
	*nident = nid;
	*nc = lenc;
	nd = 0;
	
	return g_mins + lenc + nd;
}

/* return 1 if no node in LIST share vertices with the area */
static int TC_no_cross()
{
	TC_vertex_ptr cur;
	register int i;
	
	for (i = 0; i < g_numnode; i++) {
		cur = g_LIST[i];
		if (cur->SIM_STARI <= g_mm && cur->SIM_STARJ <= g_nn && cur->SIM_BOT >= g_m1 - 1
				&& cur->SIM_RIGHT >= g_n1 - 1 && (cur->SIM_STARI < g_rl || cur->SIM_STARJ < g_cl)) {
			if (cur->SIM_STARI < g_rl)
				g_rl = cur->SIM_STARI;
			if (cur->SIM_STARJ < g_cl)
				g_cl = cur->SIM_STARJ;
			g_flag = 1;
			break;
		}
	}
	
	if (i == g_numnode)
		return 1;
	else
		return 0;
}

/* Determine the left and top boundaries of the recomputed area */
static int TC_locate(int A[], int B[], int nseq)
{
	register int i, j; /* row and column indices */
	register int c; /* best score at current point */
	register int f; /* best score ending with insertion */
	register int d;	/* best score ending with deletion */
	register int p; /* best score at (i-1, j-1) */
	register int ci, cj; /* end-point associated with c */
	register int di, dj; /* end-point associated with d */
	register int fi, fj; /* end-point associated with f */
	register int pi, pj; /* end-point associated with p */
	int cflag, rflag; /* for recomputation */
	int limit; /* the bound on j */
	
	/* Reverse pass
	 * rows
	 * CC : the scores on the current row
	 * RR and EE : the endpoints that lead to CC
	 * DD : the deletion scores
	 * SS and FF : the endpoints that lead to DD
	 * columns
	 * HH : the scores on the current columns
	 * II and JJ : the endpoints that lead to HH
	 * WW : the deletion scores
	 * XX and YY : the endpoints that lead to WW */
	for (j = g_nn; j >= g_n1; j--) {
		g_CC[j] = 0;
		g_EE[j] = j;
		g_DD[j] = -(g_q);
		g_FF[j] = j;
		if (nseq == 2 || j > g_mm)
			g_RR[j] = g_SS[j] = g_mm + 1;
		else
			g_RR[j] = g_SS[j] = j;
	}
	for (i = g_mm; i >= g_m1; i--) {
		c = p = 0;
		f = -(g_q);
		ci = fi = i;
		pi = i + 1;
		cj = fj = pj = g_nn + 1;
		if (nseq == 2 || g_n1 > i)
			limit = g_n1;
		else
			limit = i + 1;
		for (j = g_nn; j >= limit; j--) {
			f = f - g_r;
			c = c - g_qr;
			ORDER(f, fi, fj, c, ci, cj);
			c = g_CC[j] - g_qr;
			ci = g_RR[j];
			cj = g_EE[j];
			d = g_DD[j] - g_r;
			di = g_SS[j];
			dj = g_FF[j];
			ORDER(d, di, dj, c, ci, cj);
			c = 0;
			DIAG(i, j, c, p + TC_SCORE(A[i - 1], B[j - 1])); /* diagonal */
			
			if (c <= 0) {
				c = 0;
				ci = i;
				cj = j;
			}
			else {
				ci = pi;
				cj = pj;
			}
			ORDER(c, ci, cj, d, di, dj);
			ORDER(c, ci, cj, f, fi, fj);
			p = g_CC[j];
			g_CC[j] = c;
			pi = g_RR[j];
			pj = g_EE[j];
			g_RR[j] = ci;
			g_EE[j] = cj;
			g_DD[j] = d;
			g_SS[j] = di;
			g_FF[j] = dj;
			if (c > g_lmin)
				g_flag = 1;
		}
		if (nseq == 2 || i < g_n1) {
			g_HH[i] = g_CC[g_n1];
			g_II[i] = g_RR[g_n1];
			g_JJ[i] = g_EE[g_n1];
			g_WW[i] = g_DD[g_n1];
			g_XX[i] = g_SS[g_n1];
			g_YY[i] = g_FF[g_n1];
		}
	}
	for (g_rl = g_m1, g_cl = g_n1; ;) {
		for (rflag = cflag = 1; (rflag && g_m1 > 1) || (cflag && g_n1 > 1); ) {
			if (rflag && g_m1 > 1) { /* Compute one row */
				rflag = 0;
				g_m1--;
				c = p = 0;
				f = -(g_q);
				ci = fi = g_m1;
				pi = g_m1 + 1;
				cj = fj = pj = g_nn + 1;
				for (j = g_nn; j >= g_n1; j--) {
					f = f - g_r;
					c = c - g_qr;
					ORDER(f, fi, fj, c, ci, cj);
					c = g_CC[j] - g_qr;
					ci = g_RR[j];
					cj = g_EE[j];
					d = g_DD[j] - g_r;
					di = g_SS[j];
					dj = g_FF[j];
					ORDER(d, di, dj, c, ci, cj);
					c = 0;
					DIAG(g_m1, j, c, TC_SCORE(A[g_m1 - 1], B[j - 1])); /* diagonal */
					
					if (c <= 0) {
						c = 0;
						ci = g_m1;
						cj = j;
					}
					else {
						ci = pi;
						cj = pj;
					}
					ORDER(c, ci, cj, d, di, dj);
					ORDER(c, ci, cj, f, fi, fj);
					p = g_CC[j];
					g_CC[j] = c;
					pi = g_RR[j];
					pj = g_EE[j];
					g_RR[j] = ci;
					g_EE[j] = cj;
					g_DD[j] = d;
					g_SS[j] = di;
					g_FF[j] = dj;
					
					if (c > g_lmin)
						g_flag = 1;
					if (!rflag && (ci > g_rl && cj > g_cl || di > g_rl && dj > g_cl
								|| fi > g_rl && fj > g_cl))
						rflag = 1;
				}
				g_HH[g_m1] = g_CC[g_n1];
				g_II[g_m1] = g_RR[g_n1];
				g_JJ[g_m1] = g_EE[g_n1];
				g_WW[g_m1] = g_DD[g_n1];
				g_XX[g_m1] = g_SS[g_n1];
				g_YY[g_m1] = g_FF[g_n1];
				
				if (!cflag && (ci > g_rl && cj > g_cl || di > g_rl && dj > g_cl
							|| fi > g_rl && fj > g_cl))
					cflag = 1;
			}
			if (nseq == 1 && g_n1 == (g_m1 + 1) && !rflag)
				cflag = 0;
			if (cflag && g_n1 > 1) { /* Compute one column */
				cflag = 0;
				g_n1--;
				c = 0;
				f = -(g_q);
				cj = fj = g_n1;
				if (nseq == 2 || g_mm < g_n1) {
					p = 0;
					ci = fi = pi = g_mm + 1;
					pj = g_n1 + 1;
					limit = g_mm;
				}
				else {
					p = g_HH[g_n1];
					pi = g_II[g_n1];
					pj = g_JJ[g_n1];
					ci = fi = g_n1;
					limit = g_n1 - 1;
				}
				for (i = limit; i >= g_m1; i--) {
					f = f - g_r;
					c = c - g_qr;
					ORDER(f, fi, fj, c, ci, cj);
					c = g_HH[i] - g_qr;
					ci = g_II[i];
					cj = g_JJ[i];
					d = g_WW[i] - g_r;
					di = g_XX[i];
					dj = g_YY[i];
					ORDER(d, di, dj, c, ci, cj);
					c = 0;
					DIAG(i, g_n1, c, p + TC_SCORE(A[i - 1], B[g_n1 - 1]));
					
					if (c <= 0) {
						c = 0;
						ci = i;
						cj = g_n1;
					}
					else {
						ci = pi;
						cj = pj;
					}
					ORDER(c, ci, cj, d, di, dj);
					ORDER(c, ci, cj, f, fi, fj);
					p = g_HH[i];
					g_HH[i] = c;
					pi = g_II[i];
					pj = g_JJ[i];
					g_II[i] = ci;
					g_JJ[i] = cj;
					g_WW[i] = d;
					g_XX[i] = di;
					g_YY[i] = dj;
					
					if (c > g_lmin)
						g_flag = 1;
					if (!cflag && (ci > g_rl && cj > g_cl || di > g_rl && dj > g_cl
								|| fi > g_rl && fj > g_cl))
						cflag = 1;
				}
				g_CC[g_n1] = g_HH[g_m1];
				g_RR[g_n1] = g_II[g_m1];
				g_EE[g_n1] = g_JJ[g_m1];
				g_DD[g_n1] = g_WW[g_m1];
				g_SS[g_n1] = g_XX[g_m1];
				g_FF[g_n1] = g_YY[g_m1];
				
				if (!rflag && (ci > g_rl && cj > g_cl || di > g_rl && dj > g_cl
							|| fi > g_rl && fj > g_cl))
					rflag = 1;
			}
		}
		if (g_m1 == 1 && g_n1 == 1 || TC_no_cross())
			break;
	}
	g_m1--;
	g_n1--;
	
	return 1;
}

/*
 * SIM(A,B,M,N,K,V,Q,R) reports K best non-intersecting alignments of the segments of A and B
 * in order of similarity scores, where V[a][b] is the score of aligning a and b
 * and -(Q+R*i) is the score of an i-symbol indel.
 */
int TC_sim_pair_wise_lalign(TC_Alignment *in_A, int *in_ns, int **in_l_s, TC_ConstraintList *in_CL)
{
	TC_vertex_ptr cur; /* temporary pointer */
	double percent;
	int endi, endj, stari, starj; /* endpoint and startpoint */
	int score; /* the max score in LIST */
	int count; /* maximum size of list */
	int *S; /* saving operations for diff */
	int nc, nident; /* for display */
	int t1, t2, g1, g2, r1, r2;
	int a, b, c, d, e, i;
	/* cedric was here 11/2/99 */
	int CEDRIC_MAX_N_ALN = 999;
	int CEDRIC_THRESHOLD = 50;
	int *A, *B;
	int M, N, K, maxl;
	int nseq;
	int R, Q;
	TC_Alignment *DA;
	
	DA = in_A;
	g_Aln = TC_copy_aln(in_A, 0);
	
	g_l_s = in_l_s;
	g_ns = in_ns;
	g_CL = in_CL;
	K = g_CL->lalign_n_top;
	
	M = strlen(g_Aln->seq_al[g_l_s[0][0]]);
	N = strlen(g_Aln->seq_al[g_l_s[1][0]]);
	maxl = M + N + 1;
	
	g_pos = TC_aln2pos_simple(g_Aln, -1, g_ns, g_l_s);
	
	g_seqc0 = (int *)TC_sim_vcalloc(maxl, sizeof(int));
	A = (int *)TC_sim_vcalloc(maxl, sizeof(int));
	for (a = 0; a < maxl; a++)
		g_seqc0[a] = A[a] = a;
	A[M + 1] = '\0';
	
	g_seqc1 = (int *)TC_sim_vcalloc(maxl, sizeof(int));
	B = (int *)TC_sim_vcalloc(maxl, sizeof(int));
	for (a = 0; a < maxl; a++)
		g_seqc1[a] = B[a] = a;
	B[N + 1] = '\0';
	
	nseq = (g_l_s[0][0] != g_l_s[1][0]) ? 2 : 1;
	
	Q = MAX(g_CL->gop, -g_CL->gop) * SCORE_K;
	R = MAX(g_CL->gep, -g_CL->gep) * SCORE_K;
	
	if (K == CEDRIC_MAX_N_ALN)
		K--;
	else if (K < 0) {
		CEDRIC_THRESHOLD = -K;
		K = CEDRIC_MAX_N_ALN;
	}
	// XXX: what if K == 0, it causes function CRASH !!
	
	/* allocate space for all vectors */
	g_CC = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_DD = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_RR = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_SS = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_EE = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_FF = (int *)TC_sim_vcalloc(N + 1, sizeof(int));
	g_HH = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	g_WW = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	g_II = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	g_JJ = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	g_XX = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	g_YY = (int *)TC_sim_vcalloc(M + 1, sizeof(int));
	S = (int *)TC_sim_vcalloc(min(M, N) * 5 / 4 + 1, sizeof(int));
	g_row = (TC_pair_ptr *)TC_sim_vcalloc(M + 1, sizeof(TC_pair_ptr));
	
	/* set up list for each row */
	if (nseq == 2)
		for (i = 1; i <= M; i++)
			g_row[i] = PAIRNULL;
	else {
		g_z = (TC_pair_ptr)TC_sim_vcalloc(M, sizeof(TC_pair));
		for (i = 1; i <= M; i++, g_z++) {
			g_row[i] = g_z;
			g_z->COL = i;
			g_z->NEXT = PAIRNULL;
		}
	}
	
	g_q = Q;
	g_r = R;
	g_qr = g_q + g_r;
	
	g_LIST = (TC_vertex_ptr *)TC_sim_vcalloc(K, sizeof(TC_vertex_ptr));
	for (i = 0; i < K; i++)
		g_LIST[i] = (TC_vertex_ptr)TC_sim_vcalloc(1, sizeof(TC_vertex));
	
	g_numnode = g_lmin = 0;
	TC_big_pass(A, B, M, N, K, nseq);
	
	/* Report the K best alignments one by one. After each alignment is output,
	 * recompute part of the matrix. First determine the size of the area to be
	 * recomputed, then do the recomputation. */
	
	for (count = K - 1; count >= 0; count--) {
		if (g_numnode == 0) {
			TC_padd_aln(in_A);
			TC_sim_free_all();
			
			return 1;
		}
		cur = TC_findmax(); /* Return a pointer to a node with max score */
		score = cur->SIM_SCORE;
		if (K == CEDRIC_MAX_N_ALN && score < CEDRIC_THRESHOLD)
			break;
		stari = ++cur->SIM_STARI;
		starj = ++cur->SIM_STARJ;
		endi = cur->SIM_ENDI;
		endj = cur->SIM_ENDJ;
		g_m1 = cur->SIM_TOP;
		g_mm = cur->SIM_BOT;
		g_n1 = cur->SIM_LEFT;
		g_nn = cur->SIM_RIGHT;
		g_rl = endi - stari + 1;
		g_cl = endj - starj + 1;
		g_I = stari - 1;
		g_J = starj - 1;
		g_sapp = S;
		g_last = 0;
		g_al_len = 0;
		g_no_mat = 0;
		g_no_mis = 0;
		TC_diff_sim(&A[stari] - 1, &B[starj] - 1, g_rl, g_cl, g_q, g_q);
		
		g_min0 = stari;
		g_min1 = starj;
		g_max0 = stari + g_rl - 1;
		g_max1 = starj + g_cl - 1;
		TC_calcons(A + 1, B + 1, S, &nc, &nident, g_ns, g_l_s);
		percent = (double)nident * 100.0 / (double)nc;
		
		/* Min0: index of the last residue before the first in a 1..N+1 numerotation */
		if (!DA->A)
			DA->A = TC_copy_aln(g_Aln, DA->A);
		DA->A = TC_realloc_aln(DA->A, nc + 1);
		DA = DA->A;
		DA->A = 0;
		
		for (c = 0; c < 2; c++) {
			for (a = 0; a < g_ns[c]; a++) {
				e = (c == 0) ? g_min0 : g_min1;
				for (d = 0; d < e; d++) {
					DA->order[g_l_s[c][a]][1] += 1 - TC_is_gap(g_Aln->seq_al[g_l_s[c][a]][d]);
				}
			}
		}
		for (t1 = g_min0, t2 = g_min1, a = 0; a < nc; a++) {
			r1 = g_seqc0[a];
			r2 = g_seqc1[a];
			g1 = (r1 == SIM_GAP || r1 > M) ? 0 : 1;
			g2 = (r2 == SIM_GAP || r2 > N) ? 0 : 1;
			t1 += g1;
			t2 += g2;
			for (b = 0; b < g_ns[0]; b++)
				DA->seq_al[g_l_s[0][b]][a] = (g1) ? g_Aln->seq_al[g_l_s[0][b]][A[t1 - 1]] : '-';
			for (b = 0; b < g_ns[1]; b++)
				DA->seq_al[g_l_s[1][b]][a] = (g2) ? g_Aln->seq_al[g_l_s[1][b]][B[t2 - 1]] : '-';
		}
		for (b = 0; b < g_ns[0]; b++) {
			DA->seq_al[g_l_s[0][b]][a] = '\0';
		}
		for (b = 0; b < g_ns[1]; b++) {
			DA->seq_al[g_l_s[1][b]][a] = '\0';
		}
		
		DA->nseq = g_ns[0] + g_ns[1];
		DA->len_aln = nc;
		DA->score = (int)percent;
		DA->score_aln = score;
		
		if (count) {
			g_flag = 0;
			TC_locate(A, B, nseq);
			if (g_flag)
				TC_small_pass(A, B, count, nseq);
		}
	}
	TC_padd_aln(in_A);
	
	TC_sim_free_all();
	TC_free_int(g_pos);
	TC_free_aln(g_Aln);
	
	return 1;
}
