/***
 *  $Id: par_tree_aln.cpp 728 2007-03-15 04:28:37Z zola $
 **
 *  File: par_tree_aln.cpp
 *  Created: Nov 06, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Xiao Yang <xyang@iastate.edu>
 *          Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#include "par_tree_aln.hpp"

#if WITH_TREE
static void print_job(const JobNodePtr job) {
    std::cout << "\n" << job->id << " " << job->name << " " 
	      << job->l_leaf_num << " " << job->r_leaf_num << " "
	      << job->tot_leaf
	      << std::endl;
} // print_job
#else
#define print_job(job)
#endif


std::size_t traverse_phyl_tree(TC_NT_node root, std::size_t root_idx,
			       job_list_type& j_list) {
    // TODO: optimise recursion
    // std::cout << "root: " << root_idx << " ";

    TC_NT_node LT = root->left;
    TC_NT_node RT = root->right;

    if ((LT->leaf == 1) && (RT->leaf == 1)) {
	// std::cout << "case 0" << "\n";
	j_list[root_idx].l_leaf_num = 1;
	j_list[root_idx].r_leaf_num = 1;
	j_list[root_idx].in_jobs = 2;

#if DEBUG_TREE
	j_list[root_idx].name = "(";
	j_list[root_idx].name += LT->name;
	j_list[root_idx].name += ",";
	j_list[root_idx].name += RT->name;
	j_list[root_idx].name += ")";
#endif
	
	return root_idx + 1;
    }

    int new_idx = -1;
    int free_idx = root_idx + 1;
    
    if ((LT->leaf == 1) && (RT->leaf == 0)) {
	// std::cout << "case 2" << "\n";
	j_list[root_idx].r_child = &j_list[free_idx];
	j_list[root_idx].l_leaf_num = 1;
	j_list[root_idx].in_jobs = 1;
	j_list[free_idx].pNode = RT;
	j_list[free_idx].parent = &j_list[root_idx];
	new_idx = traverse_phyl_tree(RT, free_idx, j_list);

	j_list[root_idx].r_leaf_num = j_list[free_idx].l_leaf_num +
	    j_list[free_idx].r_leaf_num;

#if DEBUG_TREE
	j_list[root_idx].name = "(";
	j_list[root_idx].name += LT->name;
	j_list[root_idx].name += ",";
	j_list[root_idx].name += j_list[free_idx].name;
	j_list[root_idx].name += ")";
#endif
	// return new_idx;
    }

    if ((LT->leaf == 0) && (RT->leaf == 1)) {
	// std::cout << "case 1" << "\n";
	j_list[root_idx].l_child = &j_list[free_idx];
	j_list[root_idx].r_leaf_num = 1;
	j_list[root_idx].in_jobs = 1;
	j_list[free_idx].pNode = LT;
	j_list[free_idx].parent = &j_list[root_idx];
	new_idx = traverse_phyl_tree(LT, free_idx, j_list);

	j_list[root_idx].l_leaf_num = j_list[free_idx].l_leaf_num +
	    j_list[free_idx].r_leaf_num;

#if DEBUG_TREE
	j_list[root_idx].name = "(";
	j_list[root_idx].name += j_list[free_idx].name;
	j_list[root_idx].name += ",";
	j_list[root_idx].name += RT->name;
	j_list[root_idx].name += ")";
#endif	
	// return new_idx;
    }

    if ((LT->leaf == 0) && (RT->leaf == 0)) {
	// std::cout << "case 3" << "\n";
	j_list[root_idx].l_child = &j_list[free_idx];
	j_list[root_idx].in_jobs = 0;
	j_list[free_idx].pNode = LT;
	j_list[free_idx].parent = &j_list[root_idx];
	new_idx = traverse_phyl_tree(LT, free_idx, j_list);

	j_list[root_idx].l_leaf_num = j_list[free_idx].l_leaf_num +
	    j_list[free_idx].r_leaf_num;

#if DEBUG_TREE
	j_list[root_idx].name = "(";
	j_list[root_idx].name += j_list[free_idx].name;
	j_list[root_idx].name += ",";
#endif

	free_idx = new_idx;

	j_list[root_idx].r_child = &j_list[free_idx];
	j_list[root_idx].in_jobs = 0;
	j_list[free_idx].pNode = RT;
	j_list[free_idx].parent = &j_list[root_idx];
	new_idx = traverse_phyl_tree(RT, free_idx, j_list);

	j_list[root_idx].r_leaf_num = j_list[free_idx].l_leaf_num +
	    j_list[free_idx].r_leaf_num;

#if DEBUG_TREE
	j_list[root_idx].name += j_list[free_idx].name;
	j_list[root_idx].name += ")";
#endif
    }

    return new_idx;
} // traverse_phyl_tree


void send_fin_to(int host, const MPI_Env& cl_env) {
    int fin = -1;
    MPI_Send(&fin, 1, MPI_INT, host, PA_JOB_TAG, cl_env.comm());
} // send_fin_to


void send_job_to(JobNodePtr job, TC_Alignment*& A, 
		 int host, const MPI_Env& cl_env, TC_ConstraintList*& CL) {
    int L_nseq = job->pNode->left->nseq;
    int R_nseq = job->pNode->right->nseq;
    
    int* L_lseq = job->pNode->left->lseq;
    int* R_lseq = job->pNode->right->lseq;

    int seq_len = 0;

    std::vector<int> tmp(A->S->nseq, 0);

    for (int i = 0; i < A->S->nseq; ++i) {
	tmp[i] = strlen(A->seq_al[i]) + 1;
	seq_len += tmp[i];
    }

    // estimate size of the buffer
    int buf_size = 0;
    int msg_size = 0;

    // 3 = id, Lnseq, Rnseq
    // 3 = A->...
    MPI_Pack_size(3 + 3 + L_nseq + R_nseq + A->S->nseq,
		  MPI_INT, cl_env.comm(), &msg_size);
    buf_size += msg_size;

    MPI_Pack_size(seq_len, MPI_CHAR, cl_env.comm(), &msg_size);
    buf_size += msg_size;
    
    char* buf = new char[buf_size];

    // pack all data
    empi::MPI_ibuffer ibuf(buf_size, buf, cl_env.comm());

    // we do not send nseq
//  ibuf << job->id << L_nseq << R_nseq
//  	 << A->declared_len << A->max_n_seq << A->len_aln;

    ibuf << job->id << L_nseq << R_nseq
  	 << A->declared_len << A->len_aln;

    ibuf << L_lseq, L_lseq + L_nseq;
    ibuf << R_lseq, R_lseq + R_nseq;
    
    // pack sequences
    for (int i = 0; i < A->S->nseq; ++i) {
	ibuf << tmp[i];
	ibuf << A->seq_al[i], A->seq_al[i] + tmp[i];
    }

    MPI_Send(&buf_size, 1, MPI_INT, host, PA_JOB_TAG, cl_env.comm());

    MPI_Send((void*)ibuf.data(), ibuf.position(), MPI_PACKED,
	     host, PA_JOB_TAG + 1, cl_env.comm());

    delete[] buf;

    // store info with job
    // this will be freed when job received
    job->l_0 = new int[L_nseq];
    for (int i = 0; i < L_nseq; ++i) job->l_0[i] = CL->translation[L_lseq[i]];
    job->l_1 = new int[R_nseq];
    for (int i = 0; i < R_nseq; ++i) job->l_1[i] = CL->translation[R_lseq[i]];

#if DEBUG_MPI
    std::cout << "\nMASTER SENT PROGESSIVE JOB " << job->id
	      << " TO " << host << std::endl;
#endif

} // send_job_to


void receive_job(int& jid, int& L_nseq, int*& L_lseq, 
		 int& R_nseq, int*& R_lseq,
		 TC_Alignment*& A, const MPI_Env& cl_env) {
    MPI_Status stat;
    int buf_size = 0;

    MPI_Recv(&buf_size, 1, MPI_INT,
	     cl_env.get_root(), PA_JOB_TAG, cl_env.comm(), &stat);

    // no more job
    if (buf_size == -1) {
#if DEBUG_MPI
	std::cout << "\nWORKER " << cl_env.rank() << " GOT FIN" << std::endl;
#endif
	jid = -1;
	return;
    }

    char* buf = new char[buf_size];
    
    MPI_Recv(buf, buf_size, MPI_PACKED,
	     cl_env.get_root(), PA_JOB_TAG + 1, cl_env.comm(), &stat);

    TC_free_char(A->seq_al);

    empi::MPI_obuffer obuf(buf_size, buf, cl_env.comm());

    int len_aln;
//  obuf >> jid >> L_nseq >> R_nseq
//	 >> A->declared_len >> A->max_n_seq >> A->len_aln;

    obuf >> jid >> L_nseq >> R_nseq
	 >> A->declared_len >> A->len_aln;

    A->max_n_seq = A->S->nseq;

    L_lseq = new int[L_nseq];
    R_lseq = new int[R_nseq];
    
    obuf >> L_lseq, L_lseq + L_nseq;
    obuf >> R_lseq, R_lseq + R_nseq;

    A->seq_al = TC_declare_char(A->max_n_seq, A->declared_len);

    int len = 0;
    
    for (int i = 0; i < A->S->nseq; ++i) {
 	obuf >> len;
 	obuf >> A->seq_al[i], A->seq_al[i] + len;
    }

    delete[] buf;

#if DEBUG_MPI
    std::cout << "\nWORKER " << cl_env.rank()
	      << " RECEIVED PROGRESSIVE JOB " << jid << std::endl;
#endif

} // receive_job


void send_result(int jid, int L_nseq, int R_nseq,
		 TC_Alignment*& A, const MPI_Env& cl_env) {
    int seq_len = 0;
    
    std::vector<int> tmp(A->S->nseq, 0);

    for (int i = 0; i < A->S->nseq; ++i) {
	tmp[i] = strlen(A->seq_al[i]) + 1;
	seq_len += tmp[i];
    }

    // estimate size of the buffer
    int buf_size = 0;
    int msg_size = 0;

    // 3 = id, Lnseq, Rnseq
    // 4 = A->...
    MPI_Pack_size(3 + 4 + L_nseq + R_nseq + A->S->nseq,
		  MPI_INT, cl_env.comm(), &msg_size);
    buf_size += msg_size;

    MPI_Pack_size(seq_len, MPI_CHAR, cl_env.comm(), &msg_size);
    buf_size += msg_size;
    
    char* buf = new char[buf_size];

    // pack all data
    empi::MPI_ibuffer ibuf(buf_size, buf, cl_env.comm());

//  ibuf << jid << L_nseq << R_nseq
//       << A->declared_len << A->max_n_seq << A->len_aln << A->nseq;

    ibuf << jid << L_nseq << R_nseq
	 << A->declared_len << A->len_aln << A->nseq;

    for (int i = 0; i < A->S->nseq; ++i) {
	ibuf << tmp[i];
	ibuf << A->seq_al[i], A->seq_al[i] + tmp[i];
    }

    MPI_Send(&buf_size, 1, MPI_INT,
	     cl_env.get_root(), PA_RES_TAG, cl_env.comm());

    MPI_Send((void*)ibuf.data(), ibuf.position(), MPI_PACKED,
	     cl_env.get_root(), PA_RES_TAG + 1, cl_env.comm());

    delete[] buf;
} // send_result


int receive_result(TC_Alignment*& A, const job_list_type& jobs,
		   const MPI_Env& cl_env) {    
    MPI_Status stat;
    int buf_size = 0;

    MPI_Recv(&buf_size, 1, MPI_INT, 
	     MPI_ANY_SOURCE, PA_RES_TAG, cl_env.comm(), &stat);

    int host = stat.MPI_SOURCE;

    char* buf = new char[buf_size];
    
    MPI_Recv(buf, buf_size, MPI_PACKED,
	     host, PA_RES_TAG + 1, cl_env.comm(), &stat);
    
    int jid = -1;
    int L_nseq = 0;
    int R_nseq = 0;

    int declared_len = 0;
    int max_n_seq = 0;
    int len_aln = 0;
    int nseq = 0;

    empi::MPI_obuffer obuf(buf_size, buf, cl_env.comm());
//  obuf >> jid >> L_nseq >> R_nseq
//	 >> declared_len >> max_n_seq >> len_aln >> nseq;

    obuf >> jid >> L_nseq >> R_nseq
	 >> declared_len >> len_aln >> nseq;

    A->max_n_seq = A->S->nseq;

    // UPDATE BEGINS
    char** seq_al = 0;

    if (A->max_n_seq < max_n_seq) A->max_n_seq = max_n_seq;
    if (A->len_aln < len_aln) A->len_aln = len_aln;

    if (A->declared_len < declared_len) {
	seq_al = TC_declare_char(A->max_n_seq, declared_len);
	for (int i = 0; i < A->S->nseq; ++i) strcpy(seq_al[i], A->seq_al[i]);
	TC_free_char(A->seq_al);
	A->seq_al = seq_al;
	A->declared_len = declared_len;
    }

    // unpack alignment;
    int len = 0;

    seq_al = TC_declare_char(A->max_n_seq, A->declared_len);

    for (int i = 0; i < A->S->nseq; ++i) {
	obuf >> len;
	obuf >> seq_al[i], seq_al[i] + len;
    }

    int* l_0 = jobs[jid].l_0;
    int* l_1 = jobs[jid].l_1;

    // merge alignments
    for (int i = 0; i < L_nseq; ++i) strcpy(A->seq_al[l_0[i]], seq_al[l_0[i]]);
    for (int i = 0; i < R_nseq; ++i) strcpy(A->seq_al[l_1[i]], seq_al[l_1[i]]);

    TC_free_char(seq_al);

    delete[] jobs[jid].l_0;
    delete[] jobs[jid].l_1;

    delete[] buf;

    // update jobs tree
    JobNodePtr ptr = jobs[jid].parent;
    if (ptr != 0) ++(ptr->in_jobs);

    while (ptr != 0) {
	--(ptr->tot_leaf);
	ptr = ptr->parent;
    }
    
#if DEBUG_MPI
    std::cout << "\nMASTER GOT PROGRESSIVE RESULT " << jid
	      << " FROM " << host << std::endl;
#endif

    return host;
} // receive_result


bool create_CL_comm(int ts, const MPI_Env& mpi_env, MPI_Comm* cl_comm) {
    int j_num = ts;
    *cl_comm = MPI_COMM_NULL;

    MPI_Bcast(&j_num, 1, MPI_INT, mpi_env.get_root(), mpi_env.comm());

    if ((mpi_env.size() <= j_num) && (mpi_env.get_mem() == false)) {
	(*cl_comm) = mpi_env.comm();
	return false;
    }
    
    int p = mpi_env.get_proc();

    bool in_grp = false;
    std::vector<int> cl_hosts;

    int i;
    for (i = 0; i < mpi_env.size(); i += p) {
	if (i == mpi_env.rank()) in_grp = true;
	cl_hosts.push_back(i);
	if (cl_hosts.size() == j_num) break;
    }

    if (std::find(cl_hosts.begin(), cl_hosts.end(), mpi_env.get_root())
	== cl_hosts.end()) cl_hosts.push_back(mpi_env.get_root());
    else { 
	if (i + p < mpi_env.size()) {
	    if ((i + p) == mpi_env.rank()) in_grp = true;
	    cl_hosts.push_back(i + p);
	}
    }

    if (mpi_env.am_I_root()) in_grp = true;

    MPI_Group comm_grp, cl_grp;
    MPI_Comm_group(mpi_env.comm(), &comm_grp);
    MPI_Group_incl(comm_grp, cl_hosts.size(), &cl_hosts[0], &cl_grp);
    // do we need to free cl_grp???
    
    MPI_Comm_create(mpi_env.comm(), cl_grp, cl_comm);

    return in_grp;
} // create_CL_comm


void broadcast_CL(TC_ConstraintList*& CL, const MPI_Env& cl_env) {
    typedef short int cl_int_type;

    unsigned int L_dim[3];
    cl_int_type* snd_buf = 0;

    unsigned int tot_mem = 0;
    unsigned int block = 0;

    if (cl_env.am_I_root() == true) {
	L_dim[0] = CL->ne;
	L_dim[1] = CL->entry_len;
    }

    MPI_Bcast(L_dim, 3, MPI_UNSIGNED, cl_env.get_root(), cl_env.comm());

    if (cl_env.am_I_root() == false) {
	CL->L.free();
	CL->L.allocate(L_dim[0], L_dim[1]);
	CL->ne = L_dim[0];
	CL->max_L_len = L_dim[0];
    }

#if DEBUG_MPI
    if (cl_env.am_I_root() == true) {
	std::cout << "\n" << CL->ne << " ENTRIES TO BROADCAST"
		  << std::endl;
    }
#endif

    MPI_Bcast((void*)CL->L.tab, L_dim[0] * L_dim[1], MPI_SHORT,
	      cl_env.get_root(), cl_env.comm());


#if DEBUG_MPI
    if (cl_env.am_I_root() == false) {
	std::cout << "\nWORKER " << cl_env.rank() << " HAS "
		  << CL->ne << " ENTRIES" << std::endl;
    }
#endif	    

    print_CL(CL);

#if DEBUG_MPI
    if (cl_env.am_I_root() == true) {
	std::cout << "\nLIBRARY SENT TO " << cl_env.size() - 1 << " WORKERS" 
		  << std::endl;
    }
#endif
} // broadcast_CL


TC_NT_node* master_par_tree_aln(TC_NT_node** T, TC_Alignment*& A, 
				TC_ConstraintList*& CL,
				const MPI_Env& mpi_env) {
    // CREATE LIST OF JOBS
    std::size_t nseq = (CL->S)->nseq;
    job_list_type job_list(nseq);

    job_list[0].pNode = T[3][0];
    traverse_phyl_tree(T[3][0], 0, job_list);

    // ORAGNISE SCHEDULING QUEUE
    int ts = 0;
    int tot = 0;

    std::deque<JobNodePtr> job_queue;

    for (int i = 0; i < job_list.size(); ++i) {
	if (job_list[i].in_jobs > 1) ++ts;
	if (job_list[i].in_jobs != -1) {
	    job_queue.push_back(&job_list[i]);
	    job_list[i].id = i;
	    job_list[i].tot_leaf =
		job_list[i].l_leaf_num + job_list[i].r_leaf_num;
	    ++tot;
	}
    }

    std::make_heap(job_queue.begin(), job_queue.end(), job_node_compare_neg);

    std::cout << "\tTotal number of jobs:   " << tot << std::endl;
    std::cout << "\tJobs to schedule first: " << ts << std::endl;
    
    // CREATE MPI GROUP TO PERFORM ALIGNMENT
    // cl_comm must be removed by hand
    MPI_Comm cl_comm;
    bool free_cl_comm = create_CL_comm(ts, mpi_env, &cl_comm);

    MPI_Env cl_env(cl_comm);
    cl_env.set_proc(mpi_env.get_proc());

    int cl_root = cl_env.rank();
    MPI_Bcast(&cl_root, 1, MPI_INT, mpi_env.get_root(), mpi_env.comm());

    cl_env.set_root(cl_root);

#if DEBUG_MPI
    std::cout << "\nNEW COMMUNICATOR CREATED WITH " << cl_env.size()
	      << " NODES" << std::endl;
#endif

#if USE_RMA
    CL->residue_index.mpi_env = &mpi_env;
    CL = JZ_weight_constraint_list(CL, "t_coffee", mpi_env);
    CL = AR_index_res_constraint_list(CL, WE);
#else

    // SENT LIBRARY TO NEW GROUP
    broadcast_CL(CL, cl_env);
    CL = TC_weight_constraint_list(CL, "t_coffee");

#if USE_TIMER
    TL.add("Library Broadcasting");
#endif
    
#endif // USE_RMA

    // START SCHEDULING
    int host = 0;
    int to_recv = 0;
    JobNodePtr job = 0;

    // schedule unconstrained jobs
    int to_send = std::min(ts, cl_env.size() - 1);

    // TODO: replace by scatter
    for (int i = 0; i < to_send; ++i, ++to_recv) {
	if (host == cl_root) ++host;

	job = job_queue[0];
	print_job(job);
	std::pop_heap(job_queue.begin(), job_queue.end(), job_node_compare_neg);
	job_queue.pop_back();

	send_job_to(job, A, host, cl_env, CL);
	++host;
    }

    host_cmp hcmp;
    hcmp.ptc_proc = cl_env.get_proc();
    std::priority_queue<int, std::deque<int>, host_cmp> workers(hcmp);

    // start regular scheduling
    while (job_queue.empty() == false) {
	// RECEIVE AND UPDATE
	host = receive_result(A, job_list, cl_env);
	--to_recv;
	
	std::cout << "\t got from " << host;

	std::make_heap(job_queue.begin(), job_queue.end(), 
		       job_node_compare_neg);
		
	job = job_queue[0];


//	workers.push(host);
	

	if (job->in_jobs > 1) {
	    print_job(job);
//	    host = workers.top();
//	    workers.pop();
	    job_queue.pop_front();
	    send_job_to(job, A, host, cl_env, CL);
	    ++to_recv;
	    
	    std::cout << " sent to " << host << ", ";
	    std::cout << "(" << job->l_leaf_num << ","
		      << job->r_leaf_num<< "), "
		      << job_queue.size() << " jobs to schedule\n";
	}
	else {
	    std::cout << ", no ready jobs\n";
	}
    } // while job_queue

    // receive remaining jobs
    do {
	receive_result(A, job_list, cl_env);
	--to_recv;
    }
    while (to_recv > 0);

    for (int i = 0; i < cl_env.size(); ++i) {
	if (i != cl_env.get_root()) send_fin_to(i, cl_env);
    }

    if (free_cl_comm == true) MPI_Comm_free(&cl_comm);


    // THIS IS TO MAINTAIN COMPATIBILITY WITH SEQUENTIAL PART
    // -->
    // last job is scheduled is our solution
    int L_nseq = job->pNode->left->nseq;
    int R_nseq = job->pNode->right->nseq;

    int* L_lseq = job->pNode->left->lseq;
    int* R_lseq = job->pNode->right->lseq;

    job->l_0 = new int[L_nseq];
    for (int i = 0; i < L_nseq; ++i) job->l_0[i] = CL->translation[L_lseq[i]];
    job->l_1 = new int[R_nseq];
    for (int i = 0; i < R_nseq; ++i) job->l_1[i] = CL->translation[R_lseq[i]];

    int i = 0;
    int j = 0;
    for (i, j; i < L_nseq; ++i, ++j) {
	strncpy(A->tree_order[j], (CL->S)->name[job->l_0[i]], STRING - 1);
    }
    for (i = 0; i < R_nseq; ++i, ++j)
	strncpy(A->tree_order[j], (CL->S)->name[job->l_1[i]], STRING - 1);

    delete[] job->l_0;
    delete[] job->l_1;
    // <--

    return 0;
} // master_par_tree_aln


void worker_par_tree_aln(TC_Alignment*& A, TC_ConstraintList*& CL, 
			 const MPI_Env& mpi_env) {
    MPI_Comm cl_comm;
    bool free_cl_comm = create_CL_comm(0, mpi_env, &cl_comm);

    int cl_root;
    MPI_Bcast(&cl_root, 1, MPI_INT, mpi_env.get_root(), mpi_env.comm());

    if (cl_comm != MPI_COMM_NULL) {
	// WORKER GET CL
	MPI_Env cl_env(cl_comm);

	cl_env.set_proc(mpi_env.get_proc());
	cl_env.set_root(cl_root);

#if USE_RMA
	CL->residue_index.mpi_env = &mpi_env;
	CL = JZ_weight_constraint_list(CL, "t_coffee", mpi_env);
	CL = AR_index_res_constraint_list(CL, WE);
#else
	broadcast_CL(CL, cl_env);
	CL = TC_weight_constraint_list(CL, "t_coffee");
#endif

	// HERE MAIN COMPUTATIONAL PART STARTS
	int* translation = CL->translation;

 	int*  n_s;
 	n_s = (int*)AR_allocate(2, sizeof(int));

 	int** l_s;
 	l_s = TC_declare_int(2, A->S->nseq);

	int L_nseq = 0; // LT->nseq
	int R_nseq = 0; // RT->nseq
	
	int* L_lseq = 0; // LT->l_seq
	int* R_lseq = 0; // RT->l_seq

	int jid = -1;

	// DO AS LONG AS THERE ARE JOBS
	do {
	    // receive data
	    receive_job(jid, L_nseq, L_lseq, R_nseq, R_lseq, A, cl_env);

	    if (jid == -1) break;
	    
	    // left sub-tree
	    n_s[0] = L_nseq;
	    for (int i = 0; i < L_nseq; ++i) l_s[0][i] = translation[L_lseq[i]];
    
	    // right sub-tree
	    n_s[1] = R_nseq;
	    for (int i = 0; i < R_nseq; ++i) l_s[1][i] = translation[R_lseq[i]];

 	    print_A(A);
	    print_CL(CL);
	    
	    TC_pair_wise(A, n_s, l_s, CL);

	    delete[] L_lseq;
	    delete[] R_lseq;

#if DEBUG_MPI
	    std::cout << "\nWORKER " << cl_env.rank() << " FINISHED JOB "
		      << jid << std::endl;
#endif

 	    print_A(A);
	    print_CL(CL);

	    send_result(jid, L_nseq, R_nseq, A, cl_env);
	}
	while (true);

	TC_free_int(l_s);
	AR_free(n_s);
    } // if free_cl_comm == true
    else {
	// do this to participate in collective memory
#if USE_RMA
	CL->residue_index.mpi_env = &mpi_env;
	CL = JZ_weight_constraint_list(CL, "t_coffee", mpi_env);
	CL = AR_index_res_constraint_list(CL, WE);
#endif
    }

    if (free_cl_comm == true) MPI_Comm_free(&cl_comm);
} // worker_par_tree_aln


TC_NT_node* AR_par_tree_aln(TC_NT_node** T, TC_Alignment*& A,
			    TC_ConstraintList*& CL,
			    const MPI_Env& mpi_env) {
    TC_NT_node* res = 0;

    // perform common initialization 
    // this part of code comes directly from the sequential version
    // -->
    int nseq = (CL->S)->nseq;

    if (CL->translation) AR_free(CL->translation);
    CL->translation = (int*)AR_allocate(nseq, sizeof(int));
  
    for (int i = 0; i < nseq; ++i) {
	CL->translation[i] = TC_name_is_in_list((CL->S)->name[i],
						(CL->S)->name, nseq, 100);
    }

    A = TC_reorder_aln(A, (CL->S)->name, nseq);
    A->nseq = nseq;
    // <--

    if (mpi_env.am_I_root() == true) {
	std::cout << "\nPROGRESSIVE ALIGNMENT (Tree Based).....\n";

	print_A(A);

	// main method
	res = master_par_tree_aln(T, A, CL, mpi_env);

	print_A(A);
	std::cout << ".....[Done]\n";
    }
    else {
	worker_par_tree_aln(A, CL, mpi_env);
    }

    return res;
} // AR_par_tree_aln
