/*
 * File: util_mpi.cpp
 * Created: 25.07.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <string.h>
#include <MPI_ibuffer.hpp>
#include <MPI_obuffer.hpp>
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "config.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"



TC_Sequences *AR_bcast_seq(TC_Sequences *S, const MPI_Env &mpi_env)
{
	const unsigned int size = 5 * sizeof(int) + sizeof(S->type);
	char *buffer = new char[size];

	if (mpi_env.am_I_root()) {
		empi::MPI_ibuffer ibuf(size, buffer, mpi_env.comm());

		// order matters
		ibuf << S->nseq, S->min_len, S->max_len,
				S->max_nseq, S->contains_gap, S->type, S->type + sizeof(S->type);
	}

	MPI_Bcast(buffer, size, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());

	if (!mpi_env.am_I_root()) {
		empi::MPI_obuffer obuf(size, buffer, mpi_env.comm());
		int n, min, max;

		obuf >> n, min, max;
		S = TC_declare_sequence(min, max, n);

		obuf >> S->max_nseq, S->contains_gap, S->type, S->type + sizeof(S->type);
	}

	delete[] buffer;

	const unsigned int size2 = 4 * (sizeof(int) + STRING + 1) + sizeof(int) + S->max_len + 1 + 10;
	buffer = new char[size2];
	int file_size;
	char *file_buf;

	int pack_pos;
	int len;
	char *str;
	TC_X_template *X;

	if (mpi_env.am_I_root()) {
		pack_pos = 0;

		for (int i = 0; i < S->nseq; ++i) {
			str = S->file[i];
			len = strlen(str) + 1;
			MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());
			MPI_Pack(str, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

			str = S->seq_comment[i];
			len = strlen(str) + 1;
			MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());
			MPI_Pack(str, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

			str = S->aln_comment[i];
			len = strlen(str) + 1;
			MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());
			MPI_Pack(str, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

			str = S->name[i];
			len = strlen(str) + 1;
			MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());
			MPI_Pack(str, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

			len = S->len[i] + 1;
			MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());

			str = S->seq[i];
			MPI_Pack(str, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

			str = (S->T[i])->seq_type;
			MPI_Pack(str, 10, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

#if DEBUG_MPI
			std::cout << "\nBCASTING SEQUENCE '" << S->name[i] << "' (" << S->len[i] << ")\n";
#endif
			MPI_Bcast(buffer, pack_pos, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());
			pack_pos = 0;

			for (; *str; ++str) { // str -> TC_Templates::seq_type
				if (*str == 'P')
					X = (S->T[i])->P;
				else if (*str == 'G')
					X = (S->T[i])->G;
				else if (*str == 'R')
					X = (S->T[i])->R;
				else
					continue;

				file_buf = AR_read_file(AR_check_file_exists(X->template_file), file_size);

				len = strlen(X->template_name) + 1;
				MPI_Pack(&len, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());
				MPI_Pack(X->template_name, len, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

				MPI_Pack(X->template_type, 5, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

				MPI_Pack(&file_size, 1, MPI_INT, buffer, size2, &pack_pos, mpi_env.comm());

				if (*str == 'P')
					MPI_Pack(X->pdb_id, 5, MPI_CHAR, buffer, size2, &pack_pos, mpi_env.comm());

#if DEBUG_MPI
				std::cout << "\nBCASTING SEQUENCE's TEMPLATE '" << X->template_name << "' ("
						  << X->template_type << ")\n";
#endif
				MPI_Bcast(buffer, pack_pos, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());
				MPI_Bcast(file_buf, file_size, MPI_CHAR, mpi_env.get_root(), mpi_env.comm());

				pack_pos = 0;
				AR_free(file_buf);
				file_size = 0;
			}
		}
	}
	else { // WORKER side
		pack_pos = 0;

		for (int i = 0; i < S->nseq; ++i) {
			MPI_Bcast(buffer, size2, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());

			str = S->file[i];
			MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
			MPI_Unpack(buffer, size2, &pack_pos, str, len, MPI_CHAR, mpi_env.comm());

			str = S->seq_comment[i];
			MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
			MPI_Unpack(buffer, size2, &pack_pos, str, len, MPI_CHAR, mpi_env.comm());

			str = S->aln_comment[i];
			MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
			MPI_Unpack(buffer, size2, &pack_pos, str, len, MPI_CHAR, mpi_env.comm());

			str = S->name[i];
			MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
			MPI_Unpack(buffer, size2, &pack_pos, str, len, MPI_CHAR, mpi_env.comm());

			MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
			S->len[i] = len - 1;

			str = S->seq[i];
			MPI_Unpack(buffer, size2, &pack_pos, str, len, MPI_CHAR, mpi_env.comm());

			str = (S->T[i])->seq_type;
			MPI_Unpack(buffer, size2, &pack_pos, str, 10, MPI_CHAR, mpi_env.comm());

			pack_pos = 0;

			for (; *str; ++str) { // str -> TC_Templates::seq_type
				if (*str == 'S' || *str == 's' || *str == ' ')
					continue;

				MPI_Bcast(buffer, size2, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());

				X = (TC_X_template *)AR_allocate(1, sizeof(TC_X_template));

				MPI_Unpack(buffer, size2, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
				MPI_Unpack(buffer, size2, &pack_pos, X->template_name, len, MPI_CHAR, mpi_env.comm());

				MPI_Unpack(buffer, size2, &pack_pos, X->template_type, 5, MPI_CHAR, mpi_env.comm());

				MPI_Unpack(buffer, size2, &pack_pos, &file_size, 1, MPI_INT, mpi_env.comm());

				strncpy(X->seq_name, S->name[i], 100 - 1);
				strncpy(X->template_file, AR_tmpname(), 100 - 1);

				file_buf = new char[file_size];

				MPI_Bcast(file_buf, file_size, MPI_CHAR, mpi_env.get_root(), mpi_env.comm());

				AR_write_file(X->template_file, file_buf, file_size);

				if (*str == 'P') {
					MPI_Unpack(buffer, size2, &pack_pos, X->pdb_id, 5, MPI_CHAR, mpi_env.comm());
					(S->T[i])->P = X;
				}
				else if (*str == 'G') {
					X->S = TC_get_fasta_sequence(X->template_file);
					(S->T[i])->G = X;
				}
				else if (*str == 'R') {
					X->A = TC_main_read_aln(X->template_file, 0);
					X->A = TC_aln2profile(X->A);
					(S->T[i])->R = X;
				}

				pack_pos = 0;
				delete[] file_buf;
				file_size = 0;
			}
		}
	}
	
	delete[] buffer;

	return S;
}
