/*
 * File: util_seq.cpp
 * Created: 06.06.2006
 * Last Modified: 12.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <stdlib.h>
#include <string.h>
#include "globals.hpp"
#include "input.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



void TC_ungap(char *seq)
{
	TC_remove_charset(seq, "ungap");
}

char **TC_ungap_array(char **array, int n)
{
	int a;

	for (a = 0; a < n; a++)
		TC_ungap(array[a]);

	return array;
}

char *TC_thread_aa_seq_on_dna_seq(const char *s)
{
	char *array;
	int l, b, c;

	l = strlen(s);
	array = (char *)AR_allocate(l * 3 + 1, sizeof(char));

	for (b = 0, c = 0; b < l; b++, c += 3) {
		array[c] = s[b];
		array[c + 1] = 'o';
		array[c + 2] = 'o';
	}
	array[c] = '\0';

	return array;
}

int TC_translate_dna_codon(const char *sequence, char stop)
{
	char seq[4];
	int a, b;

	if ((b = strlen(sequence)) < 3) {
		for (a = 0; a < b; a++)
			if (!TC_is_gap(sequence[a]))
				return 'x';
		return '-';
	}
	else {
		seq[0] = tolower(sequence[0]);
		seq[1] = tolower(sequence[1]);
		seq[2] = tolower(sequence[2]);
		seq[3] = '\0';

		seq[0] = (seq[0] == 'u') ? 't' : seq[0];
		seq[1] = (seq[1] == 'u') ? 't' : seq[1];
		seq[2] = (seq[2] == 'u') ? 't' : seq[2];

	}

	if (TC_is_gap(seq[0]) || TC_is_gap(seq[1]) || TC_is_gap(seq[2]))
		return '-';
	else if (TC_strm5(seq, "gca", "gcg", "gcc", "gct", "gcn"))
		return 'a';
	else if (TC_strm2(seq, "tgc", "tgt"))
		return 'c';
	else if (TC_strm2(seq, "gac", "gat"))
		return 'd';
	else if (TC_strm2(seq, "gaa", "gag"))
		return 'e';
	else if (TC_strm2(seq, "ttc", "ttt"))
		return 'f';
	else if (TC_strm5(seq, "gga", "ggg", "ggc", "ggt", "ggn"))
		return 'g';
	else if (TC_strm2(seq, "cac", "cat"))
		return 'h';
	else if (TC_strm3(seq, "ata", "atc", "att"))
		return 'i';
	else if (TC_strm2(seq, "aaa", "aag"))
		return 'k';
	else if (TC_strm6(seq, "cta", "ctg", "ctc", "ctt", "tta", "ttg"))
		return 'l';
	else if (TC_strm(seq, "ctn"))
		return 'l';
	else if (TC_strm(seq, "atg"))
		return 'm';
	else if (TC_strm2(seq, "aac", "aat"))
		return 'n';
	else if (TC_strm5(seq, "cca", "ccg", "ccc", "cct", "ccn"))
		return 'p';
	else if (TC_strm2(seq, "cag", "caa"))
		return 'q';
	else if (TC_strm6(seq, "cga", "cgg", "cgc", "cgt", "aga", "agg"))
		return 'r';
	else if (TC_strm(seq, "cgn"))
		return 'r';
	else if (TC_strm6(seq, "tca", "tcg", "tcc", "tct", "agc", "agt"))
		return 's';
	else if (TC_strm(seq, "ccn"))
		return 's';
	else if (TC_strm5(seq, "aca", "acg", "acc", "act", "acn"))
		return 't';
	else if (TC_strm5(seq, "gta", "gtg", "gtc", "gtt", "gtn"))
		return 'v';
	else if (TC_strm(seq, "tgg"))
		return 'w';
	else if (TC_strm2(seq, "tac", "tat"))
		return 'y';
	else if (TC_strm3(seq, "tag", "taa", "tga"))
		return stop;
	else if (seq[0] == 'n' || seq[1] == 'n' || seq[2] == 'n')
		return stop;
	else {
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown codon '" << seq << "'.\n";
		return 1;
	}
}

/**
 * @brief This function builds an alphabet from the string 'seq'.
 * Alphabet does not need to be empty.
 * The total number of unique symbols is returned.
 * Gaps defined in the GAP_LIST are ignored.
 */
char *TC_get_alphabet(const char *seq, char *alphabet)
{
	const char *s;
	
	if (!seq)
		return alphabet;

	if (!alphabet)
		alphabet = (char *)AR_allocate(200, sizeof(char));
	
	for (s = seq; *s; ++s)
		if (!TC_is_gap(*s) && !TC_is_in_set(*s, alphabet + 1)) {
			alphabet[++alphabet[0]] = *s;
		}
	
	return alphabet;
}

char *TC_get_string_type(const char *S)
{
	static char *ltype;
	char *type;
	int protein = 0, dna = 0, tot = 0;
	int a, l;

	if (!ltype)
		ltype = (char *)AR_allocate(15, sizeof(char));
	type = (char *)AR_allocate(15, sizeof(char));

	l = strlen(S);
	if (l == 0) {
		strncpy(type, "UNKNOWN", 15 - 1);
		return type;
	}
	for (a = 0; a < l; a++) {
		if (!TC_is_gap(S[a])) {
			protein += (TC_is_aa(S[a]) && !TC_is_dna(S[a]));
			dna += TC_is_dna(S[a]);
			++tot;
		}
	}
	protein = (protein * 100) / tot;
	dna = (dna * 100) / tot;

	if (l < 20 && ltype && ltype[0]) {
		if (dna == 100) {
			strncpy(type, ltype, 15 - 1);
			ltype[0] = '\0';
		}
		else
			strncpy(type, "PROTEIN", 15 - 1);
	}
	else if (dna > 98)
		strncpy(type, "DNA", 15 - 1);
	else
		strncpy(type, "PROTEIN", 15 - 1);
	strncpy(ltype, type, 15 - 1);

	return type;
}

int TC_is_aa(char x)
{
	return (TC_is_in_set(x, AA_ALPHABET) && !TC_is_in_set(x, GAP_LIST));
}

int TC_is_dna(char x)
{
	return (TC_is_in_set(x, DNA_ALPHABET) && !TC_is_in_set(x, GAP_LIST));
}

int TC_is_gap(char x)
{
	return TC_is_in_set(x, GAP_LIST);
}

int *TC_seq2inv_pos(const char *seq)
{
	int *pos;
	int a, l1, l2;
	
	l1 = strlen(seq);
	for (l2 = a = 0; a < l1; a++)
		l2 += 1 - TC_is_gap(seq[a]);
	
	pos = (int *)AR_allocate(l2 + 1, sizeof(int));
	
	for (l2 = a = 0; a < l1; a++)
		if (!TC_is_gap(seq[a]))
			pos[++l2] = a + 1;
	
	return pos;
}

int TC_generic_get_seq_sim(const char *seq1, const char *seq2, int *cache, const char *mode)
{
	return TC_get_seq_sim(seq1, seq2, "-", mode);
}

int TC_get_seq_sim(const char *string1, const char *string2, const char *ignore, const char *mode)
{
	int len;
	int pos0, pos1, pos2, sim;
	int p1, p2;
	int r0, r1, r2;
	int a;

	len = strlen(string1);
	for (sim = pos1 = pos2 = pos0 = 0, a = 0; a < len; a++) {
		r1 = string1[a];
		r2 = string2[a];
		p1 = 1 - TC_is_in_set(r1, ignore);
		p2 = 1 - TC_is_in_set(r2, ignore);
		pos1 += p1;
		pos2 += p2;

		if (p1 && p2) {
			++pos0;
			if (TC_is_in_same_group_aa(r1, r2, 0, 0, mode))
				++sim;
		}
	}

	r0 = (pos0 == 0) ? 0 : (sim * MAXID) / pos0;

	return r0;
}

int *TC_get_seq_winsim(const char *string1, const char *string2, const char *ignore,
	const char *mode, int *w)
{
	int len1, len2, len;
	int left, right;
	int a, b;
	int sim = 0;
	int window;
	int r1, r2;

	len1 = strlen(string1);
	len2 = strlen(string2);
	window = atoi(mode);
	len = 2 * window + 1;

	if (len1 != len2)
		return 0;
	if (window == 0 || (window * 2 + 1) >= len1) {
		sim = TC_get_seq_sim(string1, string2, ignore, "");
		for (a = 0; a < len1; a++)
			w[a] = sim;
		return w;
	}

	for (a = 0; a < len1; a++) {
		left = MAX(0, a - window);
		right = MIN(len1, left + len);
		for (sim = 0, b = left; b < right; b++) {
			r1 = string1[b];
			r2 = string2[b];
			if (!TC_is_in_set(r1, ignore) && !TC_is_in_set(r2, ignore)) {
				if (r1 == r2)
					sim++;
			}
		}
		w[a] = (sim * MAXID) / len;
	}
	return w;
}

float TC_get_seq_fsim(const char *string1, const char *string2, const char *ignore,
		const char *similarity_set, int **matrix, int MODE)
{
	float pos = 0, sim = 0;
	int len, a, r1, r2, nr1 = 0, nr2 = 0;
	
	len = MIN(strlen(string1), strlen(string2));
	if (len == 0)
		return 0;
	
	for (a = 0; a < len; a++) {
		r1 = string1[a];
		r2 = string2[a];
		nr1 += !TC_is_gap(r1);
		nr2 += !TC_is_gap(r2);
		
		if (!TC_is_in_set(r1, ignore) && !TC_is_in_set(r2, ignore)) {
			pos++;
			if (matrix)
				sim += matrix[r1 - 'A'][r2 - 'A'];
			else if (TC_is_in_same_group_aa(r1, r2, 0, 0, similarity_set))
				sim++;
		}
	}
	
	if (MODE == UNGAPED_POSITIONS)
		return (sim * 100) / pos;
	else if (MODE == ALIGNED_POSITIONS)
		return (sim * 100) / len;
	else if (MODE == AVERAGE_POSITIONS)
		return (sim * 200) / (nr1 + nr2);
	else
		return 0;
}

/**
 * @param mode indicates which matrix will be used for the grouping
 * @param ngroup pointer to the number of groups
 * @return an array of strings containing the AA of each group
 */
char **TC_make_group_aa(int *ngroup, const char *mode)
{
	char **group_list = 0;
	const char *matrix_name = 0;

	ngroup[0] = 0;
	group_list = TC_declare_char(100, 53);

	if (mode == 0 || mode[0] == '\0')
		matrix_name = "idmat";
	else if (mode[0] == '_') {
		++mode;
		matrix_name = mode;
	}
	else if (TC_strm(mode, "simple")) {
		ngroup[0] = 6;
		strncpy(group_list[0], "avilmAVILM", 53 - 1);
		strncpy(group_list[1], "dekrDEKR", 53 - 1);
		strncpy(group_list[2], "stcnqhSTCNQH", 53 - 1);
		strncpy(group_list[3], "wfyWFY", 53 - 1);
		strncpy(group_list[4], "gG", 53 - 1);
		strncpy(group_list[5], "pP", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "clustalw")) {
		ngroup[0] = 11;
		strncpy(group_list[0], "astaASTA", 53 - 1);
		strncpy(group_list[1], "bneqkBNEQK", 53 - 1);
		strncpy(group_list[2], "cnhqkCNHQK", 53 - 1);
		strncpy(group_list[3], "dndeqDNDEQ", 53 - 1);
		strncpy(group_list[4], "eqhrkEQHRK", 53 - 1);
		strncpy(group_list[5], "fmilvFMILV", 53 - 1);
		strncpy(group_list[6], "gmilfGMILF", 53 - 1);
		strncpy(group_list[7], "hhyHHY", 53 - 1);
		strncpy(group_list[8], "ifywIFYW", 53 - 1);
		strncpy(group_list[9], "jcJC", 53 - 1);
		strncpy(group_list[10], "kpKP", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "polarity")) {
		ngroup[0] = 6;
		strncpy(group_list[0], "eqrsdnkhtEQRSDNKHT", 53 - 1);
		strncpy(group_list[1], "pP", 53 - 1);
		strncpy(group_list[2], "gG", 53 - 1);
		strncpy(group_list[3], "cC", 53 - 1);
		strncpy(group_list[4], "fywFYW", 53 - 1);
		strncpy(group_list[5], "iavlmIAVLM", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "vasiliky")) {
		ngroup[0] = 0;
		strncpy(group_list[ngroup[0]++], "rkRK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "deDE", 53 - 1);
		strncpy(group_list[ngroup[0]++], "qhQH", 53 - 1);
		strncpy(group_list[ngroup[0]++], "vilmVILM", 53 - 1);
		strncpy(group_list[ngroup[0]++], "fyFY", 53 - 1);
		strncpy(group_list[ngroup[0]++], "sS", 53 - 1);
		strncpy(group_list[ngroup[0]++], "wW", 53 - 1);
		strncpy(group_list[ngroup[0]++], "aA", 53 - 1);
		strncpy(group_list[ngroup[0]++], "cC", 53 - 1);
		strncpy(group_list[ngroup[0]++], "gG", 53 - 1);
		strncpy(group_list[ngroup[0]++], "nN", 53 - 1);
		strncpy(group_list[ngroup[0]++], "pP", 53 - 1);
		strncpy(group_list[ngroup[0]++], "tT", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "clustalw_col")) {
		ngroup[0] = 0;
		strncpy(group_list[ngroup[0]++], "staSTA", 53 - 1);
		strncpy(group_list[ngroup[0]++], "neqkNEQK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "nhqkNHQK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "ndeqNDEQ", 53 - 1);
		strncpy(group_list[ngroup[0]++], "qhrkQHRK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "milvMILV", 53 - 1);
		strncpy(group_list[ngroup[0]++], "milfMILF", 53 - 1);
		strncpy(group_list[ngroup[0]++], "hyHY", 53 - 1);
		strncpy(group_list[ngroup[0]++], "fywFYW", 53 - 1);
		strncpy(group_list[ngroup[0]++], "gG", 53 - 1);
		strncpy(group_list[ngroup[0]++], "pP", 53 - 1);
		strncpy(group_list[ngroup[0]++], "cC", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "clustalw_dot")) {
		ngroup[0] = 0;
		strncpy(group_list[ngroup[0]++], "csaCSA", 53 - 1);
		strncpy(group_list[ngroup[0]++], "atvATV", 53 - 1);
		strncpy(group_list[ngroup[0]++], "sagSAG", 53 - 1);
		strncpy(group_list[ngroup[0]++], "stnkSTNK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "stpaSTPA", 53 - 1);
		strncpy(group_list[ngroup[0]++], "sgndSGND", 53 - 1);
		strncpy(group_list[ngroup[0]++], "sndeqkSNDEQK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "ndeqhkNDEQHK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "neqhrkNEQHRK", 53 - 1);
		strncpy(group_list[ngroup[0]++], "fvlimFVLIM", 53 - 1);
		strncpy(group_list[ngroup[0]++], "hfyHFY", 53 - 1);

		return group_list;
	}
	else if (TC_strm(mode, "make_all")) {
		ngroup[0] = 1;
		strncpy(group_list[0], "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", 53 - 1);

		return group_list;
	}
	else
		matrix_name = mode;

	int **matrix;
	int a, b, c, is_in;
	char buf[28];

	matrix = TC_read_matrice(matrix_name);
	for (a = 0; a < 26; ++a) {
		if (matrix[a][a] > 0) {
			for (c = 0, b = 0; b < 26; ++b) {
				if (matrix[a][b] > 0 && matrix[b][b] > 0) {
					buf[c++] = b + 'A';
					buf[c++] = b + 'a';
				}
			}
			buf[c] = '\0';

			for (is_in = 0, b = 0; b < ngroup[0]; ++b)
				if (strcmp(buf, group_list[b]) == 0)
					is_in = 1;

			if (is_in == 0)
				strncpy(group_list[ngroup[0]++], buf, 53 - 1);
		}
	}
	TC_free_int(matrix);

	return group_list;
}

int TC_is_in_same_group_aa(char r1, char r2, int n_group, char **gl, const char *mode)
{
	static char **lgl;
	static int ln_group;
	char **gl2;
	int n_group2;
	int a;

	r1 = toupper(r1);
	r2 = toupper(r2);

	if (mode == 0)
		return (r1 == r2) ? 1 : 0;

	if (TC_strm(mode, "clean")) {
		TC_free_char(lgl);
		lgl = 0;
		ln_group = 0;

		return 0;
	}

	if (lgl == 0)
		lgl = TC_make_group_aa(&ln_group, mode);
	if (gl == 0) {
		gl2 = lgl;
		n_group2 = ln_group;
	}
	else {
		gl2 = gl;
		n_group2 = n_group;
	}

	for (a = 0; a < n_group2; a++)
		if (TC_is_in_set(r1, gl2[a]) && TC_is_in_set(r2, gl2[a]))
			return 1;

	return 0;
}
