/*
 * File: dp_fasta_nw.cpp
 * Created: 06.06.2006
 * Last Modified: 12.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "dp_fasta_nw.hpp"
#include "globals.hpp"
#include "util_algo.hpp"
#include "util_hasch.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"



static TC_hasch_data *TC_allocate_ktup_hasch_data(TC_hasch_data *e, int action)
{
	static TC_hasch_data **heap;
	static int heap_size, free_heap, a;
	
	if (action == 100) {
		std::cout << '\n' << PROGRAM << ": Hasch: heap size= " << heap_size << ", free heap= " << free_heap << '\n';
		
		return 0;
	}
	else if (action == DECLARE) {
		if (free_heap == 0) {
			free_heap = 100;
			heap_size += free_heap;
			heap = (TC_hasch_data **)AR_reallocate(heap, heap_size, sizeof(TC_hasch_data *));
			for (a = 0; a < free_heap; a++) {
				heap[a] = (TC_hasch_data *)AR_allocate(1, sizeof(TC_hasch_data));
				(heap[a])->list = (int *)AR_allocate(10, sizeof(int));
				(heap[a])->list[0] = 10;
			}
		}
		
		return heap[--free_heap];
	}
	else if (action == FREE) {
		heap[free_heap++] = e;
		e->list[1] = 0;
		
		return 0;
	}
	
	return 0;
}

static TC_hasch_data *TC_declare_ktup_hasch_data(TC_hasch_entry *e)
{
	e->data = TC_allocate_ktup_hasch_data(0, DECLARE);
	
	return e->data;
}

static TC_hasch_data *TC_free_ktup_hasch_data(TC_hasch_data *d)
{
	return TC_allocate_ktup_hasch_data(d, FREE);
}

static int TC_hasch_seq(const char *seq, int **hs, int **lu, int ktup, const char *alp)
{
	static int a[10];
	char residue;
	int alp_lu[10000];
	int alp_size;
	int i, j, l, limit, code, flag;
	
	alp_size = alp[0];
	alp++;
	
	for (i = 0; i < alp_size; i++) {
		alp_lu[alp[i]] = i;
	}
	
	l = strlen(seq);
	limit = (int)pow((double)(alp_size + 1), (double)ktup);
	
	hs[0] = (int *)AR_allocate(l + 1, sizeof(int));
	lu[0] = (int *)AR_allocate(limit + 1, sizeof(int));
	
	if (l == 0)
		return 0;
	
	for (i = 1; i <= ktup; i++)
		a[i] = (int)pow((double)(alp_size + 1), (double)(i - 1));
	for (i = 1; i <= (l - ktup + 1); ++i) {
		code = 0;
		flag = FALSE;
		for (j = 1; j <= ktup; ++j) {
			if (TC_is_gap(seq[i + j - 2])) {
				flag = TRUE;
				break;
			}
			else
				residue = alp_lu[seq[i + j - 2]];
			code += residue * a[j];
		}
		if (flag)
			continue;
		
		++code;
		if (lu[0][code])
			hs[0][i] = lu[0][code];
		lu[0][code] = i;
	}
	
	return 0;
}

static TC_hasch_t *TC_hasch_sequence(char *seq1, int ktup)
{
	TC_hasch_t *H;
	TC_hasch_entry *e;
	char c;
	int key, offset = 0, ls;
	
	ls = strlen(seq1);
	H = TC_hcreate(ls, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
	
	while (ls >= ktup) {
		c = seq1[ktup];
		seq1[ktup] = '\0';
		key = TC_string2key(seq1, 0);
		e = TC_hsearch(H, key, FIND, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
		if (e == 0) {
			e = TC_hsearch(H, key, ADD, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
			(e->data)->list[++(e->data)->list[1] + 1] = offset;
		}
		else {
			if ((e->data)->list[0] == ((e->data)->list[1] + 2)) {
				(e->data)->list[0] += 10;
				(e->data)->list = (int *)AR_reallocate((e->data)->list, (e->data)->list[0], sizeof(int));
			}
			(e->data)->list[++(e->data)->list[1] + 1] = offset;
		}
		seq1[ktup] = c;
		seq1++;
		ls--;
		offset++;
	}
	
	return H;
}

/**
 * @brief Reads in an alignmnent A, with two groups of sequences marked.
 * Weight the diagonals with the values read in the constraint list.
 */
static int **TC_evaluate_diagonals_with_clist(TC_Alignment *A, int *ns, int **l_s,
	TC_ConstraintList *CL)
{
	static short int *entry;
	int **diag;
	int **code;
	int **pos;
	int l1, l2, n_diag, s1, s2, r1, r2;
	int a, b, c, d;

	if (!entry)
		entry = (short int *)AR_allocate(CL->entry_len, sizeof(short int));

	// (call makes sense only once)
	CL = TC_index_constraint_list(CL);
	
	l1 = strlen(A->seq_al[l_s[0][0]]);
	l2 = strlen(A->seq_al[l_s[1][0]]);

	n_diag = l1 + l2 - 1;
	diag = TC_declare_int(n_diag + 2, 3);
	for (a = 0; a <= n_diag; a++)
		diag[a][0] = a;

	A->S = CL->S;
	code = TC_seq2aln_pos(A, ns, l_s);
	pos = TC_aln2pos_simple(A, -1, ns, l_s);

	for (a = 0; a < ns[0]; a++) {
		s1 = A->order[l_s[0][a]][0];
		for (b = 0; b < ns[1]; b++) {
			s2 = A->order[l_s[1][b]][0];
			for (c = CL->start_index[s1][s2], d = 0; c < CL->end_index[s1][s2]; c++, d++) {
				entry = TC_extract_entry(entry, c, CL);
				if (s1 == entry[SEQ1]) {
					r1 = code[s1][entry[R1]];
					r2 = code[s2][entry[R2]];
				}
				else if (s2 == entry[SEQ1]) {
					r2 = code[s2][entry[R1]];
					r1 = code[s1][entry[R2]];
				}
				diag[(r2 - r1 + l1)][1] +=
					(CL->get_dp_cost) (A, pos, ns[0], l_s[0], r1 - 1, pos, ns[1], l_s[1], r2 - 1, CL);
			}
		}
	}
	TC_sort_int(diag + 1, 1, 0, n_diag - 1);

	TC_free_int(code);
	TC_free_int(pos);

	return diag;
}

/*
 * Reads in an alignmnet A, with two groups of sequences marked.
 * 1 - Turn each group into a conscensus, using the group list identifier
 *   - if the group list is left empty original symbols are used
 * 2 - hash groupc the two sequences
 * 3 - score each diagonal, sort the list and return it (diag_list)
 */
static int **TC_evaluate_segments_with_ktup(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL,
	int maximise, int n_groups, char **group_list, int ktup)
{
	char *seq1, *seq2, *alphabet = 0;
	int a, b, l1, l2, n_ktup, pos_ktup1, pos_ktup2, **pos;
	int *hasched_seq1, *hasched_seq2, *lu_seq1, *lu_seq2;
	int n_diag, **diag, current_diag, **dot_list, n_dots, cost;
	int l, delta_diag, delta_res;
	
	pos = TC_aln2pos_simple(A, -1, ns, l_s);
	seq1 = TC_aln2cons_seq(A, ns[0], l_s[0], n_groups, group_list);
	seq2 = TC_aln2cons_seq(A, ns[1], l_s[1], n_groups, group_list);
	
	alphabet = TC_get_alphabet(seq1, alphabet);
	alphabet = TC_get_alphabet(seq2, alphabet);
	
	l1 = strlen(seq1);
	l2 = strlen(seq2);
	
	n_diag = l1 + l2 - 1;
	diag = TC_declare_int(n_diag + 2, 3);
	n_ktup = (int)pow((double)alphabet[0] + 1, (double)ktup);
	
	TC_hasch_seq(seq1, &hasched_seq1, &lu_seq1,ktup, alphabet);
	TC_hasch_seq(seq2, &hasched_seq2, &lu_seq2,ktup, alphabet);
	
	// evaluate the diagonals
	for (a = 0; a <= n_diag; a++)
		diag[a][0] = a;
	for (n_dots = 0, a = 1; a <= n_ktup; a++) {
		pos_ktup1 = lu_seq1[a];
		while (TRUE) {
			if (!pos_ktup1)
				break;
			pos_ktup2 = lu_seq2[a];
			while (pos_ktup2) {
				n_dots++;
				pos_ktup2 = hasched_seq2[pos_ktup2];
			}
			pos_ktup1 = hasched_seq1[pos_ktup1];
		}
	}
	
	if (n_dots == 0) {
		AR_free(seq1);
		AR_free(seq2);
		AR_free(alphabet);
		AR_free(hasched_seq1);
		AR_free(hasched_seq2);
		AR_free(lu_seq1);
		AR_free(lu_seq2);
		TC_free_int(diag);
		
		return TC_evaluate_segments_with_ktup(A, ns, l_s, CL, maximise, n_groups, group_list, ktup - 1);
	}
	
	dot_list = TC_declare_int(n_dots, 3);
	
	for (n_dots = 0, a = 1; a <= n_ktup; a++) {
		pos_ktup1 = lu_seq1[a];
		while (TRUE) {
			if (!pos_ktup1)
				break;
			pos_ktup2 = lu_seq2[a];
			while (pos_ktup2) {
				current_diag = (pos_ktup2 - pos_ktup1 + l1);
				dot_list[n_dots][0] = current_diag;
				dot_list[n_dots][1] = pos_ktup1;
				dot_list[n_dots][2] = pos_ktup2;
				pos_ktup2 = hasched_seq2[pos_ktup2];
				n_dots++;
			}
			pos_ktup1 = hasched_seq1[pos_ktup1];
		}
	}
	
	TC_hsort_list_array((void **)dot_list, n_dots, sizeof(int), 3, 0, 3);
	current_diag = (int)dot_list[0][0];
	
	for (b = 0; b < ktup; b++)
		diag[current_diag][2] += (CL->get_dp_cost)(A, pos, ns[0], l_s[0], dot_list[0][1] + b - 1, pos,
				ns[1], l_s[1], dot_list[0][2] + b - 1, CL);
	
	for (l = 0, a = 1; a < n_dots; a++) {
		delta_diag = dot_list[a][0] - dot_list[a - 1][0];
		delta_res = dot_list[a][1] - dot_list[a - 1][1];
		for (cost = 0, b = 0; b < ktup; b++)
			cost++;
		
		if (delta_diag != 0 || FABS(delta_res) > 5) {
			l = 0;
			diag[current_diag][1] = TC_best_of_a_b(diag[current_diag][2], diag[current_diag][1], 1);
			if (diag[current_diag][2] < 0);
			else
				diag[current_diag][1] = MAX(diag[current_diag][1], diag[current_diag][2]);
			diag[current_diag][2] = 0;
			current_diag = dot_list[a][0];
		}
		l++;
		diag[current_diag][2] += cost;
	}
	
	diag[current_diag][1] = TC_best_of_a_b(diag[current_diag][2], diag[current_diag][1], 1);
	TC_sort_int(diag + 1, 1, 0, n_diag - 1);
	
	AR_free(seq1);
	AR_free(seq2);
	AR_free(alphabet);
	AR_free(hasched_seq1);
	AR_free(hasched_seq2);
	AR_free(lu_seq1);
	AR_free(lu_seq2);
	TC_free_int(pos);
	TC_free_int(dot_list);
	
	return diag;
}

/*
 * Ktup comparison as in Rob Edgar, NAR, vol32, No1, 381, 2004
 */
static int **TC_evaluate_diagonals_with_ktup(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL,
	int n_groups, char **group_list, int ktup)
{
	TC_hasch_t *H1, *H2;
	TC_hasch_entry *e1, *e2;
	char character;
	char *s, *sb, *seq1, *seq2;
	int key, ls;
	int l1, l2;
	int score = 0;
	int **diag, n_diag, ktup1, ktup2, a, b, c, d, **pos;
	int n_dots = 0;
	
	pos = TC_aln2pos_simple(A, -1, ns, l_s);
	seq1 = TC_aln2cons_seq(A, ns[0], l_s[0], n_groups, group_list);
	seq2 = TC_aln2cons_seq(A, ns[1], l_s[1], n_groups, group_list);
	l1 = strlen(seq1);
	l2 = strlen(seq2);
	n_diag = l1 + l2 - 1;
	
	diag = TC_declare_int(n_diag + 2, 3);
	for (a = 0; a < n_diag + 2; a++)
		diag[a][0] = a;
	
	H1 = TC_hasch_sequence(seq1, ktup);
	H2 = TC_hasch_sequence(seq2, ktup);
	
	s = sb = (char *)AR_allocate(strlen(seq1) + strlen(seq2) + 1, sizeof(char));
	sprintf(s, "%s%s", seq1, seq2);
	ls = strlen(s);
	
	while (ls >= ktup) {
		character = s[ktup];
		s[ktup] = '\0';
		key = TC_string2key(s, 0);
		e1 = TC_hsearch(H1, key, FIND, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
		e2 = TC_hsearch(H2, key, FIND, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
		if (!e2 || !e1);
		else {
			for (b = 2; b < (e1->data)->list[1] + 2; b++)
				for (c = 2; c < (e2->data)->list[1] + 2; c++) {
					ktup1 = (e1->data)->list[b];
					ktup2 = (e2->data)->list[c];
					diag[(ktup2 - ktup1) + l1][2]++;
					for (score = 0, d = 0; d < ktup; d++)
						score += (CL->get_dp_cost)(A, pos, ns[0], l_s[0], ktup1 + d, pos, ns[1], l_s[1],
								ktup2 + d, CL);
					diag[(ktup2 - ktup1) + l1][1] += score;
					n_dots++;
				}
			(e1->data)->list[1] = (e2->data)->list[1] = 0;
		}
		s[ktup] = character;
		s++;
		ls--;
	}
	
	TC_sort_int(diag + 1, 1, 0, n_diag - 1);
	TC_hdestroy(H1, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
	TC_hdestroy(H2, TC_declare_ktup_hasch_data, TC_free_ktup_hasch_data);
	AR_free(seq1);
	AR_free(seq2);
	AR_free(sb);
	TC_free_int(pos);
	
	return diag;
}

static int **TC_evaluate_diagonals(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL,
	int maximise, int n_groups, char **group_list, int ktup)
{
	int **tot_diag;

	if (CL->L) {
		tot_diag = TC_evaluate_diagonals_with_clist(A, ns, l_s, CL);
	}
	else if (CL->use_fragments) {
		tot_diag = TC_evaluate_segments_with_ktup(A, ns, l_s, CL, maximise, n_groups, group_list, ktup);
	}
	else {
		tot_diag = TC_evaluate_diagonals_with_ktup(A, ns, l_s, CL, n_groups, group_list, ktup);
	}

	return tot_diag;
}

static int *TC_extract_N_diag(int l1, int l2, int **sorted_diag, int n_chosen_diag, int window)
{
	int *slopes;
	int *diag_list;
	int a, b, up, low, current_diag, n_diag;

	n_diag = l1 + l2 - 1;

	diag_list = (int *)AR_allocate(l1 + l2 + 1, sizeof(int));
	slopes = (int *)AR_allocate(n_diag + 1, sizeof(int));

	for (a = n_diag; a > 0 && a > (n_diag - n_chosen_diag); a--) {
		current_diag = sorted_diag[a][0];
		up = MAX(1, current_diag - window);
		low = MIN(n_diag, current_diag + window);

		for (b = up; b <= low; b++)
			slopes[b] = 1;
	}

	// flag bottom right
	up = MAX(1, 1 - window);
	low = MIN(n_diag, 1 + window);
	for (a = up; a <= low; a++)
		slopes[a] = 1;

	// flag top left
	up = MAX(1, (l1 + l2 - 1) - window);
	low = MIN(n_diag, (l1 + l2 - 1) + window);
	for (a = up; a <= low; a++)
		slopes[a] = 1;

	// flag MAIN DIAG SEQ1
	up = MAX(1, l1 - window);
	low = MIN(n_diag, l1 + window);
	for (a = up; a <= low; a++)
		slopes[a] = 1;

	// flag MAIN DIAG SEQ2
	up = MAX(1, l2 - window);
	low = MIN(n_diag, l2 + window);
	for (a = up; a <= low; a++)
		slopes[a] = 1;

	for (a = 0; a <= (l1 + l2 - 1); a++)
		if (slopes[a]) {
			diag_list[++diag_list[0]] = a;
		}

	AR_free(slopes);

	return diag_list;
}

static int *TC_flag_diagonals(int l1, int l2, int **sorted_diag, float T, int window)
{
	int *slopes;
	int *diag_list;
	int use_z_score = 1;
	int a, b, up, low, current_diag, n_diag;
	double mean;
	double sd;
	
	n_diag = l1 + l2 - 1;
	mean = TC_return_mean_int(sorted_diag, n_diag + 1, 1);
	sd = TC_return_sd_int(sorted_diag, n_diag + 1, 1, (int)mean);
	
	if (T == 0) {
		use_z_score = 1;
		T = (((double)sorted_diag[n_diag][1] - mean) / sd) / 25;
	}
	
	diag_list = (int *)AR_allocate(l1 + l2 + 1, sizeof(int));
	slopes = (int *)AR_allocate(n_diag + 1, sizeof(int));
	
	for (a = n_diag; a > 0; --a) {
		current_diag = sorted_diag[a][0];
		if (!use_z_score && sorted_diag[a][1] > T) {
			up = MAX(1, current_diag - window);
			low = MIN(n_diag, current_diag + window);
			for (b = up; b <= low; ++b)
				slopes[b] = 1;
		}
		else if (use_z_score && ((double)sorted_diag[a][1] - mean) / sd > T) {
			up = MAX(1, current_diag - window);
			low = MIN(n_diag, current_diag + window);
			for (b = up; b <= low; ++b)
				slopes[b] = 1;
		}
		else
			break;
	}
	
	for (a = 1, b = 0; a <= n_diag; ++a)
		b += slopes[a];
	
	slopes[1] = 1;
	slopes[l1 + l2 - 1] = 1;
	slopes[l2] = 1;
	
	for (a = 0; a <= (l1 + l2 - 1); ++a)
		if (slopes[a])
			diag_list[++diag_list[0]] = a;
	
	AR_free(slopes);
	
	return diag_list;
}


static int TC_make_fasta_gotoh_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL,
	int *diag)
{
	// treatment of the terminal gap penalties
	// TG_MODE = 0 --> gop and gep
	// TG_MODE = 1 --> ---     gep
	// TG_MODE = 2 --> ---     ---
	int TG_MODE, gop, l_gop, gep, l_gep, maximise;
	// variables for multiple sequence alignment
	char *buffer, *char_buf;
	char **aln, **al;
	int lenal[2], len;
	int **C, **D, **I, **trace, **pos0, **LD;
	int l1, l2, eg, ch, sub, score = 0, fop;
	int last_i, last_j, i, delta_i, j, pos_j, ala, alb, LEN, n_diag, match1, match2;
	int a, b, c, k, t;

	// prepare penalties
	gop = CL->gop * SCORE_K;
	gep = CL->gep * SCORE_K;
	TG_MODE = CL->TG_MODE;
	maximise = CL->maximise;

	n_diag = diag[0];
	l1 = lenal[0] = strlen(A->seq_al[l_s[0][0]]);
	l2 = lenal[1] = strlen(A->seq_al[l_s[1][0]]);

	/* diag:
		diag[1..n_diag] --> flaged diagonal in order;
		diag[0] = 0     --> first diagonal;
		diag[n_diag+1] = l1+l2-1;
	 */

	// numeration of the diagonals strats from the bottom right [1...l1+l2-1]
	// sequence s1 is vertical and seq s2 is horizontal
	// D contains the best Deletion  in S2 => comes from diagonal N+1
	// I contains the best Insertion in S2 => comes from diagonal N-1

	C = TC_declare_int(lenal[0] + lenal[1] + 1, n_diag + 2);
	D = TC_declare_int(lenal[0] + lenal[1] + 1, n_diag + 2);
	LD = TC_declare_int(lenal[0] + lenal[1] + 1, n_diag + 2);
	I = TC_declare_int(lenal[0] + lenal[1] + 1, n_diag + 2);
	trace = TC_declare_int(lenal[0] + lenal[1] + 1, n_diag + 2);

	al = TC_declare_char(2, lenal[0] + lenal[1] + lenal[1] + 1);

	len = MAX(lenal[0], lenal[1]) + 1;
	buffer = (char *)AR_allocate(2 * len, sizeof(char));
	char_buf = (char *)AR_allocate(2 * len, sizeof(char));

	pos0 = TC_aln2pos_simple(A, -1, ns, l_s);
	C[0][0] = 0;

	t = (TG_MODE == 0) ? gop : 0;
	for (j = 1; j <= n_diag; j++) {
		l_gop = (TG_MODE == 0) ? gop : 0;
		l_gep = (TG_MODE == 2) ? 0 : gep;

		if ((diag[j] - lenal[0]) < 0) {
			trace[0][j] = TC_UNDEFINED;
			continue;
		}
		C[0][j] = (diag[j] - lenal[0]) * l_gep + l_gop;
		D[0][j] = (diag[j] - lenal[0]) * l_gep + l_gop + gop;
	}
	D[0][j] = D[0][j - 1] + gep;

	t = (TG_MODE == 0) ? gop : 0;
	for (i = 1; i <= lenal[0]; i++) {
		l_gop = (TG_MODE == 0) ? gop : 0;
		l_gep = (TG_MODE == 2) ? 0 : gep;

		C[i][0] = C[i][n_diag + 1] = t = t + l_gep;
		I[i][0] = D[i][n_diag + 1] = t + gop;

		for (j = 1; j <= n_diag; j++) {
			C[i][j] = C[i][0];
			D[i][j] = I[i][j] = I[i][0];
		}

		for (eg = 0, j = 1; j <= n_diag; j++) {
			pos_j = diag[j] - lenal[0] + i;
			if (pos_j <= 0 || pos_j > l2) {
				trace[i][j] = TC_UNDEFINED;
				continue;
			}
			sub = (CL->get_dp_cost)(A, pos0, ns[0], l_s[0], i - 1, pos0, ns[1], l_s[1], pos_j - 1, CL);

			// 1. Identify the best insertion in S2
			l_gop = (i == lenal[0]) ? ((TG_MODE == 0) ? gop : 0) : gop;
			l_gep = (i == lenal[0]) ? ((TG_MODE == 2) ? 0 : gep) : gep;
			len = (j == 1) ? 0 : (diag[j] - diag[j - 1]);
			if (TC_a_better_than_b(I[i][j - 1], C[i][j - 1] + l_gop, maximise))
				eg++;
			else
				eg = 1;
			I[i][j] = TC_best_of_a_b(I[i][j - 1], C[i][j - 1] + l_gop, maximise) + len * l_gep;

			// 2. Identify the best deletion in S2
			l_gop = (pos_j == lenal[1]) ? ((TG_MODE == 0) ? gop : 0) : gop;
			l_gep = (pos_j == lenal[1]) ? ((TG_MODE == 2) ? 0 : gep) : gep;

			len = (j == n_diag) ? 0 : (diag[j + 1] - diag[j]);
			delta_i = ((i - len) > 0) ? (i - len) : 0;

			if (TC_a_better_than_b(D[delta_i][j + 1], C[delta_i][j + 1] + l_gop, maximise)) {
				LD[i][j] = LD[delta_i][j + 1] + 1;
			}
			else {
				LD[i][j] = 1;
			}
			D[i][j] = TC_best_of_a_b(D[delta_i][j + 1], C[delta_i][j + 1] + l_gop, maximise)
				+ len * l_gep;

			// Identify the best way
			score = C[i][j] = TC_best_int(3, maximise, &fop, I[i][j], C[i - 1][j] + sub, D[i][j]);

			fop -= 1;
			if (fop < 0)
				trace[i][j] = fop * eg;
			else if (fop > 0) {
				trace[i][j] = fop * LD[i][j];
			}
			else if (fop == 0)
				trace[i][j] = 0;

			last_i = i;
			last_j = j;
		}
	}

	/*
	            [0][Positive]
	             ^     ^
	             |    /
	             |   /
	             |  /
	             | /
	             |/
	 [Neg]<-----[*]
	*/

	i = last_i;
	j = last_j;

	ala = alb = 0;
	match1 = match2 = 0;
	while (!(match1 == l1 && match2 == l2)) {
		if (match1 == l1) {
			len = l2 - match2;
			for (a = 0; a < len; a++) {
				al[0][ala++] = 0;
				al[1][alb++] = 1;
				match2++;
			}
			k = 0;
			break;
		}
		else if (match2 == l2) {
			len = l1 - match1;
			for (a = 0; a < len; a++) {
				al[0][ala++] = 1;
				al[1][alb++] = 0;
				match1++;
			}
			k = 0;
			break;
		}
		else {
			k = trace[i][j];
		}
		if (k == 0) {
			if (match2 == l2 || match1 == l1);
			else {
				al[0][ala++] = 1;
				al[1][alb++] = 1;
				i--;
				match1++;
				match2++;
			}
		}
		else if (k > 0) {
			len = diag[j + k] - diag[j];
			for (a = 0; a < len; a++) {
				if (match1 == l1)
					break;
				al[0][ala++] = 1;
				al[1][alb++] = 0;
				match1++;
			}
			i -= len;
			j += k;
		}
		else if (k < 0) {
			k *= -1;
			len = diag[j] - diag[j - k];
			for (a = 0; a < len; a++) {
				if (match2 == l2)
					break;
				al[0][ala++] = 0;
				al[1][alb++] = 1;
				match2++;
			}
			j -= k;
		}
	}

	LEN = ala;
	c = LEN - 1;
	TC_invert_list_char(al[0], LEN);
	TC_invert_list_char(al[1], LEN);
	if (A->declared_len <= LEN)
		A = TC_realloc_aln2(A, A->max_n_seq, 2 * LEN);
	aln = A->seq_al;

	for (c = 0; c < 2; c++) {
		for (a = 0; a < ns[c]; a++) {
			ch = 0;
			for (b = 0; b < LEN; b++) {
				if (al[c][b] == 1)
					char_buf[b] = aln[l_s[c][a]][ch++];
				else
					char_buf[b] = '-';
			}
			char_buf[b] = '\0';
			strncpy(aln[l_s[c][a]], char_buf, A->declared_len - 1);
		}
	}

	A->len_aln = LEN;
	A->nseq = ns[0] + ns[1];

	TC_free_int(pos0);
	TC_free_int(C);
	TC_free_int(D);
	TC_free_int(I);
	TC_free_int(trace);
	TC_free_int(LD);
	TC_free_char(al);
	AR_free(buffer);
	AR_free(char_buf);

	return score;
}

int TC_fasta_gotoh_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	int maximise;
	// variables for multiple sequence alignment
	static char **group_list;
	static int n_groups;
	int **tot_diag;
	int *diag;
	int ktup;
	int score;
	float diagonal_threshold;
	
	// prepare penalties
	maximise = CL->maximise;
	ktup = CL->ktup;
	diagonal_threshold = CL->diagonal_threshold;
	
	if (!group_list)
		group_list = TC_make_group_aa(&n_groups, CL->matrix_for_aa_group);
	
	tot_diag = TC_evaluate_diagonals(A, ns, l_s, CL, maximise, n_groups, group_list, ktup);
	
	if (!CL->fasta_step)
		diag = TC_flag_diagonals(strlen(A->seq_al[l_s[0][0]]), strlen(A->seq_al[l_s[1][0]]),
				tot_diag, diagonal_threshold, 0);
	else
		diag = TC_extract_N_diag(strlen(A->seq_al[l_s[0][0]]), strlen (A->seq_al[l_s[1][0]]),
				tot_diag, CL->fasta_step, 0);
	
	score = TC_make_fasta_gotoh_pair_wise(A, ns, l_s, CL, diag);
	
	TC_free_int(tot_diag);
	AR_free(diag);
	
	return score;
}

int TC_cfasta_gotoh_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	int maximise;
	// variables for multiple sequence alignment
	static int n_groups;
	static char **group_list;
	int **tot_diag;
	int *diag;
	int ktup;
	int score, new_score;
	int n_chosen_diag = 20;
	int step;
	int max_n_chosen_diag;
	int l1, l2;

	// prepare penalties
	maximise = CL->maximise;
	ktup = CL->ktup;

	if (!group_list) {
		group_list = TC_make_group_aa(&n_groups, CL->matrix_for_aa_group);
	}

	l1 = strlen(A->seq_al[l_s[0][0]]);
	l2 = strlen(A->seq_al[l_s[1][0]]);

	if (!CL->fasta_step) {
	    step = MIN(l1, l2);
	    step = (int)log((double)MAX(step, 1));
	    step = MAX(step, 20);
	}
	else {
		step = CL->fasta_step;
	}

	tot_diag = TC_evaluate_diagonals(A, ns, l_s, CL, maximise, n_groups, group_list, ktup);
	max_n_chosen_diag = strlen(A->seq_al[l_s[0][0]]) + strlen(A->seq_al[l_s[1][0]]) - 1;
	n_chosen_diag += step;
	n_chosen_diag = MIN(n_chosen_diag, max_n_chosen_diag);

	diag = TC_extract_N_diag(strlen(A->seq_al[l_s[0][0]]), strlen(A->seq_al[l_s[1][0]]), tot_diag,
		n_chosen_diag, 0);

	score = TC_make_fasta_gotoh_pair_wise(A, ns, l_s, CL, diag);

	new_score = 0;
	AR_free(diag);

	while (new_score != score && n_chosen_diag < max_n_chosen_diag) {
		score = new_score;
		TC_ungap_sub_aln(A, ns[0], l_s[0]);
		TC_ungap_sub_aln(A, ns[1], l_s[1]);

		n_chosen_diag += step;
		n_chosen_diag = MIN(n_chosen_diag, max_n_chosen_diag);
		diag = TC_extract_N_diag(strlen(A->seq_al[l_s[0][0]]), strlen(A->seq_al[l_s[1][0]]), tot_diag,
			n_chosen_diag, 0);

		new_score = TC_make_fasta_gotoh_pair_wise(A, ns, l_s, CL, diag);

		AR_free(diag);
	}
	score = new_score;

	TC_free_int(tot_diag);

	return score;
}

int TC_very_fast_gotoh_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	// treatment of the terminal gap penalties
	// TG_MODE = 0 --> gop and gep
	// TG_MODE = 1 --> ---     gep
	// TG_MODE = 2 --> ---     ---
	int maximise;
	// variables for the multiple sequence alignment
	static int n_groups;
	static char **group_list;
	int **tot_diag;
	int *diag;
	int ktup;
	int score;
	
	// prepare penalties
	maximise = CL->maximise;
	ktup = CL->ktup;
	
	if (!group_list) {
		group_list = TC_make_group_aa(&n_groups, CL->matrix_for_aa_group);
	}
	CL->use_fragments = 0;
	tot_diag = TC_evaluate_diagonals(A, ns, l_s, CL, maximise, n_groups, group_list, ktup);
	
	// Note: 20 diagonals. 5 shadows on each side: tunned on Hom39, 2/2/04
	diag = TC_extract_N_diag(strlen(A->seq_al[l_s[0][0]]), strlen(A->seq_al[l_s[1][0]]), tot_diag,
			20, 5);
	
	score = TC_make_fasta_gotoh_pair_wise(A, ns, l_s, CL, diag);
	
	TC_free_int(tot_diag);
	AR_free(diag);
	
	return score;
}

int TC_precomputed_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
	int l1, l2, a, b, c;
	int nid = 0, npos = 0, id;
	int r1, r2, s1, s2;
	
	l1 = strlen(A->seq_al[l_s[0][0]]);
	l2 = strlen(A->seq_al[l_s[1][0]]);
	
	if (l1 != l2) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Improper use of the function precomputed_pair_wise.\n";
		A->score_aln = A->score = 0;
		return 0;
	}
	else if (l1 == 0) {
		A->score_aln = A->score = 0;
		return 0;
	}
	
	for (npos = 0, nid = 0, a = 0; a < ns[0]; a++) {
		s1 = l_s[0][a];
		for (b = 0; b < ns[1]; b++) {
			s2 = l_s[1][b];
			for (c = 0; c < l1; c++) {
				r1 = A->seq_al[s1][c];
				r2 = A->seq_al[s2][c];
				if (TC_is_gap(r1) || TC_is_gap(r2));
				else {
					npos++;
					nid += (r1 == r2);
				}
			}
		}
	}
	id = (npos == 0) ? 0 : ((nid * 100) / npos);
	A->score = A->score_aln = id;
	
	return A->score;
}
